/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu
  

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/




#include "PLA.h"
#include <math.h>
#include <stdio.h>
#include "PLA_Blas1.h"

/******************************************************************************/

/* Note : This routine expects the user to have allocated the space for */
/*        the 1x1 multiscalar that will be returned                     */
/* Note : More importantly, for simplicity, this assumes that the two   */
/*        vectors are aligned and distributed identically.              */
/* Note : To be tiresome.  This function also assumes that all procs. in*/
/*        the communicator of the Mscalar get the answer back.          */

int PLA_Dot(PLA_Obj X, PLA_Obj Y, PLA_Obj Result)
{
   MPI_Datatype       datatype;
   int                obj_type_x, obj_type_y;
   int                orient_x, orient_y;
   int                row_x, row_y, col_x, col_y;
   int                length_x, width_x, length_y, width_y;
   int                length_result, width_result;
   int                check_x, check_y; /* will go in enter/exit later */
   PLA_Template       template;
   PLA_Obj            Result_dup = NULL, Conformal_X = NULL;

   PLA_Obj_get_orientation(X, &orient_x);
   PLA_Obj_get_orientation(Y, &orient_y);
   PLA_Obj_global_align_row(X, &row_x);
   PLA_Obj_global_align_col(X, &col_x);
   PLA_Obj_global_align_row(Y, &row_y);
   PLA_Obj_global_align_col(Y, &col_y);
   PLA_Obj_objtype(X, &obj_type_x);
   PLA_Obj_objtype(Y, &obj_type_y);
	PLA_Obj_datatype (Result, &datatype);
	PLA_Obj_template (Result, &template);

/* The dimension check stuff will go in an enter exit routine later */
/* FOR now I want it "turned on" only here                          */
   PLA_Obj_global_length(X, &length_x);
   PLA_Obj_global_width(X, &width_x);
   PLA_Obj_global_length(Y, &length_y);
   PLA_Obj_global_width(Y, &width_y);
   PLA_Obj_global_length(Result, &length_result);
   PLA_Obj_global_width(Result, &width_result);
   if(1 == length_x) check_x = width_x;
   else if(1 == width_x) check_x = length_x;
   else {printf("X is not size 1 in either dimension PLA_Dot\n"); return 1;}
   if(1 == length_y) check_y = width_y;
   else if(1 == width_y) check_y = length_y;
   else {printf("Y is not size 1 in either dimension PLA_Dot\n"); return 1;}
   if( (1 != length_result) || (1 != width_result) )
   { printf("Result is not 1x1 in PLA_Dot\n"); return 1;}
/* If the objects are not identical -- need a copy of one so the local call will work */
   if((obj_type_x != obj_type_y) || (orient_x != orient_y) ||
      (row_x != row_y) || (col_x != col_y) )
   {
      PLA_Obj_clone(X, &Conformal_X);
      PLA_Copy(Y, Conformal_X);
   }
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
					  template, &Result_dup);
   if(NULL == Conformal_X)
      PLA_Local_dot(X, Y, Result_dup);
   else PLA_Local_dot(X, Conformal_X, Result_dup);
   PLA_Reduce(Result_dup, MPI_SUM, Result);
   PLA_Obj_free(&Result_dup);
   PLA_Obj_free(&Conformal_X);
}

int PLA_Dot_add(PLA_Obj X, PLA_Obj Y, PLA_Obj Result)
{
   MPI_Datatype       datatype;
   int                obj_type_x, obj_type_y;
   int                orient_x, orient_y;
   int                row_x, row_y, col_x, col_y;
   int                length_x, width_x, length_y, width_y;
   int                length_result, width_result;
   int                check_x, check_y; /* will go in enter/exit later */
   PLA_Template       template;
   PLA_Obj            Result_dup = NULL, Conformal_X = NULL;

   PLA_Obj_get_orientation(X, &orient_x);
   PLA_Obj_get_orientation(Y, &orient_y);
   PLA_Obj_global_align_row(X, &row_x);
   PLA_Obj_global_align_col(X, &col_x);
   PLA_Obj_global_align_row(Y, &row_y);
   PLA_Obj_global_align_col(Y, &col_y);
   PLA_Obj_objtype(X, &obj_type_x);
   PLA_Obj_objtype(Y, &obj_type_y);
	PLA_Obj_datatype (Result, &datatype);
	PLA_Obj_template (Result, &template);

/* The dimension check stuff will go in an enter exit routine later */
/* FOR now I want it "turned on" only here                          */
   PLA_Obj_global_length(X, &length_x);
   PLA_Obj_global_width(X, &width_x);
   PLA_Obj_global_length(Y, &length_y);
   PLA_Obj_global_width(Y, &width_y);
   PLA_Obj_global_length(Result, &length_result);
   PLA_Obj_global_width(Result, &width_result);
   if(1 == length_x) check_x = width_x;
   else if(1 == width_x) check_x = length_x;
   else {printf("X is not size 1 in either dimension PLA_Dot\n"); return 1;}
   if(1 == length_y) check_y = width_y;
   else if(1 == width_y) check_y = length_y;
   else {printf("Y is not size 1 in either dimension PLA_Dot\n"); return 1;}
   if( (1 != length_result) || (1 != width_result) )
   { printf("Result is not 1x1 in PLA_Dot\n"); return 1;}
/* If the objects are not identical -- need a copy of one so the local call will work */
   if((obj_type_x != obj_type_y) || (orient_x != orient_y) ||
      (row_x != row_y) || (col_x != col_y) )
   {
      PLA_Obj_clone(X, &Conformal_X);
      PLA_Copy(Y, Conformal_X);
   }
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
								  template, &Result_dup);
   if(NULL == Conformal_X)
      PLA_Local_dot(X, Y, Result_dup);
   else PLA_Local_dot(X, Conformal_X, Result_dup);
   PLA_Reduce_add(Result_dup, MPI_SUM, Result);
   PLA_Obj_free(&Result_dup);
   PLA_Obj_free(&Conformal_X);
}
int PLA_Dot_x(PLA_Obj alpha, PLA_Obj X, PLA_Obj Y, PLA_Obj beta, PLA_Obj Result)
{
   MPI_Datatype       datatype;
   int                obj_type_x, obj_type_y;
   int                orient_x, orient_y;
   int                row_x, row_y, col_x, col_y;
   int                length_x, width_x, length_y, width_y;
   int                length_result, width_result;
   int                check_x, check_y; /* will go in enter/exit later */
   int                owner_row, owner_col;
   PLA_Template       template;
   PLA_Obj            Result_dup = NULL, Conformal_X = NULL;
   PLA_Obj            alpha_dup = NULL, beta_dup = NULL;

   PLA_Obj_get_orientation(X, &orient_x);
   PLA_Obj_get_orientation(Y, &orient_y);
   PLA_Obj_global_align_row(X, &row_x);
   PLA_Obj_global_align_col(X, &col_x);
   PLA_Obj_global_align_row(Y, &row_y);
   PLA_Obj_global_align_col(Y, &col_y);
   PLA_Obj_objtype(X, &obj_type_x);
   PLA_Obj_objtype(Y, &obj_type_y);
   PLA_Obj_datatype (Result, &datatype);
   PLA_Obj_template (Result, &template);

/* The dimension check stuff will go in an enter exit routine later */
/* FOR now I want it "turned on" only here                          */
   PLA_Obj_global_length(X, &length_x);
   PLA_Obj_global_width(X, &width_x);
   PLA_Obj_global_length(Y, &length_y);
   PLA_Obj_global_width(Y, &width_y);
   PLA_Obj_global_length(Result, &length_result);
   PLA_Obj_global_width(Result, &width_result);
   if(1 == length_x) check_x = width_x;
   else if(1 == width_x) check_x = length_x;
   else {printf("X is not size 1 in either dimension PLA_Dot\n"); return 1;}
   if(1 == length_y) check_y = width_y;
   else if(1 == width_y) check_y = length_y;
   else {printf("Y is not size 1 in either dimension PLA_Dot\n"); return 1;}
   if( (1 != length_result) || (1 != width_result) )
   { printf("Result is not 1x1 in PLA_Dot\n"); return 1;}

   if((obj_type_x != obj_type_y) || (orient_x != orient_y) ||
      (row_x != row_y) || (col_x != col_y) )
   {
      PLA_Obj_clone(X, &Conformal_X);
      PLA_Copy(Y, Conformal_X);
   }
   
   PLA_Obj_owner_row(alpha, &owner_row);
   PLA_Obj_owner_col(alpha, &owner_col);
   if( (PLA_ALL_ROWS != owner_row) || (PLA_ALL_COLS != owner_col) )
   {
      PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
				 template, &alpha_dup);
      PLA_Copy(alpha, alpha_dup);
   }
   
   PLA_Obj_owner_row(beta, &owner_row);
   PLA_Obj_owner_col(beta, &owner_col);
   if( (PLA_ALL_ROWS != owner_row) || (PLA_ALL_COLS != owner_col) )
   {
      PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
				 template, &beta_dup);
      PLA_Copy(beta, beta_dup);
   }
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1,
				 template, &Result_dup);
   if(NULL == Conformal_X)
      PLA_Local_dot(X, Y, Result_dup);
   else PLA_Local_dot(X, Conformal_X, Result_dup);
   PLA_Scal(alpha_dup == NULL ? alpha : alpha_dup, Result_dup);
   PLA_Scal(beta_dup == NULL ? beta : beta_dup, Result);
   PLA_Reduce_add(Result_dup, MPI_SUM, Result);
   PLA_Obj_free(&Result_dup);
   PLA_Obj_free(&Conformal_X);
}
