/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu


  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/



#include "PLA.h"
#include "mpi.h" /* doesn't pla.h do this? */
#include <stdio.h>
#include <math.h>

#ifndef min3
#define min3(x,y,z) min(min(x,y), z)
#endif
/* Note that these defines SHOULD be in MISC_DEFS.h or PLA.h  */
/* I put them here to get around the compiler problem and to  */
/* point out that they are missing (the defines above that is)*/


/* Strange abs behavior made this necessary -- it really is strange */
double zabs(double* x)
{
   return( ( (*(double*)x)     *  (*(double*)x) ) +
           ( (*((double*)x+1)) *  (*((double*)x+1)) ) );
}
float c_abs(float* x)
{
   return( ( (*(float*)x)     *  (*(float*)x) ) +
           ( (*((float*)x+1)) *  (*((float*)x+1)) ) );
}
double dabs(double x)
{
   return(x < 0 ? -1.0*x : x);
}
float f_abs(float x)
{
   return(x < 0 ? -1.0*x : x);
}

int PLA_Iamax( PLA_Obj x, PLA_Obj xmax, PLA_Obj k );

/* This routine -- will only work on things of width == 1 */
/* Originally the "efficient" version assumed :           */
/* stride-one and assumes column major storage (so that   */
/* the objects could be checked locally by indexing -not  */
/* repeated splitting.                                    */

/* NOTE : There are at least two ways to let EVERYONE     */
/* know the xmax, k information.  For the case of columns */
/* rows (projected) - you can determine it in rows after  */
/* Allgather in rows then Bcast or you can Allgather over */
/* the whole grid and have everyone determine locally.    */
/* This routine does it the first way -- but the second   */
/* is definitely easier so it can be changed to that.     */

/* Outline of algorithm : Slide through the object one    */
/* data atom at a time.  Keeping count of how many        */
/* data atoms have been seen so far (globally).  Have     */
/* all processes communicate (All_gather) their abs(Max)  */
/* and the global index where it was seen.  Then have     */
/* all processes locally search this for the global absMax*/
/* and its global location.  Placing these values in xmax */
/* and k, respectively.  Note that xmax and k had better  */
/* be totally duplicated Mscalars.  Further note that     */
/* if we are dealing with a single row or column (of a    */
/* matrix or a Pvector/Pmvector-of-one-dimesnsion-size-1  */
/* then we do the search on the given row or column of    */
/* processors and then communicate this information to the*/
/* others.  This makes the code considerably messier      */
/* at the end.                                            */

int PLA_Iamax( PLA_Obj x, PLA_Obj xmax, PLA_Obj k )
{
   MPI_Datatype datatype_x;
   MPI_Comm     communicator;
   PLA_Template template;
   int          local_stride_x, myrank, location_temp, location,myrank2, PLA_rank, idummy;
   int          split_side, k_cur, row_or_column, x_size, k_temp=0,incx;
   int          comm_count, comm_count2, i, junk_size, holder_rank, size, owner;
   PLA_Obj    x_cur = NULL, x_top = NULL, x_rest = NULL, x_leftover = NULL;
   PLA_Obj    junk = NULL;
   void        *buf_x, *value_buffer, *index_buffer, *k_buffer;
   void        *x_max_buffer;
   PLA_Obj_obj_type objtype_x;
   PLA_Start_num zero_or_one; /* should be an int */
   double d_temp, d_temp2;
   float  f_temp, f_temp2;

   PLA_Obj_datatype( x, &datatype_x );
   PLA_Obj_objtype (x, &objtype_x);
   PLA_Obj_template( x, &template );
   PLA_Obj_local_stride( x, &local_stride_x ); /* not used */
   PLA_Temp_zero_or_one( template, &zero_or_one );
   PLA_Obj_local_buffer(k, &k_buffer);
   PLA_Obj_local_buffer(xmax, &x_max_buffer);
/* Next line is a DEBUG line -- only uses PLA_rank */
   PLA_Temp_comm_all_info(template, &communicator, &PLA_rank, &size);
   PLA_Obj_view_all( x, &x_cur );

   *((int*)k_buffer) = (int)zero_or_one;
   k_cur = *((int*)k_buffer);
   k_temp = k_cur;
   location = -1;


   if( (PLA_MATRIX == objtype_x) || (PLA_PMVECTOR == objtype_x) ||
       (PLA_PVECTOR == objtype_x))
   {
       PLA_Obj_project_onto(x, &row_or_column);
      if(PLA_PROJ_ONTO_ROW == row_or_column)
      {
printf("ROW_MATRIX in Iamax\n");
         split_side = PLA_SIDE_LEFT;
         PLA_Temp_comm_col_info(template, &communicator, &myrank, &comm_count2);
         PLA_Temp_comm_row_info(template, &communicator, &myrank2, &comm_count);
         PLA_Obj_split_size(x, PLA_SIDE_TOP, &junk_size, &holder_rank);
      }
      else if(PLA_PROJ_ONTO_COL == row_or_column)
      {
printf("COL_MATRIX in Iamax\n");
         split_side = PLA_SIDE_TOP;
         PLA_Temp_comm_row_info(template, &communicator, &myrank, &comm_count2);
         PLA_Temp_comm_col_info(template, &communicator, &myrank2, &comm_count);
         PLA_Obj_split_size(x, PLA_SIDE_LEFT, &junk_size, &holder_rank);
      }
   }
   else if ((PLA_VECTOR == objtype_x) || (PLA_MVECTOR == objtype_x))
   {
      split_side = PLA_SIDE_TOP;
      myrank2 = -1; /* see later test used */
      PLA_Temp_comm_all_info(template, &communicator, &myrank, &comm_count);
   }
   index_buffer = (void*)pla_malloc(comm_count * sizeof(int));

   if ( MPI_FLOAT == datatype_x )
   {
      *(float*) x_max_buffer = -0.0;
      value_buffer = (void*)pla_malloc(comm_count * sizeof(float));
   }
   else if (MPI_DOUBLE == datatype_x)
   {
      *(double*)x_max_buffer = -0.0;
      value_buffer = (void*)pla_malloc(comm_count * sizeof(double));
   }
   else if ( MPI_COMPLEX == datatype_x )
   {
      *(float*) x_max_buffer = -0.0;
      *((float*) x_max_buffer + 1) = -0.0;
      value_buffer = (void*)pla_malloc(comm_count * sizeof(PLA_COMPLEX));
   }
   else if (MPI_DOUBLE_COMPLEX == datatype_x)
   {
      *(double*)x_max_buffer = -0.0;
      *((double*)x_max_buffer + 1) = -0.0;
      value_buffer = (void*)pla_malloc(comm_count * sizeof(PLA_DOUBLE_COMPLEX));
   }
   while( TRUE )
   {
      PLA_Obj_split_size( x_cur, split_side, &size, &owner );
      if(size == 0) break;
      if(PLA_SIDE_TOP == split_side) /* I could just use a function ptr here */
         PLA_Obj_horz_split_2( x_cur, size,        &x_top, &x_rest );
      else
         PLA_Obj_vert_split_2( x_cur, size,        &x_top, &x_rest );
      if ( (owner == myrank2) || ((myrank == owner) && (myrank2 == -1)) )
      {
         if(PLA_SIDE_TOP == split_side)
         {
            PLA_Obj_local_length(x_top, &x_size);
            incx = 1; /* Column-major storage */
         }
         else
         {
            PLA_Obj_local_width (x_top, &x_size);
            PLA_Obj_local_ldim(x_top, &incx);
         }
         PLA_Obj_local_buffer( x_top, &buf_x );
         if(MPI_DOUBLE == datatype_x)
{
            location_temp = idamax_(&x_size, buf_x, &incx);
}
         else if (MPI_FLOAT == datatype_x)
{
            location_temp = isamax_(&x_size, buf_x, &incx);
}
         else if (MPI_COMPLEX == datatype_x)
{
            location_temp = icamax_(&x_size, buf_x, &incx);
}
         else if( MPI_DOUBLE_COMPLEX == datatype_x )
{
            location_temp = izamax_(&x_size, buf_x, &incx);
}
else printf("Illegal datatype\n");
/* HORRIBLE PATCH */
         if (location_temp != -1)  /* idamax returns -1 if nothing > 0 */
         {
            if(MPI_DOUBLE == datatype_x)
            {
               d_temp = *(((double*)buf_x)+location_temp-1);
               if(d_temp < 0.0) d_temp = d_temp * -1.0;
               if(d_temp > dabs(*(double*)x_max_buffer))
               {
                  *(double*)x_max_buffer = *(((double*)buf_x) + location_temp - 1);
                  location = location_temp + k_temp - 1;
               }
            }
            else if(MPI_FLOAT == datatype_x)
            {
               f_temp = *(((float*)buf_x)+location_temp-1);
               if(f_temp < 0.0) f_temp = f_temp * -1.0;
               if(f_temp > f_abs(*(float*)x_max_buffer))
               {
                  *(float*)x_max_buffer = *(((float*)buf_x) + location_temp - 1);
                  location = location_temp + k_temp - 1;
               }
            }
            if(MPI_DOUBLE_COMPLEX == datatype_x)
            {
               d_temp = *(((double*)buf_x)+location_temp-1);
               d_temp2 = *(((double*)buf_x)+location_temp);
               d_temp = d_temp * d_temp + d_temp2 * d_temp2;
               if(d_temp > zabs((double*)x_max_buffer))
               {
                 *(double*)x_max_buffer = *(((double*)buf_x)+location_temp-1);
                 *((double*)x_max_buffer+1)=*(((double*)buf_x)+location_temp);
                  location = location_temp + k_temp - 1;
               }
            }
            else if(MPI_COMPLEX == datatype_x)
            {
               f_temp = *(((float*)buf_x)+location_temp-1);
               f_temp2 = *(((float*)buf_x)+location_temp);
               f_temp = f_temp * f_temp + f_temp2 * f_temp2 ;
               if(f_temp > c_abs((float*)x_max_buffer))
               {
                  *(float*)x_max_buffer =*(((float*)buf_x) + location_temp - 1);
                  *((float*)x_max_buffer+1) =*(((float*)buf_x) + location_temp);
                  location = location_temp + k_temp - 1;
               }
            }
         }

         k_temp = k_temp + size;
      }
      else /* update global reference outside of self */
      {
         k_temp = k_temp + size;
      }
      PLA_Obj_view_swap(&x_cur, &x_rest);
      MPI_Barrier(MPI_COMM_WORLD);
   }
   MPI_Allgather(x_max_buffer, 1,           datatype_x,
                 value_buffer, 1,  datatype_x, communicator);
   MPI_Allgather((void*)(&location), 1,     MPI_INT,
                 index_buffer, 1,  MPI_INT,    communicator);

   if( (PLA_MATRIX == objtype_x) || (PLA_PMVECTOR == objtype_x) ||
       (PLA_PVECTOR == objtype_x))
   {
      if(PLA_PROJ_ONTO_ROW == row_or_column)
      {
/* Note -- important distinction between comm_count and comm_count2 */
         PLA_Temp_comm_col_info(template, &communicator, &idummy, &idummy);
         MPI_Bcast(value_buffer,comm_count,datatype_x,holder_rank,communicator);
         MPI_Bcast(index_buffer,comm_count,MPI_INT,   holder_rank,communicator);
      }
      else if(PLA_PROJ_ONTO_COL == row_or_column)
      {
         PLA_Temp_comm_row_info(template, &communicator, &idummy, &idummy);
         MPI_Bcast(value_buffer,comm_count,datatype_x,holder_rank,communicator);
         MPI_Bcast(index_buffer,comm_count,MPI_INT,   holder_rank,communicator);
      }
   }
/* Now everyone has a copy of values and indexes -- do local search for value */
   if ( MPI_FLOAT == datatype_x )
      *(float*) x_max_buffer = -0.0;
   else if (MPI_DOUBLE == datatype_x)
      *(double*)x_max_buffer = -0.0;
   if ( MPI_COMPLEX == datatype_x )
   {
      *(float*) x_max_buffer = -0.0;
      *((float*) x_max_buffer+1) = -0.0;
   }
   if ( MPI_DOUBLE_COMPLEX == datatype_x )
   {
      *(double*) x_max_buffer = -0.0;
      *((double*) x_max_buffer+1) = -0.0;
   }
   location = -1;
   for(i = 0; i < comm_count; i++) /* ARGH! indices */
   {
      if(MPI_DOUBLE == datatype_x)
      {
         d_temp = *(((double*)value_buffer)+i);
         if( dabs(d_temp) > dabs(*((double*)x_max_buffer)))
         {
            *((double*)x_max_buffer) = d_temp;
            *((int*)k_buffer) = *(((int*)index_buffer)+i);
         }
      }
      else if(MPI_FLOAT == datatype_x)
      {
         f_temp = *(((float*)value_buffer)+i);
         if( f_abs(f_temp) > f_abs(*((float*)x_max_buffer)))
         {
            *((float*)x_max_buffer) = abs(f_temp);
            *((int*)k_buffer) = *(((int*)index_buffer)+i);
         }
      }
      else if(MPI_DOUBLE_COMPLEX == datatype_x)
      {
         d_temp = *(((double*)value_buffer)+i*2);
         d_temp2 = *(((double*)value_buffer)+i*2+1);
         if( ( abs(d_temp)+abs(d_temp2) ) > 
                            ( abs(*((double*)x_max_buffer))+
                              abs(*((double*)x_max_buffer+1)) ) )
         {
            *((double*)x_max_buffer) = abs(d_temp);
            *((double*)x_max_buffer+1) = abs(d_temp2);
            *((int*)k_buffer) = *(((int*)index_buffer)+i);
         }
      }
      else if(MPI_COMPLEX == datatype_x)
      {
         f_temp = *(((float*)value_buffer)+i*2);
         f_temp2 = *(((float*)value_buffer)+i*2+1);
         if( ( abs(f_temp)+abs(f_temp2) ) > 
                            ( abs(*((float*)x_max_buffer))+
                              abs(*((float*)x_max_buffer+1)) ) )
         {
            *((float*)x_max_buffer) = abs(f_temp);
            *((float*)x_max_buffer+1) = abs(f_temp2);
            *((int*)k_buffer) = *(((int*)index_buffer)+i);
         }
      }
   }
   if(value_buffer) pla_free(value_buffer);
   if(index_buffer) pla_free(index_buffer);
   PLA_Obj_free(&x_cur); PLA_Obj_free(&x_top); PLA_Obj_free(&x_leftover);
   PLA_Obj_free(&x_rest); PLA_Obj_free(&junk);
   MPI_Barrier(MPI_COMM_WORLD);
   return( 0 );
}


