/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/



#include "PLA.h"
#include <stdio.h>

#ifndef min3
#define min3(x, y, z)        min(min(x, y), z)
#endif

int  PLA_Local_trsv(int uplo, int trans, int diag,
                    PLA_Obj  a, PLA_Obj x)
{
   void    *a_buf, *x_buf;
   int     n, m;
   int     lda, incx;
   char    uplo_char[1], trans_char[1], diag_char[1];
   MPI_Datatype datatype;

   PLA_Obj_datatype( a, &datatype );

   PLA_Obj_local_buffer(a, &a_buf);
   PLA_Obj_local_buffer(x, &x_buf);

   PLA_Obj_local_length( a, &n );
   PLA_Obj_local_width( a, &m );
   PLA_Obj_local_ldim( a, &lda );
   PLA_Obj_local_stride( x, &incx );

   if(PLA_UPPER_TRIANGULAR == uplo)
   	uplo_char[0] = 'U';
   else uplo_char[0] = 'L';
   if(PLA_NO_TRANSPOSE == trans)
   	trans_char[0] = 'N';
   else trans_char[0] = 'T';
   if(PLA_UNIT_DIAG == diag)
   	diag_char[0] = 'U';
   else diag_char[0] = 'N';

   if( (n*m) > 0)
   {
       if( MPI_FLOAT == datatype )
          PLA_strsv( uplo_char, trans_char, diag_char, &n, a_buf, &lda, x_buf, &incx);
       else if ( MPI_DOUBLE == datatype )
          PLA_dtrsv( uplo_char, trans_char, diag_char, &n, a_buf, &lda, x_buf, &incx);
       else if ( MPI_COMPLEX == datatype )
          PLA_ctrsv( uplo_char, trans_char, diag_char, &n, a_buf, &lda, x_buf, &incx);
       else if ( MPI_DOUBLE_COMPLEX == datatype )
          PLA_ztrsv( uplo_char, trans_char, diag_char, &n, a_buf, &lda, x_buf, &incx);
   }
}
/*************************************************************************/
int PLA_Trsv_ln( int diag, PLA_Obj a, PLA_Obj x)
{
   PLA_Obj update_matrix = NULL, a_bottom_right = NULL, a_square = NULL;
   PLA_Obj x_top = NULL, x_bottom = NULL;
   PLA_Obj one = NULL, negative_one = NULL;
   int vector_length, a_left_size, a_top_size, size;
   int a_left_owner, a_top_owner;
   MPI_Datatype datatype;
   PLA_Template template;

   PLA_Obj_datatype ( a, &datatype );
   PLA_Obj_template ( a, &template );
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &negative_one);
   PLA_Obj_set_to_one      (          one );
   PLA_Obj_set_to_minus_one( negative_one );

   PLA_Obj_view_all ( a, &a_bottom_right );
   PLA_Obj_view_all( x, &x_bottom );
   for( ; ; )
   {

       PLA_Obj_global_length( x_bottom, &vector_length );
       if ( 0 == vector_length ) break;
       PLA_Obj_split_size( a_bottom_right, PLA_SIDE_LEFT, &a_left_size, &a_left_owner );
       PLA_Obj_split_size( a_bottom_right, PLA_SIDE_TOP,  &a_top_size , &a_top_owner  );
       size = min ( a_left_size, a_top_size );
       PLA_Obj_split_4 ( a_bottom_right, size, size, &a_square,      PLA_DUMMY,
                                                     &update_matrix, &a_bottom_right);
       PLA_Obj_horz_split_2( x_bottom, size, &x_top, &x_bottom );

       PLA_Local_trsv( PLA_LOW_TRIAN, PLA_NO_TRANS, diag, a_square, x_top );
       PLA_Gemv( PLA_NO_TRANS, negative_one, update_matrix, x_top, one, x_bottom);
   }
   PLA_Obj_free( &one   ); PLA_Obj_free( &negative_one );
   PLA_Obj_free( &update_matrix ); PLA_Obj_free( &a_bottom_right ); PLA_Obj_free( &a_square );
}
/******************************************************************************/
int PLA_Trsv_lt( int diag, PLA_Obj a, PLA_Obj x)
{
   PLA_Obj update_matrix = NULL, a_top_left = NULL, a_square = NULL;
   PLA_Obj x_top = NULL, x_bottom = NULL;
   PLA_Obj one = NULL, negative_one = NULL;
   int vector_length, a_right_size, a_bottom_size, size;
   int a_right_owner, a_bottom_owner;
   MPI_Datatype datatype;
   PLA_Template template;

   PLA_Obj_datatype ( a, &datatype );
   PLA_Obj_template ( a, &template );
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &negative_one);
   PLA_Obj_set_to_one      (          one );
   PLA_Obj_set_to_minus_one( negative_one );

   PLA_Obj_view_all ( a, &a_top_left );
   PLA_Obj_view_all ( x, &x_top );
   for( ; ; )
   {
       PLA_Obj_global_length( x_top, &vector_length );
       if ( 0 == vector_length ) break;
       PLA_Obj_split_size( a_top_left, PLA_SIDE_RIGHT,  &a_right_size,  &a_right_owner );
       PLA_Obj_split_size( a_top_left, PLA_SIDE_BOTTOM, &a_bottom_size, &a_bottom_owner  );
       size = min ( a_right_size, a_bottom_size );
       PLA_Obj_split_4 ( a_top_left, -size, -size, &a_top_left,      PLA_DUMMY,
                                                   &update_matrix, &a_square);
       PLA_Obj_horz_split_2( x_top, -size, &x_top, &x_bottom );
       PLA_Local_trsv( PLA_LOW_TRIAN, PLA_TRANS, diag, a_square, x_bottom );
       PLA_Gemv( PLA_TRANS, negative_one, update_matrix, x_bottom, one, x_top);
   }
   PLA_Obj_free( &x_top ); PLA_Obj_free( &x_bottom );
   PLA_Obj_free( &one   ); PLA_Obj_free( &negative_one );
   PLA_Obj_free( &update_matrix ); PLA_Obj_free( &a_top_left ); PLA_Obj_free( &a_square );
}
/******************************************************************************/

/******************************************************************************/
int PLA_Trsv_un( int diag, PLA_Obj a, PLA_Obj x)
{
   PLA_Obj update_matrix = NULL, a_top_left = NULL, a_square = NULL;
   PLA_Obj x_top = NULL, x_bottom = NULL;
   PLA_Obj one = NULL, negative_one = NULL;
   int vector_length, a_right_size, a_bottom_size, size;
   int a_right_owner, a_bottom_owner;
   MPI_Datatype datatype;
   PLA_Template template;

   PLA_Obj_datatype ( a, &datatype );
   PLA_Obj_template ( a, &template );
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &negative_one);
   PLA_Obj_set_to_one      (          one );
   PLA_Obj_set_to_minus_one( negative_one );

   PLA_Obj_view_all ( a, &a_top_left );
   PLA_Obj_view_all ( x, &x_top );
   for( ; ; )
   {
       PLA_Obj_global_length( x_top, &vector_length );
       if ( 0 == vector_length ) break;
       PLA_Obj_split_size( a_top_left, PLA_SIDE_RIGHT,  &a_right_size,  &a_right_owner );
       PLA_Obj_split_size( a_top_left, PLA_SIDE_BOTTOM, &a_bottom_size, &a_bottom_owner  );
       size = min ( a_right_size, a_bottom_size );
       PLA_Obj_split_4 ( a_top_left, -size, -size, &a_top_left,    &update_matrix,
                                                   PLA_DUMMY,      &a_square);
       PLA_Obj_horz_split_2( x_top, -size, &x_top, &x_bottom );
       PLA_Local_trsv( PLA_UP_TRIAN, PLA_NO_TRANS, diag, a_square, x_bottom );
       PLA_Gemv( PLA_NO_TRANS, negative_one, update_matrix, x_bottom, one, x_top);
   }
   PLA_Obj_free( &x_top ); PLA_Obj_free( &x_bottom );
   PLA_Obj_free( &one   ); PLA_Obj_free( &negative_one );
   PLA_Obj_free( &update_matrix ); PLA_Obj_free( &a_top_left ); PLA_Obj_free( &a_square );
}
/******************************************************************************/
int PLA_Trsv_ut( int diag, PLA_Obj a, PLA_Obj x)
{
   PLA_Obj update_matrix = NULL, a_bottom_right = NULL, a_square = NULL;
   PLA_Obj x_top = NULL, x_bottom = NULL;
   PLA_Obj one = NULL, negative_one = NULL;
   int vector_length, a_left_size, a_top_size, size;
   int a_left_owner, a_top_owner;
   MPI_Datatype datatype;
   PLA_Template template;

   PLA_Obj_datatype ( a, &datatype );
   PLA_Obj_template ( a, &template );
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &one);
   PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &negative_one);
   PLA_Obj_set_to_one      (          one );
   PLA_Obj_set_to_minus_one( negative_one );

   PLA_Obj_view_all ( a, &a_bottom_right );
   PLA_Obj_view_all ( x, &x_bottom );
   for( ; ; )
   {
       PLA_Obj_global_length( x_bottom, &vector_length );
       if ( 0 == vector_length ) break;
       PLA_Obj_split_size( a_bottom_right, PLA_SIDE_LEFT, &a_left_size, &a_left_owner );
       PLA_Obj_split_size( a_bottom_right, PLA_SIDE_TOP,  &a_top_size , &a_top_owner  );
       size = min ( a_left_size, a_top_size );
       PLA_Obj_split_4 ( a_bottom_right, size, size, &a_square,      &update_matrix,
                                                     PLA_DUMMY,      &a_bottom_right);
       PLA_Obj_horz_split_2( x_bottom, size, &x_top, &x_bottom );
       PLA_Local_trsv( PLA_UP_TRIAN, PLA_TRANS, diag, a_square, x_top );
       PLA_Gemv( PLA_TRANS, negative_one, update_matrix, x_top, one, x_bottom);
   }
   PLA_Obj_free( &x_top ); PLA_Obj_free( &x_bottom );
   PLA_Obj_free( &one   ); PLA_Obj_free( &negative_one );
   PLA_Obj_free( &update_matrix ); PLA_Obj_free( &a_bottom_right ); 
   PLA_Obj_free( &a_square );
}
/******************************************************************************/
int PLA_Trsv( int uplo, int trans, int diag, PLA_Obj a, PLA_Obj x)
{
/* It seems that the way to make this fast is to assume that x is aligned */
/* with A.  Since PLA_Copy does not handle re-alignment right now I am    */
/*   PLA_Trsv_enter ( uplo, trans, diag, a, x ); */
   if(PLA_UPPER_TRIANGULAR == uplo)
   {
      if(PLA_NO_TRANSPOSE == trans)
      {
         PLA_Trsv_un( diag, a, x );
      }
      else /* PLA_TRANSPOSE */
      {
         PLA_Trsv_ut( diag, a, x );
      }
   }
   else /* PLA_LOWER_TRIANGULAR */
   {
      if(PLA_NO_TRANSPOSE == trans)
      {
         PLA_Trsv_ln( diag, a, x );
      }
      else /* PLA_TRANSPOSE */
      {
         PLA_Trsv_lt( diag, a, x );
      }
   }
/*   PLA_Trsv_exit  ( uplo, trans, diag, a, x ); */

}
