#include "PLA.h"

static PLA_Obj alpha_in = NULL, a_in = NULL, b_in = NULL;

int PLA_Trsm_enter ( int side, int uplo, int transa, int diag,
                     PLA_Obj alpha, PLA_Obj a, PLA_Obj b )
{
   int check_parameter, instrument, check_seq, myrow, mycol;
   char current_routine[20] = "PLA_Trsm";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check transa and transb */
      if ( PLA_NO_TRANS != transa && PLA_TRANS != transa && PLA_CONJ_TRANS != transa )
         PLA_Log_abort("PLA_Trsm_enter", "PLA_Trsm : transa invalid", PLA_ERROR_CODE, original_error);

	  if ( (PLA_UPPER_TRIANGULAR != uplo) && (PLA_LOWER_TRIANGULAR != uplo))
         PLA_Log_abort("PLA_Trsm_enter", "PLA_Trsm : uplo invalid", PLA_ERROR_CODE, original_error);

	  if ( (PLA_SIDE_LEFT != side) && (PLA_SIDE_RIGHT != side))
         PLA_Log_abort("PLA_Trsm_enter", "PLA_Trsm : side invalid", PLA_ERROR_CODE, original_error);

      if ( PLA_NO_TRANS != transa && PLA_TRANS != transa && PLA_CONJ_TRANS != transa )
         PLA_Log_abort("PLA_Trsm_enter", "PLA_Trsm : transa invalid", PLA_ERROR_CODE, original_error);

	  if ( (PLA_UNIT_DIAG != diag) && (PLA_NONUNIT_DIAG != diag))
         PLA_Log_abort("PLA_Trsm_enter", "PLA_Trsm : diag invalid", PLA_ERROR_CODE, original_error);

	  /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( a, "a", error, original_error);
      PLA_Obj_object_valid( b, "b", error, original_error);

    /* check object types */
      PLA_Obj_objtype_equal( alpha, PLA_MSCALAR, "alpha", error, original_error );
      PLA_Obj_objtype_equal( a,     PLA_MATRIX,  "a", error, original_error );
      PLA_Obj_objtype_equal( b,     PLA_MATRIX,  "b", error, original_error );

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( a,     datatype, "a", error, original_error );
      PLA_Obj_datatype_equal( b,     datatype, "b", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( a,     template, "a", error, original_error );
      PLA_Obj_template_equal( b,     template, "b", error, original_error );


    /* check dimensions */
      PLA_Obj_global_length_equal ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_global_width_equal  ( alpha, 1, "alpha", error, original_error );

	  PLA_Obj_global_length_width_equal(a, a, "a and a (sq.)", error, original_error);

      if (  (( PLA_SIDE_LEFT  == side ) && (PLA_NO_TRANSPOSE == transa))           ||
		    (( PLA_SIDE_RIGHT == side ) && ((PLA_TRANSPOSE    == transa) ||  
			                                (PLA_CONJUGATE_TRANSPOSE == transa))) ) 
               PLA_Obj_global_width_length_equal(a, b, "a and b", error, original_error );
      else
		       PLA_Obj_global_width_width_equal (a, b, "a and b", error, original_error );
   }


   PLA_Get_check_sequential ( &check_seq );
   if ( PLA_CHECK_SEQUENTIAL == check_seq )
   {
       PLA_Mscalar_create_conf_to( alpha, 0, 0, &alpha_in );
       PLA_Mscalar_create_conf_to( a,     0, 0, &a_in );
       PLA_Mscalar_create_conf_to( b,     0, 0, &b_in );

       PLA_Copy_temp( alpha, alpha_in );
       PLA_Copy_temp( a,     a_in );
       PLA_Copy_temp( b,     b_in );

       PLA_Temp_comm_col_rank( template, &myrow );
       PLA_Temp_comm_row_rank( template, &mycol );

       if ( 0 == myrow && 0 == mycol )
          PLA_Local_trsm( side, uplo, transa, diag, alpha_in, a_in, b_in );
   }

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}


int PLA_Trsm_exit ( int side, int uplo, int transa, int diag,
                    PLA_Obj alpha, PLA_Obj a, PLA_Obj b )
{
  char current_routine[20];
  int  instrument, check_seq, myrow, mycol;
  PLA_Obj b_out = NULL;
  PLA_Template template;

  PLA_Get_check_sequential ( &check_seq );
/* MAY (JAG) want to test the objects for validity here as well      */
/* in case we have a wayward pointer or some other memory corruption */
  if ( PLA_CHECK_SEQUENTIAL == check_seq )
  {
     PLA_Mscalar_create_conf_to( b,     0, 0, &b_out );
     PLA_Copy_temp(b, b_out);
     PLA_Obj_template(b, &template);
     PLA_Temp_comm_col_rank( template, &myrow );
     PLA_Temp_comm_row_rank( template, &mycol );
     if ( 0 == myrow && 0 == mycol )
     {
/* Commented out because I know we have the code to do this somewhere */
/* and I would like to avoid re-writing it.  It's not a big deal to   */
/* rewrite if we restrict what can be passed to the function though.  */
/*        diff = PLA_max_abs_diff( b_in, b_out ); */
      }
   }

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   PLA_Obj_free( &alpha_in );
   PLA_Obj_free( &a_in );
   PLA_Obj_free( &b_in );
   PLA_Obj_free( &b_out );
   return(PLA_SUCCESS); /* Actually, return fail if exceed tolerance in test */
}

int PLA_Local_trsm_enter ( int side, int uplo, int transa, int diag,
                           PLA_Obj alpha, PLA_Obj a, PLA_Obj b )
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_trsm";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check transa and transb */
      if ( PLA_NO_TRANS != transa && PLA_TRANS != transa && PLA_CONJ_TRANS != transa )
         PLA_Log_abort("PLA_Local_trsm_enter", "PLA_Local_trsm : transa invalid", PLA_ERROR_CODE, original_error);

	  if ( (PLA_UPPER_TRIANGULAR != uplo) && (PLA_LOWER_TRIANGULAR != uplo))
         PLA_Log_abort("PLA_Local_trsm_enter", "PLA_Local_trsm : uplo invalid", PLA_ERROR_CODE, original_error);

	  if ( (PLA_SIDE_LEFT != side) && (PLA_SIDE_RIGHT != side))
         PLA_Log_abort("PLA_Local_trsm_enter", "PLA_Local_trsm : side invalid", PLA_ERROR_CODE, original_error);

      if ( PLA_NO_TRANS != transa && PLA_TRANS != transa && PLA_CONJ_TRANS != transa )
         PLA_Log_abort("PLA_Local_trsm_enter", "PLA_Local_trsm : transa invalid", PLA_ERROR_CODE, original_error);

	  if ( (PLA_UNIT_DIAG != diag) && (PLA_NONUNIT_DIAG != diag))
         PLA_Log_abort("PLA_Local_trsm_enter", "PLA_Local_trsm : diag invalid", PLA_ERROR_CODE, original_error);

	  /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( a, "a", error, original_error);
      PLA_Obj_object_valid( b, "b", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( a,     datatype, "a", error, original_error );
      PLA_Obj_datatype_equal( b,     datatype, "b", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( a,     template, "a", error, original_error );
      PLA_Obj_template_equal( b,     template, "b", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal  ( alpha, 1, "alpha", error, original_error );

	  PLA_Obj_local_length_width_equal(a, a, "a and a (sq.)", error, original_error);

      if (  (( PLA_SIDE_LEFT  == side ) && (PLA_NO_TRANSPOSE == transa))           ||
		    (( PLA_SIDE_RIGHT == side ) && ((PLA_TRANSPOSE    == transa) ||  
			                                (PLA_CONJUGATE_TRANSPOSE == transa))) ) 
               PLA_Obj_local_width_length_equal(a, b, "a and b", error, original_error );
      else
		       PLA_Obj_local_width_width_equal (a, b, "a and b", error, original_error );
   }


   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}


int PLA_Local_trsm_exit ( int side, int uplo, int transa, int diag,
                    PLA_Obj alpha, PLA_Obj a, PLA_Obj b )
{
  char current_routine[20];
  int  instrument;

  PLA_Get_instrument ( &instrument );
  if ( PLA_INSTRUMENT == instrument )
  {
    /* insert instrumentation calls */
  }
  PLA_Stack_pop_current( current_routine );

  return(PLA_SUCCESS); /* Actually, return fail if exceed tolerance in test */
}
