
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/





#include "PLA.h"
/* 
PLA_Reduce --> reduces the panel duplicated on row/column
of processors to whatnot.

by Philip Alpatov 
*/
static int FIRST_WARNING = 1;

static double PLA_DUMMY_ADRESS;

#if MANUFACTURE == CRAY
  #define BF(buffer) (NULL==buffer) ? &PLA_DUMMY_ADRESS : buffer
#else
  #define BF(buffer) buffer 
#endif

int dupl_colpanel_to_vector(PLA_Obj p, PLA_Obj v, MPI_Op op)
     /*
	assumes alignments sizez and datatypes match
	appropriately  
	Note that in this case of a vector we don't have to worry
	about unpacking recieve buffer. 
	*/


{ 
  int
    error = 0,
    idummy,
    i,j,k,l,m,
    glob_p_length,
    glob_p_width,
    p_align_row;

  
  PLA_Template
    template;

  MPI_Datatype
    datatype;

  PLA_Obj
    ptemp = NULL;

  void
    *p_buf = NULL;

/* get local and global info from objects */  
  PLA_Panel_local_all(p, &datatype, &idummy, &idummy,
		     &p_buf, &idummy);
  PLA_Panel_global_all(p, (PLA_Panel_rowcol_flag*)&idummy, &glob_p_length, &glob_p_width,
		      &p_align_row, &idummy,
		      &template);
  

  error += PLA_Panel_create( datatype,
			    PLA_PROJ_ONTO_COL, 
			    PLA_ALL_COLS,
			    glob_p_length,
			    glob_p_width,
			    template,
			    p_align_row,
			    &ptemp);
  

  error += dupl_colpanel_to_dupl_colpanel(p, ptemp, op);
  
  error += PLA_Copy(ptemp, v);

  error += PLA_Obj_free(&ptemp);

  if (error != 0)
    printf("colpanel_to_vector in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);
}


int dupl_rowpanel_to_vector(PLA_Obj p, PLA_Obj v, MPI_Op op)
     /*
	assumes alignments sizes and datatypes match
	appropriately  
	Note that in this case of a vector we don't have to worry
	about unpacking recieve buffer. 
	*/
 
{
  int
    error = 0,
    idummy,
    i,j,k,l,m,
    glob_p_length,
    glob_p_width,
    p_align_col;

  
  PLA_Template
    template;

  MPI_Datatype
    datatype;

  PLA_Obj
    ptemp = NULL;

  void
    *p_buf = NULL;

/* get local and global info from objects */  
  PLA_Panel_local_all(p, &datatype, &idummy, &idummy,
		     &p_buf, &idummy);
  PLA_Panel_global_all(p, (PLA_Panel_rowcol_flag*)&idummy, &glob_p_length, 
		      &glob_p_width,
		      &idummy, &p_align_col,
		      &template);
  

  error += PLA_Panel_create( datatype,
			    PLA_PROJ_ONTO_ROW, 
			    PLA_ALL_ROWS,
			    glob_p_length,
			    glob_p_width,
			    template,
			    p_align_col,
			    &ptemp);

  error += dupl_rowpanel_to_dupl_rowpanel(p, ptemp, op);
  
  error += PLA_Copy(ptemp, v);

  error += PLA_Obj_free(&ptemp);

  if (error != 0)
    printf("rowpanel_to_vector in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);
}

/********************************************************************/







/********************************************************************/

int dupl_colpanel_to_dupl_colpanel(PLA_Obj p1, PLA_Obj p2, MPI_Op op)
/* 
   reduces panel duplicated on columns of processors to a 
   similarly duplicated panel 

   This routine will check that the alignments and sizes of
both panels match
*/

{ 
  int
    i,j, l,
    error = 0,
    count = 0,
    idummy,
    loc_p1_length, loc_p2_length, loc_p_length,
    loc_p1_width, loc_p2_width, loc_p_width,
    p1_align_row, p2_align_row, p_align_row,
    p1_align_col, p2_align_col, p_align_col,
    p1_lld, p2_lld;

  
  PLA_Duplicate_flag
    duplicate;

  MPI_Datatype
    datatype_1, datatype_2, datatype;

  MPI_Comm
    row_comm;

  PLA_Template
    template_1, template_2, template;

  void
    *recvbuffer_ptr = NULL,
    *sendbuffer_ptr = NULL,
    *p1_buf = NULL,
    *p2_buf = NULL;

  PLA_Obj
    sendbuffer = NULL,
    recvbuffer = NULL;



/* get local and global info from objects */  
  error += PLA_Panel_local_all(p1, &datatype_1, &loc_p1_length, 
			      &loc_p1_width, &p1_buf, &p1_lld);
  error += PLA_Panel_global_all(p1, (PLA_Panel_rowcol_flag*)&idummy, &idummy, &idummy,
		      &p1_align_row, &p1_align_col,
		      &template_1);
  error += PLA_Panel_local_all(p2, &datatype_2, &loc_p2_length, 
			      &loc_p2_width, &p2_buf, &p2_lld);
  error += PLA_Panel_global_all(p2, (PLA_Panel_rowcol_flag*)&idummy, &idummy, &idummy,
		      &p2_align_row, &p2_align_col,&template_2);

/* get communicators and mesh info */

  error += PLA_Temp_comm_row(template_1, &row_comm);

/*_______________________________________________________________*/


/* do the size , template, alignment and datatype checks */
  if (datatype_1 == datatype_2)
    datatype = datatype_1;
  else
    {
      printf("panel datatypes in PLA_Reduce do not match; exiting\n");
      return ++error;
    }

  if (template_1 == template_2)
    template = template_1;
  else
    {
      printf("panel templates in PLA_Reduce do not match; exiting\n");
      return ++error;
    }

  if (loc_p1_width == loc_p2_width &&
      loc_p1_length == loc_p2_length &&
      p1_align_row == p2_align_row)
    {
      loc_p_length = loc_p1_length;
      loc_p_width = loc_p1_width;
      p_align_row = p1_align_row;
    }
  else
    {
      printf("panel size or panel alignment do not match; exiting \n");
      return ++error;
    }

  
/*_______________________________________________________________*/


/* how much data is communicated */
  count = loc_p_length * loc_p_width;

/* create the sendbuffer and the recvbuffer*/
  PLA_Mscalar_create(datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    loc_p_length,
		    loc_p_width,
		    template,
		    &sendbuffer);
  PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

  PLA_Mscalar_create(datatype, 
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    loc_p_length,
		    loc_p_width,
		    template,
		    &recvbuffer);
  PLA_Obj_local_buffer( recvbuffer, &recvbuffer_ptr);


  if (count == 0)
    return error;


  /* pack the data from p1 into the sendbuffer */

  error += PLA_Obj_extract_contents( p1,
				     loc_p_length,
				     1,
				     &idummy,
				     &idummy,
				     sendbuffer_ptr);
  
      
  /* perform the reduce to all operation */
  error += MPI_Allreduce(BF(sendbuffer_ptr), BF(recvbuffer_ptr), 
			 count, datatype, op, row_comm);
      
  error += PLA_Obj_extract_contents( recvbuffer,
				     p2_lld,
				     1,
				     &idummy,
				     &idummy,
				     p2_buf);
				 

  error += PLA_Obj_free(&sendbuffer);
  error += PLA_Obj_free(&recvbuffer);
  
  if(error != 0)
    printf("colpanel_to_colpanel in PLA_Reduce completed with error = %d\n", error);

  
  return (error); 
}
/********************************************************************/








/********************************************************************/

int dupl_rowpanel_to_dupl_rowpanel(PLA_Obj p1, PLA_Obj p2, MPI_Op op)
/* 
   reduces panel duplicated on rows of processors to a 
   similarly duplicated panel 

   This routine will check that the alignments and sizes of
both panels match
*/
{
  int
    i,j, l,
    error = 0,
    count = 0,
    idummy,
    loc_p1_length, loc_p2_length, loc_p_length,
    loc_p1_width, loc_p2_width, loc_p_width,
    p1_align_row, p2_align_row, p_align_row,
    p1_align_col, p2_align_col, p_align_col,
    p1_lld, p2_lld;

  
  PLA_Duplicate_flag
    duplicate;

  MPI_Datatype
    datatype_1, datatype_2, datatype;

  MPI_Comm
    col_comm;

  PLA_Template
    template_1, template_2, template;

  void
    *recvbuffer_ptr = NULL,
    *sendbuffer_ptr = NULL,
    *p1_buf = NULL,
    *p2_buf = NULL;

  PLA_Obj
    sendbuffer = NULL,
    recvbuffer = NULL;




/* get local and global info from objects */  
  error += PLA_Panel_local_all(p1, &datatype_1, &loc_p1_length, 
			      &loc_p1_width, &p1_buf, &p1_lld);
  error += PLA_Panel_global_all(p1, (PLA_Panel_rowcol_flag*)&idummy, &idummy, &idummy,
		      &p1_align_row, &p1_align_col,
		      &template_1);
  error += PLA_Panel_local_all(p2, &datatype_2, &loc_p2_length, 
			      &loc_p2_width, &p2_buf, &p2_lld);
  error += PLA_Panel_global_all(p2, (PLA_Panel_rowcol_flag*)&idummy, &idummy, &idummy,
		      &p2_align_row, &p2_align_col,&template_2);

/* get communicators and mesh info */

  error += PLA_Temp_comm_col(template_1, &col_comm);

/*_______________________________________________________________*/


/* do the size , template, alignment and datatype checks */
  if (datatype_1 == datatype_2)
    datatype = datatype_1;
  else
    {
      printf("panel datatypes in PLA_Reduce do not match; exiting\n");
      return ++error;
    }

  if (template_1 == template_2)
    template = template_1;
  else
    {
      printf("panel templates in PLA_Reduce do not match; exiting\n");
      return ++error;
    }

  if (loc_p1_width == loc_p2_width &&
      loc_p1_length == loc_p2_length &&
      p1_align_col == p2_align_col)
    {
      loc_p_length = loc_p1_length;
      loc_p_width = loc_p1_width;
      p_align_col = p1_align_col;
    }
  else
    {
      printf("panel size or panel alignment do not match; exiting \n");
      return ++error;
    }

  
/*_______________________________________________________________*/


/* how much data is communicated */
  count = loc_p_length * loc_p_width;


  if (count == 0)
    return error;



/* create the sendbuffer and the recvbuffer*/
  PLA_Mscalar_create(datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    loc_p_width,
		    loc_p_length,
		    template,
		    &sendbuffer);
  PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

  PLA_Mscalar_create(datatype, 
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		    loc_p_width,
		    loc_p_length,
		    template,
		    &recvbuffer);
  PLA_Obj_local_buffer( recvbuffer, &recvbuffer_ptr);




  /* pack the data from p1 into the sendbuffer */

  error += PLA_Obj_extract_transpose_contents( p1,
				     loc_p_width,
				     1,
				     &idummy,
				     &idummy,
				     sendbuffer_ptr);
  
  /* perform the reduce to all operation */

  error += MPI_Allreduce(BF(sendbuffer_ptr), BF(recvbuffer_ptr), 
			 count, datatype, op, col_comm);

  error += PLA_Obj_extract_transpose_contents( recvbuffer,
				     p2_lld,
				     1,
				     &idummy,
				     &idummy,
				     p2_buf);
				 


  error += PLA_Obj_free(&sendbuffer);
  error += PLA_Obj_free(&recvbuffer);

  
  if(error != 0)
    printf("rowpanel_to_rowpanel in PLA_Reduce completed with error = %d\n", error);

  return (error); 

}
/********************************************************************/



int PLA_Pack_dupl_colpanel_vector( PLA_Obj panel, PLA_Obj vector, 
				 PLA_Obj sendbuffer, int *recvcounts)
/* 
   packs the column panel (width = 1) into a sendbuffer to be used
   by MPI_Reduce_Scatter. Also returns the recvcounts array of   
*/  
{
  int
    i,
    error = 0,
    me,
    idummy,
    z_or_1,
    size,
    np,
    nb,
    column, row,
    myrow, mycol,
    nprows, npcols,
    p,
    loc_p_width, loc_p_length,
    glob_v_length, v_align_row,
    current_vector_proc, current_vector_row,
    current_proc_col, current_proc_row;


  int 
    *displs = NULL;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    current_block = NULL,
    current_panel = NULL,
    remainder = NULL,
    current_sendbuffer = NULL;

  void
    *current_block_ptr = NULL,
    *p_buf = NULL,
    *current_panel_ptr = NULL,
    *current_sendbuffer_ptr = NULL;
  


  PLA_Panel_local_all( panel, &datatype, &loc_p_length, &loc_p_width,
		      &p_buf, &idummy);


  PLA_Vector_global_all( vector, &glob_v_length,
		        &v_align_row, &template);

  PLA_Obj_template( panel, &template);
  
  PLA_Temp_vector_owner_rank( template, v_align_row, &current_vector_proc);
  PLA_Temp_comm_col_size( template, &nprows);
  PLA_Temp_comm_row_size( template, &npcols);
  PLA_Temp_comm_col_rank( template, &myrow);
  PLA_Temp_comm_row_rank( template, &mycol);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_zero_or_one( template, (PLA_Start_num *) &z_or_1);


  np = nprows * npcols;

  /* first fill the recvcounts array */ 
  for(p = 0; p < np; p++)
    {
      column = p/nprows;
      row = p%nprows;
      if (row == myrow)
	PLA_Temp_vector_distant_length(template,
				      p,
				      glob_v_length,
				      v_align_row,
				      &recvcounts[column]);
    }

  /* might as well figure out the displacements in the sendbuffer : */
  displs = (int*)pla_malloc(npcols*sizeof(int));
  displs[0] = z_or_1;
  for(column = 1; column < npcols; column++)
    displs[column] = displs[column-1] + recvcounts[column-1];


  /*take a view into the panel */
  PLA_Panel_subpanel( panel, PLA_DIM_ALL, PLA_DIM_ALL,
		     PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		     &current_panel);

  for(;;){
    PLA_Obj_top(current_panel, &size, &current_proc_row);
    if( size == 0) break;
    
    PLA_Obj_horz_split_2( current_panel, size, &current_block, &remainder);

    current_proc_col = current_vector_proc/nprows;

    PLA_Mscalar_submscalar( sendbuffer,
			  size,
			  loc_p_width,
			  displs[current_proc_col],
			  PLA_ALIGN_FIRST,
			  &current_sendbuffer);
    PLA_Obj_local_buffer( current_sendbuffer, &current_sendbuffer_ptr);
 

    if (myrow == current_proc_row)
      {

	PLA_Obj_extract_contents( current_block,
				  size,
				  1,
				  &idummy, &idummy,
				  current_sendbuffer_ptr);
	displs[current_proc_col] += size;
      }
    
    current_vector_proc = (current_vector_proc + 1)%np;    
    PLA_Obj_swap( &current_panel, &remainder);
  }



  PLA_Obj_free( &current_sendbuffer);
  PLA_Obj_free( &remainder);
  PLA_Obj_free( &current_block);
  PLA_Obj_free( &current_panel);

  
  pla_free(displs);
  if (error != 0)
    printf("PLA_Pack_colpanel_to_vector in PLA_Reduce completed with \
error = %d\n", error);

  return (error);
  
}


int direct_dupl_colpanel_to_vector(PLA_Obj panel, PLA_Obj vector, MPI_Op op)
/* reduction of the duplicated column panel to a vector */
{
  int    
    error = 0,
    *recvcounts,
    npcols,
    loc_p_length, loc_p_width,
    idummy;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    sendbuffer = NULL;

  MPI_Comm
    row_comm;
  
  void
    *sendbuffer_ptr = NULL,
    *p_buf = NULL,
    *v_ptr = NULL;

  PLA_Panel_local_all(panel, &datatype, &loc_p_length, &loc_p_width,
		     &p_buf, &idummy);
  PLA_Obj_template( panel, &template);

  PLA_Obj_local_buffer(vector, &v_ptr);
  
  PLA_Temp_comm_row(template, &row_comm);

  PLA_Mscalar_create( datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		     loc_p_length,
		     loc_p_width,
		     template,
		     &sendbuffer);
  PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

  PLA_Temp_comm_row_size(template, &npcols);
  recvcounts = (int *)pla_malloc(npcols*sizeof(int));
 

  error += PLA_Pack_dupl_colpanel_vector( panel, vector, sendbuffer, 
					recvcounts );

  if (loc_p_length*loc_p_width != 0)
    error += MPI_Reduce_scatter( BF(sendbuffer_ptr), BF(v_ptr), recvcounts, datatype,
				op, row_comm);

  if (error != 0)
    printf("disrect_dupl_colpanel_to_vector in PLA_Reduce completed with error = %d\n",
	   error);

  pla_free( recvcounts);
  PLA_Obj_free( &sendbuffer);

  return (error);
}



int PLA_Pack_dupl_rowpanel_vector( PLA_Obj panel, PLA_Obj vector, 
			    PLA_Obj sendbuffer, int *recvcounts)
/* 
   packs the row duplicated panel (width = 1) into a sendbuffer to be used
   by MPI_Reduce_Scatter. Also returns the recvcounts array  
*/  
{
  int
    i,
    error = 0,
    me,
    idummy,
    z_or_1,
    size,
    np,
    nb,
    column, row,
    myrow, mycol,
    nprows, npcols,
    p,
    loc_p_width, loc_p_length,
    glob_v_length, v_align_row,
    current_vector_proc, current_vector_row,
    current_proc_col, current_proc_row;

  int 
    *displs = NULL;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    current_block = NULL,
    current_panel = NULL,
    remainder = NULL,
    current_sendbuffer = NULL;

  void
    *current_block_ptr = NULL,
    *p_buf = NULL,
    *current_panel_ptr = NULL,
    *current_sendbuffer_ptr = NULL;
  


  PLA_Panel_local_all( panel, &datatype, &loc_p_length, &loc_p_width,
		      &p_buf, &idummy);


  PLA_Vector_global_all( vector, &glob_v_length,
		        &v_align_row, &template);

  PLA_Obj_template( panel, &template);
  
  PLA_Temp_vector_owner_rank( template, v_align_row, &current_vector_proc);
  PLA_Temp_comm_col_size( template, &nprows);
  PLA_Temp_comm_row_size( template, &npcols);
  PLA_Temp_comm_col_rank( template, &myrow);
  PLA_Temp_comm_row_rank( template, &mycol);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_zero_or_one( template, (PLA_Start_num *) &z_or_1);


  np = nprows * npcols;

  /* first fill the recvcounts array */ 
  for(p = 0; p < np; p++)
    {
      column = p/nprows;
      row = p%nprows;
      if (column == mycol)
	PLA_Temp_vector_distant_length(template,
				      p,
				      glob_v_length,
				      v_align_row,
				      &recvcounts[row]);
    }

  /* might as well figure out the displacements in the sendbuffer : */
  displs = (int*)pla_malloc(nprows*sizeof(int));
  displs[0] = z_or_1;
  for(row = 1; row < nprows; row++)
    displs[row] = displs[row-1] + recvcounts[row-1];


  /*take a view into the panel */
  PLA_Panel_subpanel( panel, PLA_DIM_ALL, PLA_DIM_ALL,
		     PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		     &current_panel);

  for(;;){
    PLA_Obj_left(current_panel, &size, &current_proc_col);
    if( size == 0) break;
    
    PLA_Obj_vert_split_2( current_panel, size, &current_block, &remainder);

    current_proc_row = current_vector_proc%nprows;

    PLA_Mscalar_submscalar( sendbuffer,
			  loc_p_length,
			  size,
			  PLA_ALIGN_FIRST,
			  displs[current_proc_row],
			  &current_sendbuffer);
    PLA_Obj_local_buffer( current_sendbuffer, &current_sendbuffer_ptr);
    


    if (mycol == current_proc_col)
      {

	PLA_Obj_extract_contents( current_block,
				  loc_p_length,
				  1,
				  &idummy, &idummy,
				  current_sendbuffer_ptr);
	displs[current_proc_row] += size;
      }
    
    current_vector_proc = (current_vector_proc + 1)%np;    
    PLA_Obj_swap( &current_panel, &remainder);
  }



  PLA_Obj_free( &current_sendbuffer);
  PLA_Obj_free( &remainder);
  PLA_Obj_free( &current_block);
  PLA_Obj_free( &current_panel);

  
  pla_free(displs);
  if (error != 0)
    printf("PLA_Pack_colpanel_to_vector in PLA_Reduce completed with \
error = %d\n", error);

  return (error);
  
}


int direct_dupl_rowpanel_to_vector(PLA_Obj panel, PLA_Obj vector, MPI_Op op)
/* reduction of the duplicated column panel to a vector */
{
  int    
    error = 0,
    *recvcounts,
    nprows,
    loc_p_length, loc_p_width,
    idummy;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    sendbuffer = NULL;

  MPI_Comm
    col_comm;
  
  void
    *sendbuffer_ptr = NULL,
    *p_buf = NULL,
    *v_ptr = NULL;

  PLA_Panel_local_all(panel, &datatype, &loc_p_length, &loc_p_width,
		     &p_buf, &idummy);
  PLA_Obj_template( panel, &template);

  PLA_Obj_local_buffer(vector, &v_ptr);
  
  PLA_Temp_comm_col(template, &col_comm);

  PLA_Mscalar_create( datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
		     loc_p_length,
		     loc_p_width,
		     template,
		     &sendbuffer);
  PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

  PLA_Temp_comm_col_size(template, &nprows);
  recvcounts = (int *)pla_malloc(nprows*sizeof(int));


  error += PLA_Pack_dupl_rowpanel_vector( panel, vector, sendbuffer, 
					 recvcounts );

  if (loc_p_length*loc_p_width != 0)
    error += MPI_Reduce_scatter( BF(sendbuffer_ptr), BF(v_ptr), 
				 recvcounts, datatype, op, col_comm);

  if (error != 0)
    printf("direct_dupl_rowpanel_to_vector in PLA_Reduce completed \
with error = %d\n", error);

  pla_free( recvcounts);
  PLA_Obj_free( &sendbuffer);
  return (error);
}
/********************************************************************/







/********************************************************************/
int PLA_Pack_dupl_colpanel_mvector( PLA_Obj panel, PLA_Obj mvector, 
				  PLA_Obj sendbuffer, int *recvcounts)
/* 
   packs the column panel (width = 1) into a sendbuffer to be used
   by MPI_Reduce_Scatter. Also returns the recvcounts array of   
*/  
{
  int
    i,
    error = 0,
    me,
    idummy,
    z_or_1,
    size,
    np,
    nb,
    column, row,
    myrow, mycol,
    nprows, npcols,
    p,
    loc_p_width, loc_p_length,
    glob_mv_length, glob_mv_width,
    mv_align_row,
    current_vector_proc, current_vector_row,
    current_proc_col, current_proc_row;


  int 
    *displs = NULL;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    current_block = NULL,
    current_panel = NULL,
    remainder = NULL,
    current_sendbuffer = NULL;

  void
    *current_block_ptr = NULL,
    *p_buf = NULL,
    *current_panel_ptr = NULL,
    *current_sendbuffer_ptr = NULL;
  


  PLA_Panel_local_all( panel, &datatype, &loc_p_length, &loc_p_width,
		      &p_buf, &idummy);


  PLA_Mvector_global_all( mvector, &glob_mv_length, &glob_mv_width,
		        &mv_align_row, &template);

  PLA_Obj_template( panel, &template);
  
  PLA_Temp_vector_owner_rank( template, mv_align_row, &current_vector_proc);
  PLA_Temp_comm_col_size( template, &nprows);
  PLA_Temp_comm_row_size( template, &npcols);
  PLA_Temp_comm_col_rank( template, &myrow);
  PLA_Temp_comm_row_rank( template, &mycol);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_zero_or_one( template, (PLA_Start_num *) &z_or_1);


  np = nprows * npcols;

  /* first fill the recvcounts array */ 
  for(p = 0; p < np; p++)
    {
      column = p/nprows;
      row = p%nprows;
      if (row == myrow)
	PLA_Temp_vector_distant_length(template,
					p,
					glob_mv_length,
					mv_align_row,
					&recvcounts[column]);
    }

  /* multiply recvcuonts array by the panel width to get true recvcounts */
  for(column = 0; column < npcols; column++)
    recvcounts[column] *= glob_mv_width;


  /* might as well figure out the displacements in the sendbuffer : */
  displs = (int*)pla_malloc(npcols*sizeof(int));
  displs[0] = z_or_1;
  for(column = 1; column < npcols; column++)
    displs[column] = displs[column-1] + recvcounts[column-1]/glob_mv_width;


  /*take a view into the panel */
  PLA_Panel_subpanel( panel, PLA_DIM_ALL, PLA_DIM_ALL,
		     PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		     &current_panel);

  for(;;){
    PLA_Obj_top(current_panel, &size, &current_proc_row);
    if( size == 0) break;
    
    PLA_Obj_horz_split_2( current_panel, size, &current_block, &remainder);

    current_proc_col = current_vector_proc/nprows;

    PLA_Mscalar_submscalar( sendbuffer,
			  loc_p_width,
			  size,
			  PLA_ALIGN_FIRST,
			  displs[current_proc_col],
			  &current_sendbuffer);
    PLA_Obj_local_buffer( current_sendbuffer, &current_sendbuffer_ptr);
    


    if (myrow == current_proc_row)
      {

	PLA_Obj_extract_transpose_contents( current_block,
					    loc_p_width,
					    1,
					    &idummy, &idummy,
					    current_sendbuffer_ptr);
	displs[current_proc_col] += size;
      }
    
    current_vector_proc = (current_vector_proc + 1)%np;    
    PLA_Obj_swap( &current_panel, &remainder);
  }



  PLA_Obj_free( &current_sendbuffer);
  PLA_Obj_free( &remainder);
  PLA_Obj_free( &current_block);
  PLA_Obj_free( &current_panel);

  
  pla_free(displs);
  if (error != 0)
    printf("PLA_Pack_dupl_colpanel_to_mvector in PLA_Reduce completed with \
error = %d\n", error);

  return (error);
  
}









int direct_dupl_colpanel_to_mvector(PLA_Obj panel, PLA_Obj mvector, MPI_Op op)
/* reduction of the duplicated column panel to a vector */
{
  int    
    i,me,
    error = 0,
    *recvcounts,
    npcols,
    mycol,
    loc_p_length, loc_p_width,
    loc_mv_lld,
    idummy;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    sendbuffer = NULL,
    recvbuffer = NULL;

  MPI_Comm
    row_comm;
  
  void
    *sendbuffer_ptr = NULL,
    *recvbuffer_ptr = NULL,
    *p_buf = NULL,
    *mv_ptr = NULL;

  error += PLA_Panel_local_all(panel, &datatype, &loc_p_length, &loc_p_width,
			      &p_buf, &idummy);
  error += PLA_Obj_local_buffer(mvector, &mv_ptr);
  error += PLA_Obj_local_ldim( mvector, &loc_mv_lld);


  error += PLA_Obj_template( panel, &template);
  error += PLA_Temp_comm_row(template, &row_comm);
  error += PLA_Temp_comm_row_size(template, &npcols);
  error += PLA_Temp_comm_row_rank( template, &mycol);

  error += PLA_Mscalar_create( datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			     loc_p_width,
			     loc_p_length,
			     template,
			     &sendbuffer);
  error  += PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

		     

  recvcounts = (int *)pla_malloc(npcols*sizeof(int));
 

  error += PLA_Pack_dupl_colpanel_mvector( panel, mvector, sendbuffer, 
					  recvcounts );
  error += PLA_Mscalar_create( datatype, 
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			     loc_p_width,
			     recvcounts[mycol]/loc_p_width,
			     template,
			     &recvbuffer);
  error += PLA_Obj_local_buffer( recvbuffer, &recvbuffer_ptr);

  if (loc_p_length*loc_p_width != 0)
    error += MPI_Reduce_scatter( BF(sendbuffer_ptr), BF(recvbuffer_ptr), 
				 recvcounts, 
				datatype, op, row_comm);

  /*extract contents from the recievbuffer into the multiscalar */

  error += PLA_Obj_extract_transpose_contents( recvbuffer,
					       loc_mv_lld,
					       1,
					       &idummy, &idummy,
					       mv_ptr);
  pla_free( recvcounts);
  error += PLA_Obj_free( &sendbuffer);
  error += PLA_Obj_free( &recvbuffer);

  if (error != 0)
    printf("direct_dupl_colpanel_to_mvector in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);
}
/******************************************************************/







/******************************************************************/
int PLA_Pack_dupl_rowpanel_mvector( PLA_Obj panel, PLA_Obj mvector, 
				  PLA_Obj sendbuffer, int *recvcounts)
/* 
   packs the column panel (width = 1) into a sendbuffer to be used
   by MPI_Reduce_Scatter. Also returns the recvcounts array of   
*/  
{
  int
    i,
    error = 0,
    me,
    idummy,
    z_or_1,
    size,
    np,
    nb,
    column, row,
    myrow, mycol,
    nprows, npcols,
    p,
    loc_p_width, loc_p_length,
    glob_mv_length, glob_mv_width,
    mv_align_row,
    current_vector_proc, current_vector_row,
    current_proc_col, current_proc_row;


  int 
    *displs = NULL;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    current_block = NULL,
    current_panel = NULL,
    remainder = NULL,
    current_sendbuffer = NULL;

  void
    *current_block_ptr = NULL,
    *p_buf = NULL,
    *current_panel_ptr = NULL,
    *current_sendbuffer_ptr = NULL;
  


  PLA_Panel_local_all( panel, &datatype, &loc_p_length, &loc_p_width,
		      &p_buf, &idummy);


  PLA_Mvector_global_all( mvector, &glob_mv_length, &glob_mv_width,
		        &mv_align_row, &template);

  PLA_Obj_template( panel, &template);
  
  PLA_Temp_vector_owner_rank( template, mv_align_row, &current_vector_proc);
  PLA_Temp_comm_col_size( template, &nprows);
  PLA_Temp_comm_row_size( template, &npcols);
  PLA_Temp_comm_col_rank( template, &myrow);
  PLA_Temp_comm_row_rank( template, &mycol);
  PLA_Temp_nb( template, &nb);
  PLA_Temp_zero_or_one( template, (PLA_Start_num *) &z_or_1);


  np = nprows * npcols;

  /* first fill the recvcounts array */ 
  for(p = 0; p < np; p++)
    {
      column = p/nprows;
      row = p%nprows;
      if (column == mycol)
	PLA_Temp_vector_distant_length(template,
				      p,
				      glob_mv_length,
				      mv_align_row,
				      &recvcounts[row]);
    }

  /* multiply recvcuonts array by the panel width to get true recvcounts */
  for(row = 0; row < nprows; row++)
    recvcounts[row] *= glob_mv_width;


  /* might as well figure out the displacements in the sendbuffer : */
  displs = (int *)pla_malloc(nprows*sizeof(int));
  displs[0] = z_or_1;
  for(row = 1; row < nprows; row++)
    displs[row] = displs[row-1] + recvcounts[row-1]/glob_mv_width;


  /*take a view into the panel */
  PLA_Panel_subpanel( panel, PLA_DIM_ALL, PLA_DIM_ALL,
		     PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		     &current_panel);

  for(;;){
    PLA_Obj_left(current_panel, &size, &current_proc_col);
    if( size == 0) break;
    
    PLA_Obj_vert_split_2( current_panel, size, &current_block, &remainder);

    current_proc_row = current_vector_proc%nprows;

    PLA_Mscalar_submscalar( sendbuffer,
			  loc_p_length,
			  size,
			  PLA_ALIGN_FIRST,
			  displs[current_proc_row],
			  &current_sendbuffer);
    PLA_Obj_local_buffer( current_sendbuffer, &current_sendbuffer_ptr);
    


    if (mycol == current_proc_col)
      {
	PLA_Obj_extract_contents( current_block,
				  loc_p_length,
				  1,
				  &idummy, &idummy,
				  current_sendbuffer_ptr);
	displs[current_proc_row] += size;
      }
    
    current_vector_proc = (current_vector_proc + 1)%np;    
    PLA_Obj_swap( &current_panel, &remainder);
  }



  PLA_Obj_free( &current_sendbuffer);
  PLA_Obj_free( &remainder);
  PLA_Obj_free( &current_block);
  PLA_Obj_free( &current_panel);

  
  pla_free(displs);
  if (error != 0)
    printf("PLA_Pack_dupl_rowpanel_to_mvector in PLA_Reduce completed with \
error = %d\n", error);

  return (error);
  
}




int direct_dupl_rowpanel_to_mvector(PLA_Obj panel, 
				    PLA_Obj mvector, MPI_Op op)
/* reduction of the duplicated column panel to a vector */
{
  int    
    error = 0,
    *recvcounts,
    nprows,
    myrow,
    loc_p_length, loc_p_width,
    loc_mv_lld,
    idummy;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    sendbuffer = NULL,
    recvbuffer = NULL;

  MPI_Comm
    col_comm;
  
  void
    *sendbuffer_ptr = NULL,
    *recvbuffer_ptr = NULL,
    *p_buf = NULL,
    *mv_ptr = NULL;

  error += PLA_Panel_local_all(panel, &datatype, &loc_p_length, &loc_p_width,
			      &p_buf, &idummy);
  error += PLA_Obj_local_buffer(mvector, &mv_ptr);
  error += PLA_Obj_local_ldim( mvector, &loc_mv_lld);


  error += PLA_Obj_template( panel, &template);
  error += PLA_Temp_comm_col(template, &col_comm);
  error += PLA_Temp_comm_col_size(template, &nprows);
  error += PLA_Temp_comm_col_rank( template, &myrow);

  error += PLA_Mscalar_create( datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			     loc_p_length,
			     loc_p_width,
			     template,
			     &sendbuffer);
  error  += PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

		     

  recvcounts = (int *)pla_malloc(nprows*sizeof(int));
 

  error += PLA_Pack_dupl_rowpanel_mvector( panel, mvector, sendbuffer, 
					  recvcounts );

  error += PLA_Mscalar_create( datatype, 
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			      loc_p_length,
			      recvcounts[myrow]/loc_p_length,
			      template,
			      &recvbuffer);
  error += PLA_Obj_local_buffer( recvbuffer, &recvbuffer_ptr);


  if (loc_p_length*loc_p_width != 0)
    error += MPI_Reduce_scatter( BF(sendbuffer_ptr), BF(recvbuffer_ptr), 
				 recvcounts, 
				 datatype, op, col_comm);

  /*extract contents from the recievbuffer into the multiscalar */

  error += PLA_Obj_extract_transpose_contents( recvbuffer,
					       loc_mv_lld,
					       1,
					       &idummy, &idummy,
					       mv_ptr);

  pla_free( recvcounts);
  error += PLA_Obj_free( &sendbuffer);
  error += PLA_Obj_free( &recvbuffer);

  if (error != 0)
    printf("dupl_rowpanel_to_mvector in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);
}
/*******************************************************************/






/*******************************************************************/

int dupl_colpanel_to_matrix(PLA_Obj p,PLA_Obj m, MPI_Op op)
{
/* reduces duplicated panel to a matrix. matrix and panel have to be
row aligned and have matching dimentions */

  int
    error = 0,
    idummy,
    glob_p_length,
    glob_p_width,
    p_align_row;

  
  PLA_Template
    template;

  MPI_Datatype
    datatype;

  PLA_Obj
    temp = NULL;

  void
    *p_buf = NULL;
    

  PLA_Panel_local_all(p, &datatype, &idummy, &idummy,
		     &p_buf, &idummy);
  PLA_Panel_global_all(p, (PLA_Panel_rowcol_flag*)&idummy, 
		      &glob_p_length, &glob_p_width,
		      &p_align_row, &idummy,
		      &template);
  

  error += PLA_Mvector_create( datatype,
			    glob_p_length,
			    glob_p_width,
			    template,
			    p_align_row,
			    &temp);


  error += direct_dupl_colpanel_to_mvector(p, temp, op);
  
  error += PLA_Copy(temp, m);

  error += PLA_Obj_free(&temp);

  if (error != 0)
    printf("dupl_colpanel_to_matrix in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);
}
/*****************************************************************/







/*****************************************************************/
int direct_dupl_colpanel_to_colmatrix(PLA_Obj panel, PLA_Obj matrix, 
				   MPI_Op op)
/* reduction of the duplicated column panel to a matrix */
{
  int    
    error = 0,
    *recvcounts,
    size,
    owner_col,
    mycol,
    loc_p_length, loc_p_width,
    loc_m_lld,
    idummy;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    panel_view = NULL,
    matrix_view = NULL,
    current_matrix = NULL,
    current_panel = NULL,
    remainder_matrix = NULL,
    remainder_panel = NULL,
    sendbuffer = NULL,
    recvbuffer = NULL;

  MPI_Comm
    row_comm;
  
  void
    *sendbuffer_ptr = NULL,
    *recvbuffer_ptr = NULL,
    *p_buf = NULL,
    *current_m_ptr = NULL;



  error += PLA_Panel_local_all(panel, &datatype, &loc_p_length, &loc_p_width,
			      &p_buf, &idummy);

  error += PLA_Obj_local_ldim( matrix, &loc_m_lld);


  error += PLA_Obj_template( panel, &template);
  error += PLA_Temp_comm_row(template, &row_comm);
  error += PLA_Temp_comm_row_rank( template, &mycol);

		     

  PLA_Matrix_submatrix(matrix,
		      PLA_DIM_ALL, PLA_DIM_ALL,
		      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		      &matrix_view);

  PLA_Panel_subpanel(panel, PLA_DIM_ALL, PLA_DIM_ALL,
		    PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		    &panel_view);



  for(;;){

    error += PLA_Obj_left(matrix_view, &size, &owner_col);
    
    if (size == 0) break;

    error += PLA_Obj_vert_split_2(matrix_view, size, &current_matrix, 
				   &remainder_matrix);

    error += PLA_Obj_local_buffer(current_matrix, &current_m_ptr);

    error += PLA_Obj_vert_split_2(panel_view, size, &current_panel, 
				   &remainder_panel);
    
    error += PLA_Mscalar_create( datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			       loc_p_length,
			       size,
			       template,
			       &sendbuffer);
    error  += PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

    error += PLA_Obj_extract_contents( current_panel,
				        loc_p_length,
				        1,
				        &idummy, &idummy,
				        sendbuffer_ptr);

    error += PLA_Mscalar_create( datatype, 
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			        loc_p_length,
			        size,
			        template,
			        &recvbuffer);
    error += PLA_Obj_local_buffer( recvbuffer, &recvbuffer_ptr);
    
    if (loc_p_length*size != 0)
      error += MPI_Reduce( BF(sendbuffer_ptr), BF(recvbuffer_ptr), 
			   size*loc_p_length, 
			   datatype, op, owner_col, row_comm);
    

    /*extract contents from the recievbuffer into the current mutrix */
    
    if(owner_col == mycol)
      error += PLA_Obj_extract_contents( recvbuffer,
					 loc_m_lld,
					 1,
					 &idummy, &idummy,
					 current_m_ptr);

    PLA_Obj_swap(&matrix_view, &remainder_matrix);
    PLA_Obj_swap(&panel_view, &remainder_panel);
    
    error += PLA_Obj_free( &sendbuffer);
    error += PLA_Obj_free( &recvbuffer);
  }



  error += PLA_Obj_free(&remainder_matrix);
  error += PLA_Obj_free(&remainder_panel);
  error += PLA_Obj_free(&current_matrix);
  error += PLA_Obj_free(&current_panel);
  error += PLA_Obj_free(&panel_view);
  error += PLA_Obj_free(&matrix_view);


  if (error != 0)
    printf("dupl_colpanel_to_matrix in PLA_Reduce completed with error = %d\n",
	   error);


  return (error);
}
/******************************************************************/








/********************************************************************/

int dupl_rowpanel_to_matrix(PLA_Obj p, PLA_Obj m, MPI_Op op)

/* reduces duplicated panel to a matrix. matrix and panel have to be
column aligned and have matching dimentions */
 
{
  int
    error = 0,
    idummy,
    glob_p_length,
    glob_p_width,
    p_align_col;

  
  PLA_Template
    template;

  MPI_Datatype
    datatype;

  PLA_Obj
    temp = NULL;

  void
    *p_buf = NULL;

/* get local and global info from objects */  
  PLA_Panel_local_all(p, &datatype, &idummy, &idummy,
		     &p_buf, &idummy);
  PLA_Panel_global_all(p, (PLA_Panel_rowcol_flag*)&idummy, &glob_p_length, 
		      &glob_p_width,
		      &idummy, &p_align_col,
		      &template);
  

  error += PLA_Mvector_create( datatype,
			    glob_p_length,
			    glob_p_width,
			    template,
			    p_align_col,
			    &temp);

  error += direct_dupl_rowpanel_to_mvector(p, temp, op);
  
  error += PLA_Copy(temp, m);

  error += PLA_Obj_free(&temp);

  if (error != 0)
    printf("rowpanel_to_vector in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);
}

/********************************************************************/








/********************************************************************/
int direct_dupl_rowpanel_to_rowmatrix(PLA_Obj panel, PLA_Obj matrix, 
				   MPI_Op op)
/* reduction of the duplicated column panel to a matrix */
{
  int    
    error = 0,
    size,
    owner_row,
    nprows,
    myrow,
    loc_p_length, loc_p_width,
    loc_m_lld,
    idummy;

  MPI_Datatype
    datatype;

  PLA_Template
    template;

  PLA_Obj
    panel_view = NULL,
    matrix_view = NULL,
    current_matrix = NULL,
    current_panel = NULL,
    remainder_matrix = NULL,
    remainder_panel = NULL,
    sendbuffer = NULL,
    recvbuffer = NULL;

  MPI_Comm
    col_comm;
  
  void
    *sendbuffer_ptr = NULL,
    *recvbuffer_ptr = NULL,
    *p_buf = NULL,
    *current_m_ptr = NULL;



  error += PLA_Panel_local_all(panel, &datatype, &loc_p_length, &loc_p_width,
			      &p_buf, &idummy);

  error += PLA_Obj_local_ldim( matrix, &loc_m_lld);


  error += PLA_Obj_template( panel, &template);
  error += PLA_Temp_comm_col(template, &col_comm);
  error += PLA_Temp_comm_col_rank( template, &myrow);

		     

  PLA_Matrix_submatrix(matrix,
		      PLA_DIM_ALL, PLA_DIM_ALL,
		      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		      &matrix_view);

  PLA_Panel_subpanel(panel, PLA_DIM_ALL, PLA_DIM_ALL,
		    PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
		    &panel_view);



  for(;;){

    error += PLA_Obj_top(matrix_view, &size, &owner_row);
    
    if (size == 0) break;

    error += PLA_Obj_horz_split_2(matrix_view, size, &current_matrix, 
				   &remainder_matrix);

    error += PLA_Obj_local_buffer(current_matrix, &current_m_ptr);

    error += PLA_Obj_horz_split_2(panel_view, size, &current_panel, 
				   &remainder_panel);
    
    error += PLA_Mscalar_create( datatype,
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			       size,
			       loc_p_width,
			       template,
			       &sendbuffer);
    error  += PLA_Obj_local_buffer( sendbuffer, &sendbuffer_ptr);

    error += PLA_Obj_extract_contents( current_panel,
				        size,
				        1,
				        &idummy, &idummy,
				        sendbuffer_ptr);


    error += PLA_Mscalar_create( datatype, 
                    PLA_ALL_ROWS, PLA_ALL_COLS,
			        size,
			        loc_p_width,
			        template,
			        &recvbuffer);
    error += PLA_Obj_local_buffer( recvbuffer, &recvbuffer_ptr);
    
    if (loc_p_width*size != 0)
      error += MPI_Reduce( BF(sendbuffer_ptr), BF(recvbuffer_ptr), 
			   size*loc_p_width, 
			   datatype, op, owner_row, col_comm);
    

    /*extract contents from the recievbuffer into the current mutrix */
    
    if(owner_row == myrow)
      error += PLA_Obj_extract_contents( recvbuffer,
					 loc_m_lld,
					 1,
					 &idummy, &idummy,
					 current_m_ptr);

    error += PLA_Obj_swap(&matrix_view, &remainder_matrix);
    error += PLA_Obj_swap(&panel_view, &remainder_panel);
    
    error += PLA_Obj_free( &sendbuffer);
    error += PLA_Obj_free( &recvbuffer);
  }


  error += PLA_Obj_free(&remainder_matrix);
  error += PLA_Obj_free(&remainder_panel);
  error += PLA_Obj_free(&current_matrix);
  error += PLA_Obj_free(&current_panel);
  error += PLA_Obj_free(&matrix_view);
  error += PLA_Obj_free(&panel_view);


  if (error != 0)
    printf("direct dupl_rowpanel_to_matrix in PLA_Reduce completed with error = %d\n",
	   error);


  return (error);
}
/******************************************************************/







/******************************************************************/

int dupl_colpanel_to_rowmatrix(PLA_Obj panel, PLA_Obj matrix, MPI_Op op)
{
/* reduces duplicated column panel to a row oriented matrix. 
matrix and panel have to be row aligned and have matching dimentions */

  int
    error = 0,
    idummy,
    glob_p_length,
    glob_p_width,
    p_align_row;

  
  PLA_Template
    template;

  MPI_Datatype
    datatype;

  PLA_Obj
    temp = NULL;

  void
    *p_buf = NULL;
    

  error += PLA_Panel_local_all(panel, &datatype, &idummy, &idummy,
			      &p_buf, &idummy);
  error += PLA_Panel_global_all(panel, (PLA_Panel_rowcol_flag*)&idummy, 
			       &glob_p_length, &glob_p_width,
			       &p_align_row, &idummy,
			       &template);
  

  error += PLA_Mvector_create( datatype,
			    glob_p_length,
			    glob_p_width,
			    template,
			    p_align_row,
			    &temp);


  error += direct_dupl_colpanel_to_mvector(panel, temp, op);
  
  error += PLA_Copy(temp, matrix);

  error += PLA_Obj_free(&temp);

  if (error != 0)
    printf("dupl_colpanel_to_rowmatrix in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);

}
/********************************************************************/









/********************************************************************/
int dupl_rowpanel_to_colmatrix(PLA_Obj panel, PLA_Obj matrix, MPI_Op op)
{

/* reduces duplicated rowpanel to a column_oriented matrix. matrix and 
panel have to be column aligned and have matching dimentions */

  int
    error = 0,
    idummy,
    glob_p_length,
    glob_p_width,
    p_align_col;

  
  PLA_Template
    template;

  MPI_Datatype
    datatype;

  PLA_Obj
    temp = NULL;

  void
    *p_buf = NULL;

/* get local and global info from objects */  
  error += PLA_Panel_local_all(panel, &datatype, &idummy, &idummy,
			      &p_buf, &idummy);
  error += PLA_Panel_global_all(panel, (PLA_Panel_rowcol_flag*)&idummy, 
			       &glob_p_length, 
			       &glob_p_width,
			       &idummy, &p_align_col,
			       &template);
  

  error += PLA_Mvector_create( datatype,
			    glob_p_width,
			    glob_p_length,
			    template,
			    p_align_col,
			    &temp);

  error += direct_dupl_rowpanel_to_mvector(panel, temp, op);
  
  error += PLA_Copy(temp, matrix);

  error += PLA_Obj_free(&temp);

  if (error != 0)
    printf("dupl_rowpanel_to_colmatrix in PLA_Reduce completed with error = %d\n",
	   error);

  return (error);
}
/********************************************************************/







/********************************************************************/
int dupl_mscalar_to_mscalar( PLA_Obj dms, MPI_Op op, PLA_Obj ms)
{
  int
    idummy,
    lld,
    nprows, root,
    owner_col_to, owner_row_to,
    owner_col_from, owner_row_from,
    global_length, global_width;

  PLA_Template template;

  PLA_Obj
    from = NULL, to = NULL;

  void
    *from_buffer = NULL, *to_buffer = NULL, 
    *dms_buffer = NULL, *ms_buffer = NULL;

  MPI_Comm 
    comm_all;

  MPI_Datatype
    datatype;

  PLA_Obj_template( dms, &template );
  PLA_Temp_comm1D( template, &comm_all);

  PLA_Obj_owner_col( dms, &owner_col_from );
  PLA_Obj_owner_row( dms, &owner_row_from );
  PLA_Obj_owner_col( ms, &owner_col_to );
  PLA_Obj_owner_row( ms, &owner_row_to );

  PLA_Obj_local_buffer( dms, &dms_buffer );
  PLA_Obj_local_buffer( ms, &ms_buffer );


  PLA_Obj_global_length( dms, &global_length);
  PLA_Obj_global_width( dms, &global_width);

  PLA_Obj_datatype( dms, &datatype);
  
  
  PLA_Mscalar_create_conf_to( dms, PLA_ALL_ROWS, PLA_ALL_COLS,
			     &from);
  PLA_Obj_local_buffer( from, &from_buffer );
  PLA_Obj_local_ldim( from, &lld);

  PLA_Mscalar_create_conf_to( ms, PLA_ALL_ROWS, PLA_ALL_COLS,
			     &to);
  PLA_Obj_local_buffer( to, &to_buffer );
  if (ms_buffer == 0)
    PLA_Obj_set_to_zero(from);
  
  PLA_Obj_get_local_contents( dms, PLA_NO_TRANSPOSE, &idummy, &idummy,
			        from_buffer, lld, 1);

  if(owner_col_to != PLA_ALL_COLS && 
     owner_row_to != PLA_ALL_ROWS)
    {
      PLA_Temp_comm_col_size( template, &nprows);
      root = owner_col_to * nprows + owner_row_to;
      MPI_Reduce( from_buffer, to_buffer, global_length*global_width,
		 datatype, op, root, comm_all ); 
    }
  else
    {
      MPI_Allreduce( from_buffer, to_buffer, global_length*global_width,
		datatype, op, comm_all ); 
    }
  PLA_Copy(to,ms);

  PLA_Obj_free(&from); PLA_Obj_free(&to);
  return(0);
}
/********************************************************************/




/********************************************************************/
int reduce_misaligned(PLA_Obj panel, PLA_Obj object, MPI_Op op)
{

  int
    error = 0,
    shift,
    obj_type,
    mv_width,
    glob_p_length, glob_p_width,
    p_align_row, p_align_col,
    p_rowcol, o_rowcol,
    p_align, obj_align;

  PLA_Template
    template;

  MPI_Datatype
    datatype;

  PLA_Obj
    temp = NULL;



  error += PLA_Obj_objtype(object, (PLA_Obj_obj_type*)&obj_type);
  error += PLA_Obj_datatype(panel, &datatype);
  error += PLA_Panel_global_all(panel, (PLA_Panel_rowcol_flag *)&p_rowcol,
			       &glob_p_length, &glob_p_width,
			       &p_align_row, &p_align_col,
			       &template);

  if (PLA_PROJ_ONTO_COL == p_rowcol)
    p_align = p_align_row;
  else /* PLA_PROJ_ONTO_ROW == rowcol */
    p_align = p_align_col;

  if (obj_type == PLA_PMVECTOR)
    if (PLA_PROJ_ONTO_COL == p_rowcol)
      mv_width = glob_p_width;
    else /* PLA_PROJ_ONTO_ROW == rowcol */
      mv_width = glob_p_length;
/*
  printf("glob_mv_length = %d, glob_mb_width = %d, mv_align = %d\n",
	 glob_p_length, glob_p_width, p_align);
*/
  error += PLA_Mvector_create_conf_to(panel,
					mv_width,
					&temp);

  error += PLA_Reduce(panel, op, temp);



  if (PLA_MATRIX == obj_type || 
      PLA_PVECTOR == obj_type ||
      PLA_PMVECTOR == obj_type )
    error += PLA_Obj_project_onto(object, 
				      (PLA_Panel_rowcol_flag *)&o_rowcol);
  else
    o_rowcol = PLA_PROJ_ONTO_COL;


  if (PLA_PROJ_ONTO_COL == o_rowcol)
    error += PLA_Obj_global_align_row(object, &obj_align);
  else /* PLA_PROJ_ONTO_ROW == o_rowcol */
    error += PLA_Obj_global_align_col(object, &obj_align);


  shift = obj_align - p_align;
/*
  printf("obj_align = %d, p_align = %d, shift = %d\n",obj_align, p_align, shift);
*/

  error += pla_shift(&temp, shift);

  error += PLA_Copy(temp, object);  
  error += PLA_Obj_free(&temp);

  if (error != 0)
    printf("reduce_misaligned in PLA_Reduce completed with error = %d\n",
	   error);
  
  return (error);

}
/********************************************************************/












/********************************************************************/

int PLA_Reduce(PLA_Obj p, MPI_Op op, PLA_Obj o)
{
/* this routine requires p to be of a panel type */ 
  int
    p_align,o_align;


  PLA_Panel_rowcol_flag
    panel_type;

  PLA_Obj_obj_type
    dest_obj_type, source_obj_type;

  PLA_Panel_rowcol_flag
    dest_panel_type;
  
  PLA_Duplicate_flag
    dest_duplicate_flag;


  PLA_Obj_project_onto( p, &panel_type);
  PLA_Obj_objtype( o, &dest_obj_type);
  PLA_Obj_objtype( p, &source_obj_type );

  if( source_obj_type == PLA_MSCALAR ){
    dupl_mscalar_to_mscalar( p, op, o);
    return(0);
  }


  if ((PLA_PVECTOR == dest_obj_type) || 
      (PLA_PMVECTOR == dest_obj_type) ||
      (PLA_MATRIX == dest_obj_type)){
    PLA_Obj_project_onto( o, &dest_panel_type);
    PLA_Obj_annotate_duplicate( o, &dest_duplicate_flag);
  }
  else /* if destination object is vector or a multivector */
    dest_panel_type = PLA_PROJ_ONTO_COL;

/* determine if the objects are misaligned. If yes then call the misaligned 
reduce (slow), if no proceed to the switch statement */

  if (PLA_PROJ_ONTO_COL == panel_type)
    PLA_Obj_global_align_row(p, &p_align);
  else
    PLA_Obj_global_align_col(p, &p_align);


  
  if (PLA_PROJ_ONTO_COL == dest_panel_type)
    PLA_Obj_global_align_row(o, &o_align);
  else
    PLA_Obj_global_align_col(o, &o_align);


  if (o_align != p_align)
    {
      if (FIRST_WARNING){
	printf("reducing to the misaligned object\n");
	printf("this warning would not be broadcasted again\n");
	FIRST_WARNING = 0;
      }
      reduce_misaligned(p,o,op);
      return(0);
    }

  switch( panel_type )
    {
    case PLA_PROJ_ONTO_COL:
      {
	switch( dest_obj_type)
	  {
	  case PLA_VECTOR :
	    direct_dupl_colpanel_to_vector(p,o,op);
	    /* colpanel_to_vector(p,o,op); */      
	    break;
	  case PLA_MVECTOR :
	    direct_dupl_colpanel_to_mvector(p,o,op);
	    break;
	  case PLA_PVECTOR :
	  case PLA_PMVECTOR :
	    switch(dest_panel_type)
	      {
	      case PLA_PROJ_ONTO_ROW :
		dupl_colpanel_to_rowmatrix(p,o, op);
		break;
	      case PLA_PROJ_ONTO_COL:
		switch( dest_duplicate_flag )
		  {
		  case PLA_ALL_COLS :
		    dupl_colpanel_to_dupl_colpanel(p,o,op);
		    break;
		  case PLA_NO_DUPLICATE :
		    /* a strange case, I don't think it will be 
		       implemented ever  */
		    break;
		  }
		break;		
	      }
	    break;
	  case PLA_MATRIX :
	    switch(dest_panel_type)
	      {
	      case PLA_PROJ_ONTO_COL:
		/*	    dupl_colpanel_to_matrix(p,o,op); */
		direct_dupl_colpanel_to_colmatrix(p,o,op);
		break;
	      case PLA_PROJ_ONTO_ROW:
		dupl_colpanel_to_rowmatrix(p,o,op);
		break;
	      }
	    break;
	  }	
	break;
      }
      break;
    case PLA_PROJ_ONTO_ROW:
      {
	switch( dest_obj_type)
	  {
	  case PLA_VECTOR :
	    direct_dupl_rowpanel_to_vector(p,o,op);
	    /*    rowpanel_to_vector(p,o,op);  */
	    break;
	  case PLA_MVECTOR :
	    direct_dupl_rowpanel_to_mvector(p,o,op);
	    break;
	  case PLA_PVECTOR :
	  case PLA_PMVECTOR :
	    switch(dest_panel_type)
	      {
	      case PLA_PROJ_ONTO_ROW :
		dupl_rowpanel_to_dupl_rowpanel(p,o,op);
		break;
	      case PLA_PROJ_ONTO_COL:
		dupl_rowpanel_to_colmatrix(p,o,op);
		break;
	      }
	    break;
	  case PLA_MATRIX :
	    switch(dest_panel_type)
	      {
	      case PLA_PROJ_ONTO_ROW:
		/*	    dupl_rowpanel_to_matrix(p,o,op); */
		direct_dupl_rowpanel_to_rowmatrix(p,o,op);
		break;
	      case PLA_PROJ_ONTO_COL:
		dupl_rowpanel_to_colmatrix(p,o,op);
		break;
	      }
	    break;
	  }
	break;
      }
    }
return 0;
}

