

/****************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/





#include "PLA.h"
#include <stdio.h>

/* Incredibly thick-skulled patch for PLA_Reduce so that  */
/* the user can reduce and add to the destination object */

int PLA_Reduce_add(PLA_Obj Source, MPI_Op Operation, PLA_Obj Destination)
{
   PLA_Obj_obj_type object_type;
   int global_length, global_width, global_align_row;
   int global_align_column, orientation, global_align;
   int project, owner_row, owner_col;
   PLA_Template  template = NULL;
   MPI_Datatype datatype;
   PLA_Obj One = NULL, temporary_object = NULL;
   double d_one = 1.0;
   float s_one = 1.0;

   PLA_Obj_objtype(Destination, &object_type);
   PLA_Obj_datatype(Destination, &datatype);
   PLA_Obj_template(Destination, &template);
   PLA_Obj_global_info(Destination, &global_length, &global_width, 
                       &project, &owner_row, &owner_col,
                       &global_align_row, &global_align_column);
   if(object_type == PLA_VECTOR)
   {
      PLA_Vector_create(datatype, global_length, template, global_align_row,
                         &temporary_object);
   }
   else if(object_type == PLA_MVECTOR)
   {
      PLA_Mvector_create(datatype, global_length, global_width, template, global_align_row,
                          &temporary_object);
   }
   else if(object_type == PLA_MATRIX)
   {
      PLA_Obj_project_onto(Destination, &orientation);
      PLA_Matrix_create(datatype, global_length, global_width, template, global_align_row,
                         global_align_column, &temporary_object);
      PLA_Obj_set_orientation(temporary_object, orientation);
   PLA_Obj_objtype(Source, &object_type);
   PLA_Obj_datatype(Source, &datatype);
   PLA_Obj_template(Source, &template);
   PLA_Obj_global_info(Source, &global_length, &global_width, 
                       &project, &owner_row, &owner_col,
                       &global_align_row, &global_align_column);
   }
   else if(object_type == PLA_PVECTOR || object_type == PLA_PMVECTOR)
   {
      PLA_Obj_project_onto(Destination, &orientation);
/* If I query a Pmvector do I get the projected length and width that I use to specify it's */
/* creation or the unprojected?                                                             */
      if(orientation == PLA_PROJ_ONTO_ROW) global_align = global_align_column;
      else if(orientation == PLA_PROJ_ONTO_COL) global_align = global_align_row;
      else
      {
         fprintf(stdout, "Error on panel orientation in PLA_Reduce_add\n");
         fflush(stdout);
         return(-1);
      }
      if (orientation == PLA_PROJ_ONTO_COL)
         PLA_Pmvector_create(datatype, orientation, PLA_ALL_COLS, 
                 global_length, global_width, template, global_align, 
                  &temporary_object);
      else
         PLA_Pmvector_create(datatype, orientation, PLA_ALL_ROWS, 
                 global_length, global_width, template, global_align, 
                  &temporary_object);
   }
   else if(object_type == PLA_MSCALAR)
   {
      PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, global_length, global_width, template, &temporary_object);
   }
   else
   {
      fprintf(stdout, "Error in call to PLA_Reduce_add -- unrecognized ");
      fprintf(stdout, "object type\n");
      fflush(stdout);
      return(-1); /* Temporary */
   }
   PLA_Reduce(Source, Operation, temporary_object);
/*baker 8.30.96*/
   PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, template, &One);
   PLA_Obj_set_to_one(One);
   PLA_Axpy(One, temporary_object, Destination);
   PLA_Obj_free(&One);
   PLA_Obj_free(&temporary_object);
   return(0);
}

