
/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/

#include "PLA.h"
#include "MISC_DEFS.h"
/* the defines below should be done in PLA.h */
int pla_shape_obj(int shape, PLA_Obj obj)
/*
   puts zeros into the obj's buffer as dictated by shape.
   shape can be PLA_SHAPE_UPPER_TRAPEZOIDAL and PLA_SHAPE_LOWER_TRAPEZOIDAL
   by Philip Alpatov philip@physics.utexas.edu
*/
{
   int
   error = 0, i, size, length, width;

   PLA_Obj
       current_obj = NULL,  zero_mask = NULL;

   error += PLA_Obj_global_length(obj, &length);
   error += PLA_Obj_global_width(obj, &width);

   if (PLA_SHAPE_LOWER_TRAPEZOIDAL == shape && length >= width)
      error += PLA_Obj_view(obj, width, width,
                            PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &current_obj);
   else if (PLA_SHAPE_LOWER_TRAPEZOIDAL == shape && length < width)
   {
      error += PLA_Obj_view(obj, length, length,
                            PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &current_obj);
      error += PLA_Obj_view_shift(current_obj,               0,
                                                 width-length,     width-length,
                                                               0);
   }
   else if (PLA_SHAPE_UPPER_TRAPEZOIDAL == shape && length >= width)
   {
      error += PLA_Obj_view(obj, width, width,
                            PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &current_obj);
      error += PLA_Obj_view_shift(current_obj,          length-width,
                                                      0,                0,
                                                          length-width);
   }
   else if (PLA_SHAPE_UPPER_TRAPEZOIDAL == shape && length < width)
      error = PLA_Obj_view(obj, length, length,
                           PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &current_obj);
   else if(PLA_SHAPE_GENERAL == shape)
      /* do nothing but avoid error flag */ ;
   else
   {
      printf("illegal combination of shape/orientation in pla_shape_obj\n");
      return(++error);
   }
   switch (shape)
   {
   case PLA_SHAPE_LOWER_TRAPEZOIDAL:
      if ( error == 0)
      {
         PLA_Obj_view(obj, 0, 1, PLA_ALIGN_FIRST,PLA_ALIGN_FIRST, &zero_mask);
         PLA_Obj_global_length(obj, &size);
         for(i = 0; i < (size - 1); i++)
         {
            PLA_Obj_view_shift(zero_mask,              0,
                                             1,                    1,
                                                       1);
            PLA_Obj_set_to_zero(zero_mask);
         }
      }
      break;
   case PLA_SHAPE_UPPER_TRAPEZOIDAL:
      if ( error == 0)
      {
         PLA_Obj_global_length(obj, &size);
         PLA_Obj_view(obj,size,0,PLA_ALIGN_FIRST,PLA_ALIGN_FIRST, &zero_mask);
         PLA_Obj_view_shift(zero_mask,                 1,
                                          0,                        1,
                                                       0);
         for(i = 0; i < (size - 1); i++)
         {
            PLA_Obj_set_to_zero(zero_mask);
            PLA_Obj_view_shift(zero_mask,              1,
                                            1,                    1,
                                                       0);
         }
      }
      break;
   }
   PLA_Obj_free ( & current_obj );
   PLA_Obj_free ( & zero_mask );
   if (error != 0)
      printf("pla_shape_obj completed with error = %d\n", error);
   return(error);
}

/* Scales                                                     */
int pla_scal_shape_obj(int shape, PLA_Obj alpha, PLA_Obj obj)
{
   int
   error = 0, i, size, length, width;

   PLA_Obj
       square_obj = NULL,  chunk_obj = NULL, scale_mask = NULL;

   error += PLA_Obj_global_length(obj, &length);
   error += PLA_Obj_global_width(obj, &width);

   if (PLA_SHAPE_LOWER_TRAPEZOIDAL == shape && length >= width)
      error += PLA_Obj_horz_split_2(obj, width, &square_obj,
                                                  &chunk_obj);
   else if (PLA_SHAPE_LOWER_TRAPEZOIDAL == shape && length < width)
      error += PLA_Obj_vert_split_2(obj, -length, &chunk_obj,&square_obj);
   else if (PLA_SHAPE_UPPER_TRAPEZOIDAL == shape && length >= width)
      error += PLA_Obj_horz_split_2(obj, -width, &chunk_obj,
                                                 &square_obj);
   else if (PLA_SHAPE_UPPER_TRAPEZOIDAL == shape && length < width)
      error += PLA_Obj_vert_split_2(obj, length, &square_obj, &chunk_obj);
      /* General is a simple case and handled separately */
   else if (PLA_SHAPE_GENERAL == shape)
      error += PLA_Scal(alpha, obj);

   else
   {
      printf("illegal combination of shape/orientation in pla_scal_shape_obj\n");
      return(++error);
   }
   switch (shape)
   {
   case PLA_SHAPE_UPPER_TRAPEZOIDAL:
      PLA_Scal(alpha, chunk_obj);
      if ( error == 0)
      {
         PLA_Obj_view(square_obj, 1, 1,
                      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &scale_mask);
         PLA_Obj_global_length(obj, &size);
         for(i = 0; i < size; i++)
         {
            PLA_Scal(alpha, scale_mask);
            if(i < (size - 1))
               PLA_Obj_view_shift(scale_mask,             0,
                                               1,                    1,
                                                          1);
         }
      }
      break;
   case PLA_SHAPE_LOWER_TRAPEZOIDAL:
      PLA_Scal(alpha, chunk_obj);
      if ( error == 0)
      {
         PLA_Obj_global_length(obj, &size);
         PLA_Obj_view(square_obj, size, 1,
                      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST, &scale_mask);
         for(i = 0; i < size; i++)
         {
            PLA_Scal(alpha, scale_mask);
            if( i < (size - 1))
               PLA_Obj_view_shift(scale_mask,             1,
                                               1,                    1,
                                                          0);
         }
      }
      break;
   }
   PLA_Obj_free ( & square_obj );
   PLA_Obj_free ( & chunk_obj );
   PLA_Obj_free ( & scale_mask );
   if (error != 0)
      printf("pla_scal_shape_obj completed with error = %d\n", error);
   return(error);
}


int PLA_Reduce_x(int shape, PLA_Obj panel,
PLA_Obj alpha, PLA_Obj o)
{
/*
   reduces the duplicated panel of specified shape to a object o
   while adding the scaled object itself
   by Philip Alpatov philip@phisics.utexas.edu

   o = reduce(alpha * shape(dupl_panel)) + beta*o
*/
   int
   error = 0,
   owner_row,
   owner_col,
   source_orientation,
   destination_orientation,
   second_shape;

   PLA_Obj
       alpha_dup  = NULL,
       temp_panel = NULL;

   PLA_Obj_get_orientation(panel, &source_orientation);
   PLA_Obj_get_orientation(o,     &destination_orientation);
   if(source_orientation != destination_orientation)
   {
      if(PLA_SHAPE_UPPER_TRAPEZOIDAL == shape)
         second_shape == PLA_SHAPE_LOWER_TRAPEZOIDAL;
      else if(PLA_SHAPE_LOWER_TRAPEZOIDAL == shape)
         second_shape == PLA_SHAPE_UPPER_TRAPEZOIDAL;
      else second_shape == shape;
   }
   else second_shape = shape;
   PLA_Obj_owner_row(alpha, &owner_row);
   PLA_Obj_owner_col(alpha, &owner_col);
   if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS )
   {
      PLA_Mscalar_create_conf_to( alpha, PLA_ALL_ROWS, PLA_ALL_COLS,&alpha_dup);
      PLA_Copy( alpha, alpha_dup );
   }
   

   error += PLA_Obj_clone(panel, &temp_panel);
   error += PLA_Copy(panel, temp_panel);
   error += pla_shape_obj(shape, temp_panel);
   error += pla_scal_shape_obj(second_shape,alpha_dup==NULL?alpha:alpha_dup,o);
   error += PLA_Reduce_add(temp_panel, MPI_SUM, o);
   error += PLA_Obj_free(&temp_panel);
   error += PLA_Obj_free(&alpha_dup);

   if (error != 0)
      printf("PLA_Reduce_x completed with error = %d\n", error);

   return(error);
}

