/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/



#include "PLA_Machines.h"


#ifndef  PLA_H
#define  PLA_H

/* 
   This header should be included in all source code using the PLA_Blas 
   subroutines.  It contains definitions of constants and structures 
   used in those subroutines. 
   */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#ifndef NULL
#define NULL 0
#endif

/* 
   Got this idea a little late.  Maybe we should have an PLA_NULL to initialize
   things to.  Just to keep things under our control.
   */
#ifndef PLA_NULL
#define PLA_NULL 0
#endif

#ifndef PLA_DUMMY
#define PLA_DUMMY PLA_NULL
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define  PLA_ALIGN_FIRST            -1
#define  PLA_DIM_ALL                -2
#define  PLA_LEADING_DIM_TIGHT_FIT  -3
#define  PLA_ALL_COLS               -4
#define  PLA_ALL_ROWS               -5
#define  PLA_INHERIT                -6


enum { PLA_ZERO, PLA_ONE };
typedef int PLA_Start_num;

enum {
  PLA_NO_DUPLICATE
};
typedef int PLA_Duplicate_flag;

enum {
  PLA_UNSPECIFIED_PANEL=11,
  PLA_PROJ_ONTO_ROW, 
  PLA_PROJ_ONTO_COL,
  PLA_ROWVECTOR,    /* Added TEMPORARILY by JAG on 07/07/96 PLA.h */
  PLA_COLVECTOR     /* to compile a stable version  JAG  PLA.h */
};
typedef int PLA_Panel_rowcol_flag;

typedef struct PLA_Template_def *PLA_Template;

enum  { 
  PLA_MSCALAR=21,
  PLA_VECTOR, 
  PLA_MVECTOR, 
  PLA_MATRIX,
  PLA_PVECTOR,
  PLA_PMVECTOR
};
/* Needed for some of the error test routines */
#define PLA_ANY_VECTOR 31

typedef int PLA_Obj_obj_type;

typedef struct PLA_Obj_def *PLA_Obj;

union pre_base_complex{
   double not_used;
   float  val[2];
};

struct base_complex {
   union pre_base_complex value;
};

struct base_double_complex {
   double real;
   double imaginary;
};

typedef struct base_complex          PLA_COMPLEX;
typedef struct base_double_complex   PLA_DOUBLE_COMPLEX;


#include "PLA_External.h"
#include "PLA_Init.h"
#include "PLA_BLAS.h"
#include "PLA_Temp.h"
#include "PLA_Obj.h"
#include "PLA_Obj_set.h"
#include "PLA_Conformal.h"
#include "PLA_Mscalar.h"
#include "PLA_Vector.h"
#include "PLA_Mvector.h"
#include "PLA_Pmvector.h"
#include "PLA_Matrix.h"
#include "PLA_Panel.h"
#include "PLA_BLAS.h"    /* Machine independence */
#include "MISC_DEFS.h"  /* PLA_TRANS, etc. defined */
#include "PLA_Wrapper.h" /* Wrapper Routines */
#include "PLA_copy.h"
#include "PLA_Reduce.h"
#include "PLA_Gemm.h"
#include "PLA_Trsv.h"
#include "PLA_Syrk.h"
#include "PLA_Chol.h"
#include "PLA_UTIL.h"
  
#endif

