
/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu


  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/




#include "PLA.h"
#include <math.h>
#include <stdio.h>
#include "PLA_Local_BLAS1.h"

#if MANUFACTURE != CRAY

extern double PLA_ddot(int *, double *, int *, double *, int*);
extern float PLA_sdot(int *, float *, int *, float *, int*);
/* extern PLA_COMPLEX PLA_cdotu(int *, PLA_COMPLEX *, int *, PLA_COMPLEX *, int*);
extern PLA_DOUBLE_COMPLEX  PLA_zdotu(int *, PLA_DOUBLE_COMPLEX *, int *, 
                                    PLA_DOUBLE_COMPLEX *, int*); */

extern double PLA_dnrm2(int *, double *, int *);
extern float PLA_snrm2(int *, float *, int *);
extern PLA_COMPLEX PLA_cnrm2(int *, PLA_COMPLEX *, int *);
extern PLA_DOUBLE_COMPLEX PLA_znrm2(int *, PLA_DOUBLE_COMPLEX *, int *);

extern double PLA_dasum(int *, double *, int *);
extern float PLA_sasum(int *, float *, int *);
extern PLA_COMPLEX PLA_casum(int *, PLA_COMPLEX *, int *);
extern PLA_DOUBLE_COMPLEX PLA_zasum(int *, PLA_DOUBLE_COMPLEX *, int *);

extern int PLA_idamax(int *, double *, int *);
extern int PLA_isamax(int *, float *, int *);
extern int PLA_icamax(int *, PLA_COMPLEX *, int *);
extern int PLA_izamax(int *, PLA_DOUBLE_COMPLEX *, int *);

#endif

int PLA_Local_copy_enter(PLA_Obj x, PLA_Obj y)
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_copy";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( x, "x", error, original_error);
      PLA_Obj_object_valid( y, "y", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( x, "x", error, original_error );
      PLA_Obj_datatype( x, &datatype );
      PLA_Obj_datatype_equal( y, datatype, "y", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( x, &template );
      PLA_Obj_template_equal( y, template, "y", error, original_error );

    /* check dimensions */
         PLA_Obj_local_length_length_equal( x, y, "x and y", error, original_error );
         PLA_Obj_local_width_width_equal  ( x, y, "x and y", error, original_error );
   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_copy_exit(PLA_Obj x, PLA_Obj y)
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}



/* Note : I am really unsure what to do if the strides are not */
/*        the same.  Should y's field indicate the new stride? */
/*        Should the function exit?  PLA_BAD_STRIDE?           */

int PLA_Local_copy(PLA_Obj x, PLA_Obj y)
{
   int original_error = 0;


   int error_value;
   int local_length_x, local_width_x;
   int ld_y_buf, stride_y_buf;
   int rows_in_x_buf, cols_in_x_buf;

   int local_length_y, local_width_y;

   MPI_Datatype datatype_x, datatype_y;

   void *y_buf;
void *x_buf;
PLA_Obj_local_buffer(x, &x_buf);
   PLA_Local_copy_enter(x, y);

   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);
   PLA_Obj_local_length(y, &local_length_y);
   PLA_Obj_local_width (y, &local_width_y);
   PLA_Obj_datatype(y, &datatype_y);

   PLA_Obj_local_buffer(y, &y_buf);
   PLA_Obj_local_stride(y, &stride_y_buf);
   PLA_Obj_local_ldim  (y, &ld_y_buf);
/* Note -- type independence is handled in PLA_Obj_get_local_contents -- obviously */
   if(local_length_x * local_width_x > 0)
      PLA_Obj_get_local_contents(x, PLA_NO_TRANS, 
             &rows_in_x_buf, &cols_in_x_buf, y_buf, ld_y_buf, stride_y_buf);
   PLA_Local_copy_exit(x, y);
}

/******************************************************************************/


int PLA_Local_swap_enter(PLA_Obj x, PLA_Obj y)
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_swap";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( x, "x", error, original_error);
      PLA_Obj_object_valid( y, "y", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( x, "x", error, original_error );
      PLA_Obj_datatype( x, &datatype );
      PLA_Obj_datatype_equal( y, datatype, "y", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( x, &template );
      PLA_Obj_template_equal( y, template, "y", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_length_equal( x, y, "x and y", error, original_error );
      PLA_Obj_local_width_width_equal  ( x, y, "x and y", error, original_error );
   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_swap_exit(PLA_Obj x, PLA_Obj y)
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

/* Note : I am really unsure what to do if the strides are not */
/*        the same.  Should y's field indicate the new stride? */
/*        Should the function exit?  PLA_BAD_STRIDE?           */

int PLA_Local_swap(PLA_Obj x, PLA_Obj y)
{
   int original_error = 0;


   int error_value;
   int local_length_x, local_width_x;
   int ld_x_buf, stride_x_buf;
   int rows_in_x_buf, cols_in_x_buf;
   int size_of_data;

   int local_length_y, local_width_y;
   int ld_y_buf, stride_y_buf;
   int rows_in_y_buf, cols_in_y_buf;

   MPI_Datatype datatype_x, datatype_y;

   void *y_buf, *temp_buf, *x_buffer, *x_buffer_cur;

   PLA_Local_swap_enter(x, y);
   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);

   PLA_Obj_local_length(y, &local_length_y);
   PLA_Obj_local_width (y, &local_width_y);
   PLA_Obj_datatype(y, &datatype_y);

   /* PLA_Obj_local_buffer(x, &x_buf); */
   PLA_Obj_local_stride(x, &stride_x_buf);
   PLA_Obj_local_ldim  (x, &ld_x_buf);

   PLA_Obj_local_buffer(y, &y_buf);
   PLA_Obj_local_stride(y, &stride_y_buf);
   PLA_Obj_local_ldim  (y, &ld_y_buf);

   MPI_Type_size( datatype_x, &size_of_data);

   if(local_length_x * local_width_x > 0)
   {
      temp_buf = (void*) pla_malloc(local_length_x*local_width_y*size_of_data);
      if(NULL == temp_buf)
         PLA_Log_abort("PLA_Local_swap", "could not allocate storage", PLA_BAD_MALLOC, original_error);
/* ld_y_buf is used arbitrarily -- the temp buffer just makes for faster      */
/* copies if it is exactly like either the x or the y buffer in how we view it*/
/* Put y in temp */
      PLA_Obj_get_local_contents(y, PLA_NO_TRANS, &rows_in_y_buf, &cols_in_y_buf,
                                 temp_buf, ld_y_buf, stride_y_buf);
/* Put x in y */
      PLA_Obj_get_local_contents(x, PLA_NO_TRANS, &rows_in_x_buf, &cols_in_x_buf,
                                 y_buf, ld_y_buf, stride_y_buf);
/* Put temp in x */
      PLA_Obj_set_local_contents(PLA_NO_TRANS, rows_in_y_buf, cols_in_y_buf,
                                 temp_buf, ld_y_buf, stride_y_buf, x);
   }
   PLA_Local_swap_exit(x, y);
   return(PLA_SUCCESS);
}



/******************************************************************************/
/******************************************************************************/


int PLA_Local_scal_enter(PLA_Obj alpha, PLA_Obj x)
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_scal";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( x, "x", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x, template, "x", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal  ( alpha, 1, "alpha", error, original_error );
   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_scal_exit(PLA_Obj alpha, PLA_Obj x)
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

/* Why is there a case for the object being row-oriented?  Well, the optimization*/
/* isn't at all complete.  The idea is that such objects will TEND to be narrow -*/
/* well, Pvectors are anyway -- in that direction (so a series of large BLAS gscals*/
/* will be faster).                                                                */

int PLA_Local_scal(PLA_Obj alpha, PLA_Obj x)
{
   int original_error = 0;


   int error_value, i;
   int object_type_x;
   int local_length_x, local_width_x, ld_x;
   int ld_x_buf, stride_x, orientation_x;

   int local_length_alpha, local_width_alpha;

   MPI_Datatype datatype_x, datatype_alpha;

   void *x_buffer, *x_buffer_cur, *alpha_buffer;

   PLA_Local_scal_enter(alpha, x);
   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);
   PLA_Obj_local_ldim(x, &ld_x);
   PLA_Obj_local_stride(x, &stride_x);
   PLA_Obj_objtype(x,  &object_type_x);
   if((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
                                         (PLA_PMVECTOR == object_type_x))
      PLA_Obj_project_onto(x, &orientation_x);

   PLA_Obj_local_length(alpha, &local_length_alpha);
   PLA_Obj_local_width (alpha, &local_width_alpha);
   PLA_Obj_datatype(alpha, &datatype_alpha);

   PLA_Obj_local_buffer(x, &x_buffer);
   x_buffer_cur = x_buffer;
   PLA_Obj_local_buffer(alpha, &alpha_buffer);
   if(local_length_x * local_width_x > 0)
   {
      if(((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
          (PLA_PMVECTOR == object_type_x)) && (PLA_PROJ_ONTO_ROW == orientation_x))
      {
         if(MPI_DOUBLE == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_dscal(&local_width_x, (double*)alpha_buffer, (double*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((double*)x_buffer_cur)+stride_x);
            }
         else if(MPI_FLOAT == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_sscal(&local_width_x, (float*)alpha_buffer, (float*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((float*)x_buffer_cur)+stride_x);
            }
         else if(MPI_COMPLEX == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_cscal(&local_width_x, (PLA_COMPLEX*)alpha_buffer,
                         (PLA_COMPLEX*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((PLA_COMPLEX*)x_buffer_cur)+stride_x);
            }
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_zscal(&local_width_x, (PLA_DOUBLE_COMPLEX*)alpha_buffer, (PLA_DOUBLE_COMPLEX*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((PLA_DOUBLE_COMPLEX*)x_buffer_cur)+stride_x);
            }

      }
      else
      {
         if(MPI_DOUBLE == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_dscal(&local_length_x, (double*)alpha_buffer, (double*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((double*)x_buffer_cur)+ld_x);
            }
         else if(MPI_FLOAT == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_sscal(&local_length_x, (float*)alpha_buffer, (float*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((float*)x_buffer_cur)+ld_x);
            }
/* I realize that this MAY not work on the Cray -- not really sure */
         else if(MPI_COMPLEX == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_cscal(&local_length_x, (PLA_COMPLEX*)alpha_buffer,
                         (PLA_COMPLEX*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((PLA_COMPLEX*)x_buffer_cur)+ld_x); /*complex == 2 floats */
            }
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_zscal(&local_length_x, (PLA_DOUBLE_COMPLEX*)alpha_buffer,
                         (PLA_DOUBLE_COMPLEX*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((PLA_DOUBLE_COMPLEX*)x_buffer_cur)+ld_x); /* dcomp == 2 doubles */
            }
      }
   }
   PLA_Local_scal_exit(alpha, x);
   return(PLA_SUCCESS);
}



/******************************************************************************/
/******************************************************************************/


int PLA_Local_inv_scal_enter(PLA_Obj alpha, PLA_Obj x)
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_inv_scal";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( x, "x", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x, template, "x", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal  ( alpha, 1, "alpha", error, original_error );
   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_inv_scal_exit(PLA_Obj alpha, PLA_Obj x)
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

int PLA_Local_inv_scal(PLA_Obj alpha, PLA_Obj x)
{
   int original_error = 0;


   int error_value, i;
   int object_type_x;
   int local_length_x, local_width_x, ld_x;
   int ld_x_buf, stride_x, orientation_x;

   int local_length_alpha, local_width_alpha;

   MPI_Datatype datatype_x, datatype_alpha;

   void *x_buffer, *x_buffer_cur, *alpha_buffer;
   float  f_inv_val;
   double d_inv_val;
   PLA_COMPLEX        c_inv_val;
   PLA_DOUBLE_COMPLEX z_inv_val;


   PLA_Local_inv_scal_enter(alpha, x);
   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);
   PLA_Obj_local_ldim(x, &ld_x);
   PLA_Obj_local_stride(x, &stride_x);
   PLA_Obj_objtype(x,  &object_type_x);
   if((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
                                          (PLA_PMVECTOR == object_type_x))
      PLA_Obj_project_onto(x, &orientation_x);

   PLA_Obj_local_length(alpha, &local_length_alpha);
   PLA_Obj_local_width (alpha, &local_width_alpha);
   PLA_Obj_datatype(alpha, &datatype_alpha);

   PLA_Obj_local_buffer(x, &x_buffer);
   x_buffer_cur = x_buffer;
   PLA_Obj_local_buffer(alpha, &alpha_buffer);
   if(MPI_DOUBLE == datatype_x)
      d_inv_val = 1.0/(*(double*)alpha_buffer);
   else if(MPI_FLOAT  == datatype_x)
      f_inv_val = 1.0/(*(float*) alpha_buffer);
   else if(MPI_COMPLEX == datatype_x)
   {
         c_inv_val.value.val[0] = 1.0/(*(float*)alpha_buffer);
         c_inv_val.value.val[1] = 1.0/(*((float*)alpha_buffer+1));
   }
   else if(MPI_DOUBLE_COMPLEX == datatype_x)
   {
      z_inv_val.real = 1.0/(*(double*)alpha_buffer);
      z_inv_val.imaginary = 1.0/(*((double*)alpha_buffer+1));
   }
   if(local_length_x * local_width_x > 0)
   {
      if(((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
          (PLA_PMVECTOR == object_type_x)) && (PLA_PROJ_ONTO_ROW == orientation_x))
      {
         if(MPI_DOUBLE == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_dscal(&local_width_x, &d_inv_val, (double*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((double*)x_buffer_cur)+stride_x);
            }
         else if(MPI_FLOAT == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_sscal(&local_width_x, &f_inv_val, (float*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((float*)x_buffer_cur)+stride_x);
            }
         else if(MPI_COMPLEX == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_cscal(&local_width_x, &c_inv_val, (PLA_COMPLEX*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((PLA_COMPLEX*)x_buffer_cur)+stride_x);
            }
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            for(i = 0; i < local_length_x; i++)
            {
               PLA_zscal(&local_width_x, &z_inv_val, (PLA_DOUBLE_COMPLEX*)x_buffer_cur, &ld_x);
               x_buffer_cur = (void*)(((PLA_DOUBLE_COMPLEX*)x_buffer_cur)+stride_x);
            }
      }
      else
      {
         if(MPI_DOUBLE == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_dscal(&local_length_x, &d_inv_val, (double*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((double*)x_buffer_cur)+ld_x);
            }
         else if(MPI_FLOAT == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_sscal(&local_length_x, &f_inv_val, (float*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((float*)x_buffer_cur)+ld_x);
            }
         else if(MPI_COMPLEX == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_cscal(&local_length_x, &c_inv_val, (PLA_COMPLEX*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((PLA_COMPLEX*)x_buffer_cur)+ld_x);
            }
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            for(i = 0; i < local_width_x; i++)
            {
               PLA_zscal(&local_length_x, &z_inv_val, (PLA_DOUBLE_COMPLEX*)x_buffer_cur, &stride_x);
               x_buffer_cur = (void*)(((PLA_DOUBLE_COMPLEX*)x_buffer_cur)+ld_x);
            }
      }
   }
   PLA_Local_inv_scal_exit(alpha, x);
   return(PLA_SUCCESS);
}

/******************************************************************************/
/******************************************************************************/


int PLA_Local_axpy_enter(PLA_Obj alpha, PLA_Obj x, PLA_Obj y)
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_axpy";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( x, "x", error, original_error);
      PLA_Obj_object_valid( y, "y", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );
      PLA_Obj_datatype_equal( y, datatype, "y", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x, template, "x", error, original_error );
      PLA_Obj_template_equal( y, template, "y", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal  ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal   ( alpha, 1, "alpha", error, original_error );

      PLA_Obj_local_length_length_equal( x, y, "x and y", error, original_error );
      PLA_Obj_local_width_width_equal  ( x, y, "x and y", error, original_error );
   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_axpy_exit(PLA_Obj alpha, PLA_Obj x, PLA_Obj y)
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

int PLA_Local_axpy(
            PLA_Obj Alpha,
            PLA_Obj X,
            PLA_Obj Y)

{
   int                Llength_X, Lwidth_X, lld_X, Lstride_X;
   int                Llength_Y, Lwidth_Y, lld_Y, Lstride_Y;
   PLA_Obj_obj_type  Object_type_Alpha;
   MPI_Datatype       Datatype_X, Datatype_Y, Datatype_Alpha;
   void               *X_Buffer, *Y_Buffer, *Alpha_Buffer;
   int                PLA_Error = 0, idummy, i;

   PLA_Local_axpy_enter(Alpha, X, Y);
/* Do all inquiries for the objects */
   PLA_Error += PLA_Obj_local_all(X, &Datatype_X, &Llength_X, &Lwidth_X,
                                  &X_Buffer, &lld_X);
   PLA_Error += PLA_Obj_local_stride(X, &Lstride_X);
   PLA_Error += PLA_Obj_local_all(Y, &Datatype_Y, &Llength_Y, &Lwidth_Y,
                                  &Y_Buffer, &lld_Y);
   PLA_Error += PLA_Obj_local_stride(Y, &Lstride_Y);
   PLA_Error += PLA_Obj_local_all(Alpha, &Datatype_Alpha, &idummy, &idummy,
                                  &Alpha_Buffer, &idummy);

   for(i = 0; ( i < Lwidth_X ) && ( Llength_Y * Lwidth_Y > 0 ) ; i++)
   {
      if(Datatype_X == MPI_DOUBLE)
      {
         PLA_daxpy(&Llength_X, (double*)Alpha_Buffer, (double*)X_Buffer,
                  &Lstride_X, (double*)Y_Buffer, &Lstride_Y);
         X_Buffer = (void*) (((double*)X_Buffer) + lld_X);
         Y_Buffer = (void*) (((double*)Y_Buffer) + lld_Y);
      }
      else if(Datatype_X == MPI_FLOAT)
      {
/*baker 8.30.96*/
         PLA_saxpy(&Llength_X, (float*)Alpha_Buffer, (float*)X_Buffer,
                   &Lstride_X, (float*)Y_Buffer, &Lstride_Y);
         X_Buffer = (void*) (((float*)X_Buffer) + lld_X);
         Y_Buffer = (void*) (((float*)Y_Buffer) + lld_Y);
      }
      else if(Datatype_X == MPI_DOUBLE_COMPLEX)
      {
         PLA_zaxpy(&Llength_X, (PLA_DOUBLE_COMPLEX*)Alpha_Buffer, (PLA_DOUBLE_COMPLEX*)X_Buffer,
                   &Lstride_X, (PLA_DOUBLE_COMPLEX*)Y_Buffer, &Lstride_Y);
         X_Buffer = (void*) (((PLA_DOUBLE_COMPLEX*)X_Buffer) + lld_X);
         Y_Buffer = (void*) (((PLA_DOUBLE_COMPLEX*)Y_Buffer) + lld_Y);
      }
      else if(Datatype_X == MPI_COMPLEX)
      {
         PLA_caxpy(&Llength_X, (PLA_COMPLEX*)Alpha_Buffer, (PLA_COMPLEX*)X_Buffer,
                   &Lstride_X, (PLA_COMPLEX*)Y_Buffer, &Lstride_Y);
         X_Buffer = (void*) (((PLA_COMPLEX*)X_Buffer) + lld_X);
         Y_Buffer = (void*) (((PLA_COMPLEX*)Y_Buffer) + lld_Y);
      }
   }
   PLA_Local_axpy_exit(Alpha, X, Y);
   return PLA_Error;
}


/******************************************************************************/
/******************************************************************************/


int PLA_Local_dot_enter( PLA_Obj x, PLA_Obj y, PLA_Obj alpha )
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_dot";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( x, "x", error, original_error);
      PLA_Obj_object_valid( y, "y", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );
      PLA_Obj_datatype_equal( y, datatype, "y", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x, template, "x", error, original_error );
      PLA_Obj_template_equal( y, template, "y", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal  ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal   ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal   ( x, 1, "x", error, original_error);
      PLA_Obj_local_width_equal   ( y, 1, "y", error, original_error);
      PLA_Obj_local_length_length_equal  ( x, y, "x and y", error, original_error );
   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_dot_exit( PLA_Obj x, PLA_Obj y, PLA_Obj alpha )
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

int PLA_Local_dot( PLA_Obj x, PLA_Obj y, PLA_Obj alpha )
{
   int original_error = 0;

   int error_value;
   int object_type_x, object_type_y;
   int local_length_x, local_width_x, ld_x, size_x, incx;
   int ld_x_buf, stride_x, orientation_x;

   int local_length_y, local_width_y, ld_y, size_y, incy;
   int ld_y_buf, stride_y, orientation_y;

   int local_length_alpha, local_width_alpha;

   MPI_Datatype datatype_x, datatype_y, datatype_alpha;

   void *alpha_buffer, *x_buffer, *y_buffer;


/*   PLA_Local_dot_enter( x, y, alpha ); */
   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);
   PLA_Obj_local_ldim(x, &ld_x);
   PLA_Obj_local_stride(x, &stride_x);
   PLA_Obj_objtype(x,  &object_type_x);
   /*if((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
                                         (PLA_PMVECTOR == object_type_x))*/
      PLA_Obj_project_onto(x, &orientation_x);

   PLA_Obj_local_length(y, &local_length_y);
   PLA_Obj_local_width (y, &local_width_y);
   PLA_Obj_datatype(y, &datatype_y);
   PLA_Obj_local_ldim(y, &ld_y);
   PLA_Obj_local_stride(y, &stride_y);
   PLA_Obj_objtype(y,  &object_type_y);
   /*if((PLA_MATRIX == object_type_y) || (PLA_PVECTOR == object_type_y) ||
                                       (PLA_PMVECTOR == object_type_y)) */
      PLA_Obj_project_onto(y, &orientation_y); 

   PLA_Obj_set_to_zero(alpha);
   PLA_Obj_local_buffer(alpha, &alpha_buffer);
   PLA_Obj_local_buffer(x, &x_buffer);
   PLA_Obj_local_buffer(y, &y_buffer);

   if(PLA_PROJ_ONTO_COL == orientation_x)
   {
      size_x = local_length_x;
      incx = stride_x;
   }
   else
   {
      size_x = local_width_x;
      incx = ld_x;
   }

   if(PLA_PROJ_ONTO_COL == orientation_y)
      incy = stride_y;
   else
      incy = ld_y;

   if(local_length_x * local_width_x > 0)
   {
      if(MPI_DOUBLE == datatype_x)
         *(double*)alpha_buffer =
           PLA_ddot(&size_x, (double*)x_buffer, &incx, (double*)y_buffer, &incy);
      else if(MPI_FLOAT == datatype_x)
         *(float*)alpha_buffer =
           PLA_sdot(&size_x, (float*)x_buffer, &incx, (float*)y_buffer, &incy);
/*      else if(MPI_COMPLEX == datatype_x)
         *(PLA_COMPLEX*)alpha_buffer =
           PLA_cdotu(&size_x, (PLA_COMPLEX*)x_buffer, &incx, (PLA_COMPLEX*)y_buffer, &incy);
      else if(MPI_DOUBLE_COMPLEX == datatype_x)
         *(PLA_DOUBLE_COMPLEX*)alpha_buffer =
           PLA_zdotu(&size_x, (PLA_DOUBLE_COMPLEX*)x_buffer, &incx,
                             (PLA_DOUBLE_COMPLEX*)y_buffer, &incy); */
   }
/*    PLA_Local_dot_exit ( x, y, alpha ); */

   return(PLA_SUCCESS);
}
/******************************************************************************/
/******************************************************************************/
int PLA_Local_nrm2_enter( PLA_Obj x, PLA_Obj alpha )
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_nrm2";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( x, "x", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x, template, "x", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal  ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal   ( alpha, 1, "alpha", error, original_error );
/* Special case -- New test length OR width == 1  -- left out for now JAG
      PLA_Obj_local_width_equal   ( x, 1, "x", error, original_error);    */

   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_nrm2_exit( PLA_Obj x, PLA_Obj alpha )
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

int PLA_Local_nrm2( PLA_Obj x, PLA_Obj alpha )
{
   int original_error = 0;

   int error_value;
   int object_type_x;
   int local_length_x, local_width_x, ld_x;
   int ld_x_buf, stride_x, orientation_x;

   int local_length_alpha, local_width_alpha;

   MPI_Datatype datatype_x, datatype_alpha;

   void *alpha_buffer, *x_buffer;

   PLA_Local_nrm2_enter( x, alpha );
   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);
   PLA_Obj_local_ldim(x, &ld_x);
   PLA_Obj_objtype(x,  &object_type_x);
   if((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
                                         (PLA_PMVECTOR == object_type_x))
      PLA_Obj_project_onto(x, &orientation_x);

   PLA_Obj_local_length(alpha, &local_length_alpha);
   PLA_Obj_local_width (alpha, &local_width_alpha);
   PLA_Obj_datatype(alpha, &datatype_alpha);

   PLA_Obj_set_to_zero(alpha);
   PLA_Obj_local_buffer(alpha, &alpha_buffer);
   PLA_Obj_local_buffer(x, &x_buffer);

   if(local_length_x * local_width_x > 0)
   {
      if(((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
          (PLA_PMVECTOR == object_type_x)) && (PLA_PROJ_ONTO_ROW == orientation_x))
      {
         if(MPI_DOUBLE == datatype_x)
            *(double*)alpha_buffer = PLA_dnrm2(&local_width_x, (double*)x_buffer, &ld_x);
         else if(MPI_FLOAT == datatype_x)
            *(float*)alpha_buffer = PLA_snrm2 (&local_width_x, (float*)x_buffer, &ld_x);
/*JAG         else if(MPI_COMPLEX == datatype_x)
            *(PLA_COMPLEX*)alpha_buffer = PLA_cnrm2 (&local_width_x, (PLA_COMPLEX*)x_buffer, &ld_x);
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            *(PLA_DOUBLE_COMPLEX*)alpha_buffer = PLA_znrm2 (&local_width_x, (PLA_DOUBLE_COMPLEX*)x_buffer, &ld_x); */
      }
      else
      {
         if(MPI_DOUBLE == datatype_x)
               *(double*)alpha_buffer = PLA_dnrm2(&local_length_x, (double*)x_buffer, &stride_x);
         else if(MPI_FLOAT == datatype_x)
               *(float*)alpha_buffer = PLA_snrm2 (&local_length_x, (float*)x_buffer, &stride_x);
/* JAG         else if(MPI_COMPLEX == datatype_x)
            *(PLA_COMPLEX*)alpha_buffer = PLA_cnrm2 (&local_length_x, (PLA_COMPLEX*)x_buffer, &stride_x);
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            *(PLA_DOUBLE_COMPLEX*)alpha_buffer = PLA_znrm2 (&local_length_x, (PLA_DOUBLE_COMPLEX*)x_buffer, &stride_x); */
      }
   }
   PLA_Local_nrm2_exit ( x, alpha );
}



/******************************************************************************/
/******************************************************************************/
int PLA_Local_asum_enter( PLA_Obj x, PLA_Obj alpha )
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_asum";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( alpha, "alpha", error, original_error);
      PLA_Obj_object_valid( x, "x", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
      PLA_Obj_datatype( alpha, &datatype );
      PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( alpha, &template );
      PLA_Obj_template_equal( x, template, "x", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal  ( alpha, 1, "alpha", error, original_error );
      PLA_Obj_local_width_equal   ( alpha, 1, "alpha", error, original_error );
/* Special case -- New test length OR width == 1  -- left out for now JAG
      PLA_Obj_local_width_equal   ( x, 1, "x", error, original_error);    */

   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_asum_exit( PLA_Obj x, PLA_Obj alpha )
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

int PLA_Local_asum( PLA_Obj x, PLA_Obj alpha )
{
   int original_error = 0;

   int error_value;
   int object_type_x;
   int local_length_x, local_width_x, ld_x;
   int ld_x_buf, stride_x, orientation_x;

   int local_length_alpha, local_width_alpha;

   MPI_Datatype datatype_x, datatype_alpha;

   void *alpha_buffer, *x_buffer;


   PLA_Local_asum_enter( x, alpha );

   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);
   PLA_Obj_local_ldim(x, &ld_x);
   PLA_Obj_objtype(x,  &object_type_x);
   if((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
                                         (PLA_PMVECTOR == object_type_x))
      PLA_Obj_project_onto(x, &orientation_x);

   PLA_Obj_local_length(alpha, &local_length_alpha);
   PLA_Obj_local_width (alpha, &local_width_alpha);
   PLA_Obj_datatype(alpha, &datatype_alpha);

   PLA_Obj_set_to_zero(alpha);
   PLA_Obj_local_buffer(alpha, &alpha_buffer);
   PLA_Obj_local_buffer(x, &x_buffer);

   if(local_length_x * local_width_x > 0)
   {
      if(((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
          (PLA_PMVECTOR == object_type_x)) && (PLA_PROJ_ONTO_ROW == orientation_x))
      {
         if(MPI_DOUBLE == datatype_x)
            *(double*)alpha_buffer = PLA_dasum(&local_width_x, (double*)x_buffer, &ld_x);
         else if(MPI_FLOAT == datatype_x)
            *(float*)alpha_buffer = PLA_sasum (&local_width_x, (float*)x_buffer, &ld_x);
/*JAG         else if(MPI_COMPLEX == datatype_x)
            *(PLA_COMPLEX*)alpha_buffer = PLA_casum (&local_width_x, (PLA_COMPLEX*)x_buffer, &ld_x);
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            *(PLA_DOUBLE_COMPLEX*)alpha_buffer = PLA_zasum (&local_width_x, (PLA_DOUBLE_COMPLEX*)x_buffer, &ld_x); */
      } 
      else
      {
         if(MPI_DOUBLE == datatype_x)
               *(double*)alpha_buffer = PLA_dasum(&local_length_x, (double*)x_buffer, &stride_x);
         else if(MPI_FLOAT == datatype_x)
               *(float*)alpha_buffer = PLA_sasum (&local_length_x, (float*)x_buffer, &stride_x);
/*JAG         else if(MPI_COMPLEX == datatype_x)
            *(PLA_COMPLEX*)alpha_buffer = PLA_casum (&local_length_x, (PLA_COMPLEX*)x_buffer, &stride_x);
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
            *(PLA_DOUBLE_COMPLEX*)alpha_buffer = PLA_zasum (&local_length_x, (PLA_DOUBLE_COMPLEX*)x_buffer, &stride_x); */
      }
   }
   PLA_Local_asum_exit ( x, alpha );
   return(PLA_SUCCESS);
}



/******************************************************************************/
/******************************************************************************/
int PLA_Local_iamax_enter( PLA_Obj x, PLA_Obj xmax, PLA_Obj k )
{
   int check_parameter, instrument;
   char current_routine[20] = "PLA_Local_iamax";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {
   /* check objects for validity */
      PLA_Obj_object_valid( x, "x", error, original_error);
      PLA_Obj_object_valid( xmax, "xmax", error, original_error);
      PLA_Obj_object_valid( k, "k", error, original_error);

    /* check datatypes */
      PLA_Obj_datatype_valid( x, "x", error, original_error );
      PLA_Obj_datatype( xmax, &datatype );
      PLA_Obj_datatype_equal( xmax, datatype, "xmax", error, original_error );
      PLA_Obj_datatype_equal( k,    MPI_INT,  "k",    error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( x, &template );
      PLA_Obj_template_equal( xmax, template, "xmax", error, original_error );
      PLA_Obj_template_equal( k,    template, "k", error, original_error );

    /* check dimensions */
      PLA_Obj_local_length_equal  ( xmax, 1, "xmax", error, original_error );
      PLA_Obj_local_width_equal   ( xmax, 1, "xmax", error, original_error );
      PLA_Obj_local_length_equal  ( k,    1, "k", error, original_error );
      PLA_Obj_local_width_equal   ( k,    1, "k", error, original_error );
/* Special case -- New test length OR width == 1  -- left out for now JAG
      PLA_Obj_local_width_equal   ( x, 1, "x", error, original_error);    */

   }
   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   return(original_error);
}

int PLA_Local_iamax_exit( PLA_Obj x, PLA_Obj xmax, PLA_Obj k )
{
   int instrument;
   char current_routine[20];

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   return(PLA_SUCCESS);
}

int PLA_Local_iamax( PLA_Obj x, PLA_Obj xmax, PLA_Obj k )
{
   int original_error = 0;
   int error_value, object_type_x;
   int local_length_x, local_width_x, ld_x;
   int ld_x_buf, stride_x, orientation_x;



   int local_length_xmax, local_width_xmax;

   MPI_Datatype datatype_x, datatype_xmax, datatype_k;

   void *xmax_buffer, *x_buffer, *k_buffer;

   PLA_Local_iamax_enter( x, xmax, k );
   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width (x, &local_width_x);
   PLA_Obj_datatype(x, &datatype_x);
   PLA_Obj_local_ldim(x, &ld_x);
   PLA_Obj_objtype(x,  &object_type_x);
   if((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
                                         (PLA_PMVECTOR == object_type_x))
      PLA_Obj_project_onto(x, &orientation_x);

   PLA_Obj_local_length(xmax, &local_length_xmax);
   PLA_Obj_local_width (xmax, &local_width_xmax);
   PLA_Obj_datatype(xmax, &datatype_xmax);

   PLA_Obj_datatype(k, &datatype_k);

   PLA_Obj_set_to_zero(xmax);
   PLA_Obj_local_buffer(xmax, &xmax_buffer);
   PLA_Obj_local_buffer(k, &k_buffer);
   PLA_Obj_local_buffer(x, &x_buffer);
   *(int*)k_buffer = -1; /* I don't think that the set_all_minus_one does ints */

   if(local_length_x * local_width_x > 0)
   {
      if(((PLA_MATRIX == object_type_x) || (PLA_PVECTOR == object_type_x) ||
          (PLA_PMVECTOR == object_type_x)) && (PLA_PROJ_ONTO_ROW == orientation_x))
      {
         if(MPI_DOUBLE == datatype_x)
         {
            *(int*)k_buffer = PLA_idamax(&local_width_x, (double*)x_buffer, &ld_x);
            *(double*)xmax_buffer = * (((double*)x_buffer)+((*(int*)k_buffer)-1)*ld_x);
         }
         else if(MPI_FLOAT == datatype_x)
         {
            *(int*)k_buffer = PLA_isamax (&local_width_x, (float*)x_buffer, &ld_x);
            *(float*)xmax_buffer = * (((float*)x_buffer)+((*(int*)k_buffer)-1)*ld_x);
         }
/*JAG          else if(MPI_COMPLEX == datatype_x)
         {
            *(int*)k_buffer = PLA_icamax (&local_width_x, (PLA_COMPLEX*)x_buffer, &ld_x);
            *(PLA_COMPLEX*)xmax_buffer = * (((PLA_COMPLEX*)x_buffer)+((*(int*)k_buffer)-1)*ld_x);
         }
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
         {
            *(int*)k_buffer = PLA_icamax (&local_width_x, (PLA_COMPLEX*)x_buffer, &ld_x);
            *(PLA_DOUBLE_COMPLEX*)xmax_buffer = * (((PLA_DOUBLE_COMPLEX*)x_buffer)+((*(int*)k_buffer)-1)*ld_x);
         } */
      }
      else
      {
         if(MPI_DOUBLE == datatype_x)
         {
            *(int*)k_buffer = PLA_idamax(&local_length_x, (double*)x_buffer, &stride_x);
            *(double*)xmax_buffer = * (((double*)x_buffer)+((*(int*)k_buffer)-1)*stride_x);
         }
         else if(MPI_FLOAT == datatype_x)
         {
            *(int*)k_buffer = PLA_isamax (&local_length_x, (float*)x_buffer, &stride_x);
            *(float*)xmax_buffer = * (((float*)x_buffer)+((*(int*)k_buffer)-1)*stride_x);
         }
/* JAG        else if(MPI_COMPLEX == datatype_x)
         {
            *(int*)k_buffer = PLA_icamax (&local_width_x, (PLA_COMPLEX*)x_buffer, &stride_x);
            *(PLA_COMPLEX*)xmax_buffer = * (((PLA_COMPLEX*)x_buffer)+((*(int*)k_buffer)-1)*stride_x);
         }
         else if(MPI_DOUBLE_COMPLEX == datatype_x)
         {
            *(int*)k_buffer = PLA_icamax (&local_width_x, (PLA_COMPLEX*)x_buffer, &stride_x);
            *(PLA_DOUBLE_COMPLEX*)xmax_buffer = * (((PLA_DOUBLE_COMPLEX*)x_buffer)+((*(int*)k_buffer)-1)*stride_x);
         } */
      }
   }
   PLA_Local_iamax_exit ( x, xmax, k );
   return(PLA_SUCCESS);
}


