/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "mmpi.h"

/* ------------------------------------------------------------------ */

#define DEFAULT_BASE_NAME	"MMPI.LogP"

static char MMPI_Log_base_name[256] = { '\0' };

/* ------------------------------------------------------------------ */

int MMPI_Log_init( const char * const base_name )
{
  if ( MMPI_Log_base_name[0] ) return MPI_ERR_UNKNOWN ;
  strcpy( MMPI_Log_base_name , base_name );
  return MPI_SUCCESS ;
}

int MMPI_Log_file_d(void)
{
  static int Log_File_d = -1 ;

  if ( Log_File_d == -1 ) {
    char buf[16];
    char buf2[16] ;

    int PNum , PMe , R , i , d ;

    if ( MPI_SUCCESS == ( R = MPI_Comm_size( MPI_COMM_WORLD , &PNum ) ) &&
         MPI_SUCCESS == ( R = MPI_Comm_rank( MPI_COMM_WORLD , &PMe  ) ) ) {

      for ( i = 10 , d = 1 ; i < PNum ; i *= 10 , d++ );

      sprintf(buf2,"%%0%dd",d);
      sprintf(buf,buf2,PMe);

      if ( ! MMPI_Log_base_name[0] )
        strcpy( MMPI_Log_base_name , DEFAULT_BASE_NAME );

      strcat( MMPI_Log_base_name , buf );

      Log_File_d = creat( MMPI_Log_base_name, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
    }
  }
  return Log_File_d ;
}

/* ------------------------------------------------------------------ */

FILE * MMPI_Log_file(void)
{
  static FILE * Log_File = 0 ;

  if ( NULL == Log_File ) {

    int fd = MMPI_Log_file_d();

    if ( -1 == fd || NULL == ( Log_File = fdopen( fd , "a" ) ) )
      return Log_File = stderr ;
  }

  return Log_File ;
}

/* ------------------------------------------------------------------ */

void MMPI_Log_abort(
  const char * const who ,
  const char * const msg ,
  const int mpi_why )
{
  char mpi_msg[ MPI_MAX_ERROR_STRING ];
  int mpi_msg_len ;

  MPI_Error_string( mpi_why, mpi_msg, &mpi_msg_len );

  mpi_msg[mpi_msg_len] = '\0' ;

  fprintf(MMPI_Log_file(),"\n(MMPI_Log_abort) %s %s : %d %s\n",
    who,msg,mpi_why,mpi_msg);

  MPI_Abort( MPI_COMM_WORLD , mpi_why );

  exit(-1); /* If MPI_Abort doesn't kill, exit */
}

void MMPI_Log_message(
  const char * const who ,
  const char * const msg )
{
  fprintf(MMPI_Log_file(),"%s %s\n",who,msg);
  fflush(MMPI_Log_file());
}


