
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/
#include "PLA.h"
/* SHOULD the buffer size be a long???? */
int PLA_Matrix_buf_size(MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align_row,
		      int               global_align_col,
		      int              *bytes_required)
{
   int local_length, local_width, data_size;
   int zero_or_one;

   PLA_Temp_zero_or_one (template, &zero_or_one);
   if (PLA_ALIGN_FIRST == global_align_row) global_align_row = zero_or_one;
   if (PLA_ALIGN_FIRST == global_align_col) global_align_col = zero_or_one;
   MPI_Type_size(datatype, &data_size);
   PLA_Temp_matrix_local_length_width(template, 
                                      global_length,
                                      global_width,
                                      global_align_row,
                                      global_align_col,
                                      &local_length,
                                      &local_width);
   *bytes_required = local_length * local_width * data_size;
   return(PLA_SUCCESS);
}

/* Buffer flags
1 : A buffer is passed for use but the object does not exist
0 : A buffer is not passed and one should be malloced
-1: A buffer is not passed and one should NOT be malloced
*/

int PLA_Matrix_create (MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align_row,
		      int               global_align_col,
		      PLA_Obj          *new_matrix)
{
   void *data_buffer = NULL;
   int  buffer_flag =  0; /* indicates that no buffer is passed but one should be made */
   int  buffer_size =  0; /* unimportant in this case -- may be "rolled into" the flag */
   int  retval;
   retval = PLA_Matrix_create_base(datatype, global_length,global_width,template,
	                             global_align_row, global_align_col,
								 buffer_flag, buffer_size, data_buffer,
								 new_matrix);
   /* One of my new fields */
   (*new_matrix)->allocation_type = PLA_HEAP;       /* vs. PLA_USER_BUFFER */
   (*new_matrix)->has_data_space  = PLA_HAS_SPACE;  /* vs. PLA_NO_SPACE */
   return(retval);
}

int PLA_Matrix_create_buf (MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align_row,
		      int               global_align_col,
		      int               buffer_size,
		      void             *data_buffer,
		      PLA_Obj          *new_matrix)
{
   int  buffer_flag = 1; /* indicates that a buffer is passed for use */
   int  retval;
   retval = PLA_Matrix_create_base(datatype, global_length,global_width,template,
	                             global_align_row, global_align_col,
								 buffer_flag, buffer_size, data_buffer,
								 new_matrix);
   (*new_matrix)->allocation_type = PLA_USER_BUFFER;    /* vs. PLA_HEAP */
   (*new_matrix)->has_data_space  = PLA_HAS_SPACE;      /* vs. PLA_NO_SPACE */
   (*new_matrix)->length_master_buffer_byte = buffer_size;    /* overwrite what _base did */
   return(retval);
}

int PLA_Matrix_create_no_buf (MPI_Datatype      datatype,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               global_align_row,
		      int               global_align_col,
		      PLA_Obj          *new_matrix)
{
   int  buffer_flag = -1; /* indicates that a buffer is passed for use */
   void *data_buffer = NULL;
   int  buffer_size =  0; /* unimportant in this case -- may be "rolled into" the flag */
   int  retval;
   retval = PLA_Matrix_create_base(datatype, global_length,global_width,template,
	                             global_align_row, global_align_col,
								 buffer_flag, buffer_size, data_buffer,
								 new_matrix);
   (*new_matrix)->allocation_type = PLA_UNDEFINED;    /* not defined until space attached  */
   (*new_matrix)->has_data_space  = PLA_NO_SPACE;    /* vs. PLA_HAS_SPACE */
   (*new_matrix)->length_master_buffer_byte = -1;    /* overwrite what _base did */
   return(retval);
}

int PLA_Matrix_attach_buf(PLA_Obj matrix, int buffer_size, void *data_buffer)
{
   int data_size, buf_size, local_buffer_size_required;
   MPI_Datatype datatype;

   PLA_Obj_datatype(matrix, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = matrix->local_length * matrix->local_width;
   local_buffer_size_required = data_size * buf_size;
   if(buffer_size < local_buffer_size_required)
   {
      printf("PLA_Matrix_attach_buf was called with an attach buffer that was too small\n");
      return(1);
   }
   else matrix->master_buffer   = data_buffer;
   matrix->allocation_type = PLA_USER_SUPPLIED;
   matrix->has_data_space  = PLA_HAS_SPACE;
   matrix->length_master_buffer_byte = buffer_size;    /* overwrite what _base did */
   return(PLA_SUCCESS);
}

int PLA_Matrix_attach(PLA_Obj matrix)
{
   /* Need to modularize the matrix create routine */
   /* This routine allocs buffer space for a passed in object -- actually,
      I am considering making them all just call PLA_Object_attach
      because they all do the same thing */
   int data_size, buf_size;
   MPI_Datatype datatype;

   PLA_Obj_datatype(matrix, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = matrix->local_length * matrix->local_width;
   if (0 != buf_size)
   {
      MPI_Type_size (datatype, &data_size);
      matrix->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
      if (matrix->master_buffer == NULL)
      {
         printf (" Could not allocate buffer of %d bytes in PLA_Matrix_attach\n",
	      buf_size*(int)data_size);
         return(1);
      }
   }
   else matrix->master_buffer = NULL;

   matrix->allocation_type    = PLA_HEAP;
   matrix->has_data_space     = PLA_HAS_SPACE;
   return(PLA_SUCCESS);
}

int PLA_Matrix_create_base (MPI_Datatype      datatype,
		      int              global_length,
		      int              global_width,
		      PLA_Template     template,
		      int              global_align_row,
		      int              global_align_col,
			   int              buffer_flag,
			   int             external_buffer_size,
			   void            *external_buffer,
		      PLA_Obj         *new_matrix)
{

  PLA_Obj       matrix = NULL;
  int           buf_size;
  int           error=0;
  int           data_size = 0;
  int           zero_or_one;
  int           index; /* for getting a free Obj_struct */
  int           local_buffer_size_required = -1;

  if(! PLA_Initialized_object_structs())
     PLA_Init_object_structs(PLA_DEFAULT_OBJ_COUNT); /* some reasonable number of objects */

  if (NULL != *new_matrix) error+=PLA_Obj_free(new_matrix);

  index = PLA_Get_free_object_struct();

  if(-1 == index)
  {
     printf("Could not get an allocated object structure\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
/*printf("Matrix : Index is %5d and place is %ld\n", index, &PLA_GLOBAL_OBJ_STRUCT[index]); */
  matrix = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);

  error += PLA_Temp_zero_or_one (template, &zero_or_one);
  if (PLA_ALIGN_FIRST == global_align_row) global_align_row = zero_or_one;
  if (PLA_ALIGN_FIRST == global_align_col) global_align_col = zero_or_one;

  error += PLA_Obj_init (matrix);
  matrix->obj_type         = PLA_MATRIX;
  matrix->datatype         = datatype;
  matrix->global_length    = matrix->master_length    = global_length;
  matrix->global_width     = matrix->master_width     = global_width;
  matrix->local_stride     = 1;
  matrix->template         = template;
  matrix->global_align_row = matrix->master_align_row = global_align_row;
  matrix->global_align_col = matrix->master_align_col = global_align_col;

  /*
     Determine the local length;
     */
  error += PLA_Temp_matrix_local_length_width (template,
					      global_length,
					      global_width,
					      global_align_row,
					      global_align_col,
					      &(matrix->local_length),
					      &(matrix->local_width));
  matrix->local_lld  = matrix->local_length;
  /*
     Deal with buffer allocation :
     Note : buf_size will still appear to be "tight fit" for error checking
            purposes, despite the fact that an external buffer may be supplied.
            Also, if the object is to be created without a buffer, this will
            have a negative value to catch errors.
  */

  if(-1 == buffer_flag) /* Create an object with NO!!! buffer */
     buf_size = -1;
  else buf_size = matrix->local_length * matrix->local_width;

  if(0 == buffer_flag)
  {
     if (0 != buf_size)
     {
        MPI_Type_size (datatype, &data_size);
        matrix->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
        if (matrix->master_buffer == NULL)
        {
           printf ("Could not allocate buffer of %d bytes in PLA_Matrix_create\n",
	                 buf_size*(int)data_size);
           return(1);
        }
     }
     else
        matrix->master_buffer = NULL;
  }
  else if(1 == buffer_flag)
  {
     MPI_Type_size (datatype, &data_size);
     local_buffer_size_required = data_size * buf_size;
     if(external_buffer_size < local_buffer_size_required)
     {
        printf("PLA_Matrix_create_buf was called with an attached buffer that was too small\n");
        return(1);
     }
     matrix->master_buffer = external_buffer;
  }
  matrix->length_master_buffer_byte = buf_size*(int)data_size;
  matrix->buffer = matrix->master_buffer;
/* The views will also have to be taken from some set-aside space */
  index = PLA_Get_free_views_struct();
  if(-1 == index)
  {
     printf("Could not get an allocated views structure\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
  matrix->views  = (int *) (&PLA_GLOBAL_VIEWS_STRUCT[index]);
  (*(matrix->views))= 1;

  /* Set the first and last block size in both dimentions */
  error += PLA_Temp_matrix_first_row_block   (template,
					     global_length,
					     global_align_row,
					     &(matrix->owner_top_block),
					     &(matrix->length_top_block));
  error += PLA_Temp_matrix_last_row_block    (template,
					     global_length,
					     global_align_row,
					     &(matrix->owner_bottom_block),
					     &(matrix->length_bottom_block));
  error += PLA_Temp_matrix_first_col_block   (template,
					     global_width,
					     global_align_col,
					     &(matrix->owner_left_block),
					     &(matrix->width_left_block));
  error += PLA_Temp_matrix_last_col_block    (template,
					     global_width,
					     global_align_col,
					     &(matrix->owner_right_block),
					     &(matrix->width_right_block));
  *new_matrix = matrix;
  return(error);
}






int PLA_Matrix_submatrix (PLA_Obj old_matrix, 
			 int      global_length, 
			 int      global_width,
			 int      matrix_align_row,
			 int      matrix_align_col,
			 PLA_Obj *new_matrix)
{
  
  
  PLA_Obj      matrix;
  PLA_Template   template;
  int           global_align_row, global_align_col; 
  int           local_length, local_width;
  int           local_align_row, local_align_col, offset;
  int           error=0;
  PLA_Start_num  zero_or_one;
  int           data_size; 
  int           index; 

  if ( old_matrix->obj_type != PLA_MATRIX) {
    printf ("Invalid object passed to PLA_Matrix_submatrix: NonMatrix\n");
    return (1);
  }  
  if (*new_matrix != NULL) error+=PLA_Obj_free(new_matrix);

  template = old_matrix->template;
  error += PLA_Temp_zero_or_one (template, &zero_or_one);

  if (old_matrix == *new_matrix) {
  /* Well now things are tough.  They really want to just shift the view
     not take a submatrix.  So they should have called view_shift instead
     of submatrix.  But we are a forgiving programmer and will call
     view_shift for them. */
    int   
      length_top,
      width_left,
      width_right,
      length_bottom;
    if (PLA_DIM_ALL == global_length) global_length = old_matrix->global_length;
    if (PLA_DIM_ALL == global_width ) global_width  = old_matrix->global_width;
    if (PLA_ALIGN_FIRST == matrix_align_row) matrix_align_row = zero_or_one;
    if (PLA_ALIGN_FIRST == matrix_align_col) matrix_align_col = zero_or_one;

    length_top    = matrix_align_row - zero_or_one;
    width_left    = matrix_align_col - zero_or_one;
    width_right   = matrix->global_width - global_width - width_left;
    length_bottom = matrix->global_length- global_length- length_top; 
    error+= PLA_Matrix_view_shift (old_matrix, 
                                   length_top, 
                                   width_left,
                                   width_right,
                                   length_bottom);
    return(error);
  }

  if (NULL != *new_matrix ) 
    matrix = *new_matrix;
  else {
    index = PLA_Get_free_object_struct();
    if(-1 == index)
    {
       printf("Could not allocate obj_struct in matrix_submatrix\n");
       return(1);
    }
    matrix = (PLA_Obj) (& PLA_GLOBAL_OBJ_STRUCT[index]);
/* printf("Sub Matrix : Index is %5d and place is %ld\n", index, &PLA_GLOBAL_OBJ_STRUCT[index]); */
  matrix = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);
  }
  

  /* 
     By default copy all the fields in one blow.  Then modify
     the ones that need modifying.
     */
  *matrix = *old_matrix;

  if (global_length == PLA_DIM_ALL) global_length = matrix->global_length;
  else                             matrix->global_length = global_length;
  if (global_width  == PLA_DIM_ALL) global_width  = matrix->global_width;
  else                             matrix->global_width  = global_width;
  
  if (matrix_align_row == PLA_ALIGN_FIRST) matrix_align_row = zero_or_one;
  if (matrix_align_col == PLA_ALIGN_FIRST) matrix_align_col = zero_or_one;

  global_align_row        = 
    matrix->global_align_row + matrix_align_row - (int)zero_or_one;
  matrix->global_align_row = global_align_row;
  
  global_align_col        = 
    matrix->global_align_col + matrix_align_col - (int)zero_or_one;

  matrix->global_align_col = global_align_col;

  error += PLA_Temp_matrix_local_length_width (template,
					      global_length, 
					      global_width, 
					      global_align_row, 
					      global_align_col, 
					      &local_length, 
					      &local_width);
  matrix->local_length = local_length;
  matrix->local_width  = local_width;

  /* Set the first and last block size in both dimentions */
  
  error += PLA_Temp_matrix_first_row_block   (template,
					     global_length,
					     global_align_row,
					     &(matrix->owner_top_block),
					     &(matrix->length_top_block));
  error += PLA_Temp_matrix_last_row_block    (template,
					     global_length,
					     global_align_row,
					     &(matrix->owner_bottom_block),
					     &(matrix->length_bottom_block));
  error += PLA_Temp_matrix_first_col_block   (template,
					     global_width,
					     global_align_col,
					     &(matrix->owner_left_block),
					     &(matrix->width_left_block));
  error += PLA_Temp_matrix_last_col_block    (template,
					     global_width,
					     global_align_col,
					     &(matrix->owner_right_block),
					     &(matrix->width_right_block));
  /*
     Determine how much of the local matrix to skip at the beginning.
     Look at it this way:
     
     Global Template Distrobution
     --------------------------------------------------------------
     |        ^                                                   |     
     |        |                                                   |     
     |       C|                                                   |     
     |        |                                                   |     
     |   D    v         Origional Master Matrix                   |     
     |<------>=============================================       |     
     |        I           ^                               I       |     
     |        I           |                               I       |     
     |        I           |                               I       |     
     |        I          A|                               I       |     
     |        I           |                               I       |     
     |        I           |                               I       |     
     |        I     B     v          Submatrix            I       |     
     |        I<--------->============================    I       |     
     |        I           I                          I    I       |     
     |        I           I                          I    I       |     
     |        I           I                          I    I       |     
     |        I           I                          I    I       |     
     |        I           I                          I    I       |     
     |        I           I                          I    I       |     
     |        I           I                          I    I       |     
     |        I           ============================    I       |     
     |        I                                           I       |     
     |        I                                           I       |     
     |        =============================================       |     
     |                                                            |     
     --------------------------------------------------------------
     
     So I need to determine the local part of A and B given the global
     size of A and B which are matrix_align_row and matrix_align_col
     respectively.   C and D are the origional offsets
     of the master matrix within the template.  So If I ask to compute
     the local size of a matrix of global size AxB with offsets into
     the template of C and D, I should wind up with how much to skip.
     
     Don't get confused by the variable names in the local_length_width
     call, they have nothing to do with how the subroutine interprets
     them.  Matrix_align_row, col will be treated like the dimentions
     of a global matrix.  local_align_row, col are the return dimentions
     of the local matrix, they are not based on zero_or_one.
     
     There are some tricky things.  Like matrix_align_row is an offset
     so it is measured based on zero_or_one, as is matrix_align_col.  
     To make things simpler convert each to a size.  Hence each will
     be starting from 1.  
     
     Since matrix_align_row and matrix_align_col start from 1, we have
     to decrement the local_align_row and local_align_col on the way out.
     
     Note that this also works for submatrix of a submatrix.
     
     
     */

  if (local_length != 0 &&
      local_width  != 0) {

    matrix_align_row -= (int) zero_or_one;
    matrix_align_col -= (int) zero_or_one;

    PLA_Temp_matrix_local_length_width (template,
				       matrix_align_row, 
				       matrix_align_col,
				       old_matrix->global_align_row,
				       old_matrix->global_align_col,
				       &local_align_row, 
				       &local_align_col);

    /* 
       Now have to compute the offset into the buffer.  We want to find
       row local_align and column align_col on a buffer of leading 
       dimention Matrix->lld.
       */
    offset =  local_align_col * matrix->local_lld + local_align_row;
  
    MPI_Type_size (matrix->datatype, &data_size);
    offset *= (int) data_size;

    matrix->buffer = (void *) ( (char *)(matrix->buffer) + offset );
  }
  else
    matrix->buffer = NULL;

  (*(matrix->views))++;
  *new_matrix = matrix;

  return(error);
}



int PLA_Matrix_view_shift (PLA_Obj matrix, 
			  int      length_top, 
			  int      width_left,
			  int      width_right,
			  int      length_bottom)
{
  PLA_Template   template;
  int           global_align_row, global_align_col; 
  int           global_length, global_width;
  int           local_length, local_width;
  int           local_align_row, local_align_col, offset;
  int           error=0;
  int           data_size; 
  

  if (matrix->obj_type != PLA_MATRIX) {
    printf ("Invalid object passed to PLA_Matrix_submatrix: NonMatrix\n");
    return (1);
  }  

  template = matrix->template;

  global_align_row  = matrix->global_align_row + length_top;
  global_align_col  = matrix->global_align_col + width_left;
  global_length     = matrix->global_length + length_bottom-length_top;
  global_width      = matrix->global_width  + width_right-width_left;

  matrix->global_length = global_length;
  matrix->global_width  = global_width;

  if (matrix->master_align_row > global_align_row || 
      matrix->master_align_col > global_align_col) {
    printf 
      ("Matrix_view_shift shifted before the beginning of the matrix\n");
    printf 
      ("Origional alignment (%d,%d) plus offset (%d,%d)\n",
       matrix->global_align_row,matrix->global_align_col,
       length_top, width_left);
    printf
      ("gives offset (%d,%d)\n",
       global_align_row, global_align_col);
    printf
      ("Origional creation alignment is (%d,%d)\n",
       matrix->master_align_row, matrix->master_align_col);
  }
  if (0>global_length || 0>global_width) {
    printf
      ("Matrix_view_shift reduced view to less than zero\n");
    printf 
      ("Origional size (%d,%d) less top/left offset (%d,%d)\n",
       matrix->global_length,matrix->global_width,
       length_top,width_left);
    printf
      ("plus bottom/right offset (%d,%d) gives negative size (%d,%d)\n",
       length_bottom, width_right, global_length, global_width);
  }
  if (global_align_row+global_length > 
         matrix->master_align_row+matrix->master_length   ||
      global_align_col+global_width  >
         matrix->master_align_col+matrix->master_width) {
    printf
      ("Matrix_view_shift shifted past the end of the matrix\n");
    printf
      ("New alignment (%d,%d) plus length (%d,%d)\n",
       global_align_row,global_align_col,
       global_length, global_width);
    printf
      ("Exceeds origional alignment plus length (%d,%d),(%d,%d)\n",
       matrix->master_align_row, matrix->master_align_col,
       matrix->master_length, matrix->master_width);
  }




  error += PLA_Temp_matrix_local_length_width (template,
					      global_length, 
					      global_width, 
					      global_align_row, 
					      global_align_col, 
					      &local_length, 
					      &local_width);
  matrix->local_length = local_length;
  matrix->local_width  = local_width;

  /* Set the first and last block size in both dimentions */
  
  error += PLA_Temp_matrix_first_row_block   (template,
					     global_length,
					     global_align_row,
					     &(matrix->owner_top_block),
					     &(matrix->length_top_block));
  error += PLA_Temp_matrix_last_row_block    (template,
					     global_length,
					     global_align_row,
					     &(matrix->owner_bottom_block),
					     &(matrix->length_bottom_block));
  error += PLA_Temp_matrix_first_col_block   (template,
					     global_width,
					     global_align_col,
					     &(matrix->owner_left_block),
					     &(matrix->width_left_block));
  error += PLA_Temp_matrix_last_col_block    (template,
					     global_width,
					     global_align_col,
					     &(matrix->owner_right_block),
					     &(matrix->width_right_block));
  /*
     Determine how much of the local matrix to skip at the beginning.
     On a submatrix things just get smaller untill they are eventually
     zero.  But on a view shift things can get larger and can 
     go from zero length to the whole matrix. 
     So this is different from the submatrix code.  The master_align_row
     and master_align_col were added to the OBJ just to be able
     to compute the alignment for the views.
     */

  if (local_length != 0 &&
      local_width  != 0) {
    
    PLA_Temp_matrix_local_length_width (template,
				       global_align_row -
				       matrix->master_align_row, 
				       global_align_col -
				       matrix->master_align_col,
				       matrix->master_align_row,
				       matrix->master_align_col,
				       &local_align_row, 
				       &local_align_col);
    /* 
       Now have to compute the offset into the buffer.  We want to find
       row local_align and column align_col on a buffer of leading 
       dimention Matrix->lld.
       */
    offset =  local_align_col * matrix->local_lld + local_align_row;
  
    MPI_Type_size (matrix->datatype, &data_size);
    offset *= (int) data_size;

    matrix->buffer = (void *) ( (char *)(matrix->master_buffer) + offset );

    if (matrix->buffer < matrix->master_buffer) {
      printf ("Error in PLA_Matrix_view_shift.\n");
      printf ("Computed beginning of buffer, %p, is less than\n",
	      matrix->buffer);
      printf ("actual beginning of buffer, %p\n", 
	      matrix->master_buffer);
    }
  }
  else
    matrix->buffer = NULL;

  matrix->global_align_row = global_align_row;
  matrix->global_align_col = global_align_col;

  return(error);
}





int PLA_Matrix_local_all (PLA_Obj          matrix,
			 MPI_Datatype *    datatype,
			 int *             local_length,
			 int *             local_width,
			 void **           buffer,
			 int  *            local_lld)
{
  int error=0;

  error += PLA_Obj_datatype       (matrix,  datatype);
  error += PLA_Obj_local_length   (matrix,  local_length);
  error += PLA_Obj_local_width    (matrix,  local_width);
  error += PLA_Obj_local_buffer         (matrix,  buffer); 
  error += PLA_Obj_local_ldim      (matrix,  local_lld); 
  return (error);
}


int PLA_Matrix_global_all (PLA_Obj          matrix,
			  int *             global_length,
			  int *             global_width,
			  int  *            global_align_row,
			  int  *            global_align_col,
			  PLA_Template *     template)
     
{
  int error=0;

  error += PLA_Obj_global_length            (matrix,  global_length);
  error += PLA_Obj_global_width             (matrix,  global_width);
  error += PLA_Obj_global_align_row  (matrix,  global_align_row); 
  error += PLA_Obj_global_align_col  (matrix,  global_align_col); 
  error += PLA_Obj_template          (matrix,  template); 
  return (error);
}

  
     
