
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/



#include <stdio.h>
#include <mpi.h>
#include "PLA.h"
#include "PLA_Obj_prefix_mult.h"

/* note : to change reductions, change lines marked by --->operation<--- */

int PLA_Vector_prefix_mult ( PLA_Obj vector )
{
   int
      i, j, k, l,
      nb,
      my_pe,
      n_pes,
      global_align,
      global_length,
      local_length,
      local_stride,
      nblocks;

   double
    * buffer,
    * bufin,
    * bufout;

   MPI_Datatype
      datatype;

   MPI_Comm
      comm;

   PLA_Template
      template;

   PLA_Vector_local_all ( vector, & datatype, & local_length, ( void ** ) & buffer, & local_stride );
   PLA_Vector_global_all ( vector, & global_length, & global_align, & template );
   PLA_Temp_nb ( template, & nb );

   nblocks = global_length / nb;
   if ( global_length % nb ) nblocks++;

   bufin = ( double * ) calloc ( nblocks, sizeof ( double ) );
   bufout = ( double * ) calloc ( nblocks, sizeof ( double ) );

   PLA_Temp_comm1D ( template, & comm );
   PLA_Temp_total_num_proc ( template, & n_pes );
   PLA_Temp_me1D ( template, & my_pe );

   k = 0;
   for ( i = 1 ; i < local_length ; i++ )
   {
      if ( i % nb == 0 )
      { 
         bufin [ my_pe + k * n_pes ] = buffer [ i - 1 ];
         k++;
      }
      else
         buffer [ i ] *= buffer [ i - 1 ]; /*--->operation<---*/
   }
   bufin [ my_pe + k * n_pes ] = buffer [ local_length - 1 ];

   /* this allreduce collects the contribution of all distributed blocks
      into a single array - no reduction operation are performed here!!! */

   MPI_Allreduce ( bufin, bufout, nblocks, MPI_DOUBLE, MPI_SUM, comm );

   for ( i = 1 ; i < nblocks ; i++ )
      bufout [ i ] *= bufout [ i - 1 ]; /*--->operation<---*/

   k = 0;
   for ( i = 0 ; i < local_length ; i++ )
   {
      if ( i % nb == 0 )
      {
         l = my_pe + k * n_pes - 1;
         k++;
      }
      if ( l > -1 ) buffer [ i ] *= bufout [ l ]; /*--->operation<---*/
   }

   pla_free ( bufin );
   pla_free ( bufout );

   return ( 0 );
}

int PLA_Mvector_prefix_mult ( PLA_Obj mvector )
{
   int
      i, j, k, l,
      nb,
      my_pe,
      n_pes,
      global_align,
      global_length,
      global_width,
      local_length,
      local_width,
      local_stride,
      local_ld,
      nblocks;

   double
    * buffer,
    * bufin,
    * bufout;

   MPI_Datatype
      datatype;

   MPI_Comm
      comm;

   PLA_Template
      template;

   PLA_Mvector_local_all ( mvector, & datatype, & local_length, & local_width,
                          ( void ** ) & buffer, & local_stride, & local_ld );
   PLA_Mvector_global_all ( mvector, & global_length, & global_width,
                                    & global_align, & template );
   PLA_Temp_nb ( template, & nb );

   nblocks = global_length / nb;
   if ( global_length % nb ) nblocks++;

   bufin = ( double * ) calloc ( nblocks * global_width, sizeof ( double ) );
   bufout = ( double * ) calloc ( nblocks * global_width, sizeof ( double ) );

   PLA_Temp_comm1D ( template, & comm );
   PLA_Temp_total_num_proc ( template, & n_pes );
   PLA_Temp_me1D ( template, & my_pe );

#define BUFIN(i,j) bufin [ ( j ) * nblocks + ( i ) ]
#define BUFOUT(i,j) bufout [ ( j ) * nblocks + ( i ) ]
#define BUFFER(i,j) buffer [ ( j ) * local_ld + ( i ) ]


   for  ( j = 0 ; j < global_width ; j++ )
   {
      k = 0;
      for ( i = 1 ; i < local_length ; i++ )
      {
         if ( i % nb == 0 )
         { 
            BUFIN ( my_pe + k * n_pes, j ) = BUFFER ( i - 1, j );
            k++;
         }
         else
            BUFFER ( i, j ) *= BUFFER ( i - 1, j ); /*--->operation<---*/
      }
      BUFIN ( my_pe + k * n_pes, j ) = BUFFER ( local_length - 1, j );
   }

   /* this allreduce collects the contribution of all distributed blocks
      into a single array - no reduction operation are performed here!!! */

   MPI_Allreduce ( bufin, bufout, nblocks * global_width, MPI_DOUBLE, MPI_SUM, comm );

   for ( j = 0 ; j < global_width ; j++ )
      for ( i = 1 ; i < nblocks ; i++ )
         BUFOUT ( i, j ) *= BUFOUT ( i - 1, j ); /*--->operation<---*/

   for ( j = 0 ; j < global_width ; j++ )
   {
      k = 0;
      for ( i = 0 ; i < local_length ; i++ )
      {
         if ( i % nb == 0 )
         {
            l = my_pe + k * n_pes - 1;
            k++;
         }
         if ( l > -1 ) BUFFER ( i, j ) *= BUFOUT ( l, j ); /*--->operation<---*/
      }
   }

#undef BUFIN
#undef BUFOUT
#undef BUFFER

   pla_free ( bufin );
   pla_free ( bufout );

   return ( 0 );
}
