#include "PLA.h"
#include "POOCLA.h"

void PLA_test_ooc_dchol ( int my_pe, PLA_Template template )
{
  int length, width, ldim;
  double * a, time, mflops;

  PLA_Obj a_obj=NULL;
  POOCLA_Obj a_ooc=NULL;
  MPI_Comm comm1d;
  int      i, j, n, np, int_uplo;

  PLA_Temp_comm_all ( template, & comm1d );
  MPI_Comm_size( comm1d, &np );

  if ( my_pe == 0 )
  {
     printf ( "n\n" );
     scanf ( "%d", & n );
  }

  MPI_Bcast ( & n, 1, MPI_INT, 0, comm1d );

  a = pla_calloc ( n * n, sizeof ( double ) );

  if ( my_pe == 0 )
     PLA_test_dgenmatrix ( "D", 10.0e+0, n, n, a, n );

  MPI_Bcast ( a, n * n, MPI_DOUBLE, 0, comm1d );


/*
 * create in-core and ooc matrices
 */

  PLA_Matrix_create ( MPI_DOUBLE, n, n, template, 0, 0, & a_obj );
  PLA_Obj_set_to_minus_one ( a_obj );
  PLA_Test_gdist_to_local ( a, n, a_obj );

  POOCLA_create ( template, MPI_DOUBLE, n, n, 0, 0,
                   PLA_PMVECTOR, PLA_PROJ_ONTO_ROW, 1, "Mat_A", & a_ooc );
  POOCLA_write ( a_ooc, a_obj );
  POOCLA_validate_filestatus ( a_ooc );

/*
 * initialize scaling parameters and constants
 */
  int_uplo =  PLA_LOWER_TRIANGULAR;

/*
 * calculate serial chol and keep result in a
 */
  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime();

  PLA_dchol ( n, a, n );

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime() - time;
  mflops = 1.0e-6 * n * n * n / ( 3.0e+0 * time * np );

  if ( my_pe == 0 ) printf ( "Serial : %f seconds at %f mflops\n", time, mflops );

/*
 * calculate incore chol and keep result in a_obj
 */
  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime();

  PLA_Chol ( int_uplo, a_obj );

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime() - time;
  mflops = 1.0e-6 * n * n * n / ( 3.0e+0 * time * np );

  if ( my_pe == 0 ) printf ( "Incore : %f seconds at %f mflops\n", time, mflops );

/*
 * calculation ooc chol, results on disk
 */
  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime();

  POOCLA_Chol ( int_uplo, a_ooc );

  MPI_Barrier( MPI_COMM_WORLD );
  time = MPI_Wtime() - time;
  mflops = 1.0e-6 * n * n * n / ( 3.0e+0 * time * np );

  if ( my_pe == 0 ) printf ( "Out-of-Core : %f seconds at %f mflops\n", time, mflops );

/*
 * read ooc solution into a_obj
 */
  POOCLA_read ( a_ooc, PLA_MATRIX, PLA_UNDEFINED, PLA_UNDEFINED, & a_obj );
  PLA_test_compare ( a, n, a_obj );


  PLA_Obj_free ( & a_obj );
  POOCLA_free ( & a_ooc );
  pla_free ( a );
}
