#include "PLA.h"
#include "POOCLA.h"

int POOCLA_NB_ROW_INCORE = 0;
int POOCLA_NB_COL_INCORE = 0;
int POOCLA_INSTRUMENT_IO = 0;
double POOCLA_START_TIME = 0.0e+0;
double POOCLA_READ_TIME  = 0.0e+0;
double POOCLA_WRITE_TIME = 0.0e+0;

/*
 * The creation routine does not check datatype!!!!!
 */

int POOCLA_create ( PLA_Template   template,
                     MPI_Datatype   datatype,
                     int            universal_length,
                     int            universal_width,
                     int            universal_align_row,
                     int            universal_align_col,
                     int            objtype,
                     int            proj_onto,
                     int            owner,
                     char *         basefilename,
                     POOCLA_Obj *      ooc_object )
{
   int my_pe, r_dim, c_dim, my_row, my_col, nb, filestatus = 0;
   MPI_Comm comm_all = NULL;
   POOCLA_Obj new_object = NULL;
   char mode[3], routine_name [ 16 ] = "POOCLA_create";

   PLA_Stack_push_current ( routine_name );

   POOCLA_free ( ooc_object );
   new_object = ( POOCLA_Obj ) pla_calloc ( 1, sizeof ( POOCLA_OBJECT_STRUCT ) );
   if ( new_object == NULL )
   {
      PLA_Log_message ( routine_name, "POOCLA_OBJECT_STRUCT allocation not successful" );
      PLA_Stack_pop_current ( routine_name );
      return ( PLA_RETURN_LOG_MESSAGE );
   }
   * ooc_object = new_object;

   PLA_Temp_nb ( template, & nb );
   PLA_Temp_comm_all ( template, & comm_all );
   PLA_Temp_comm_all_rank ( template, & my_pe );
   PLA_Temp_comm_row_size ( template, & c_dim );
   PLA_Temp_comm_row_rank ( template, & my_col );
   PLA_Temp_comm_col_size ( template, & r_dim );
   PLA_Temp_comm_col_rank ( template, & my_row );

   if ( my_pe == 0 )
   {
      FILE * descfile = NULL;
      descfile = fopen ( basefilename, "rb" );

      if ( descfile )
      {
         int r_dim_old, c_dim_old, nb_old;

         fread ( & nb_old, sizeof ( int ), 1, descfile );
         fread ( & r_dim_old, sizeof ( int ), 1, descfile );
         fread ( & c_dim_old, sizeof ( int ), 1, descfile );
         fread ( & new_object->universal_length, sizeof ( int ), 1, descfile );
         fread ( & new_object->universal_width, sizeof ( int ), 1, descfile );
         fread ( & new_object->universal_align_row, sizeof ( int ), 1, descfile );
         fread ( & new_object->universal_align_col, sizeof ( int ), 1, descfile );
         fread ( & new_object->objtype, sizeof ( int ), 1, descfile );
         fread ( & new_object->proj_onto, sizeof ( int ), 1, descfile );
         fread ( & new_object->owner, sizeof ( int ), 1, descfile );

         filestatus += ( nb_old != nb );
         filestatus += ( r_dim_old != r_dim );
         filestatus += ( c_dim_old != c_dim );
         filestatus += ( new_object->universal_length != universal_length );
         filestatus += ( new_object->universal_width != universal_width );
         filestatus += ( new_object->universal_align_row != universal_align_row );
         filestatus += ( new_object->universal_align_col != universal_align_col );
         filestatus += ( new_object->objtype != objtype );
         filestatus += ( new_object->proj_onto != proj_onto );
         filestatus += ( new_object->owner != owner );
      }
      else
      {
         descfile = fopen ( basefilename, "wb" );
         if ( descfile == NULL )
            PLA_Log_abort ( routine_name, "Error opening new descfile",
                            PLA_ERROR_CODE, 0 );

         fwrite ( & nb, sizeof ( int ), 1, descfile );
         fwrite ( & r_dim, sizeof ( int ), 1, descfile );
         fwrite ( & c_dim, sizeof ( int ), 1, descfile );
         fwrite ( & universal_length, sizeof ( int ), 1, descfile );
         fwrite ( & universal_width, sizeof ( int ), 1, descfile );
         fwrite ( & universal_align_row, sizeof ( int ), 1, descfile );
         fwrite ( & universal_align_col, sizeof ( int ), 1, descfile );
         fwrite ( & objtype, sizeof ( int ), 1, descfile );
         fwrite ( & proj_onto, sizeof ( int ), 1, descfile );
         fwrite ( & owner, sizeof ( int ), 1, descfile );

         filestatus = -1;
      }

      fclose ( descfile );
      descfile = NULL;
   }

   MPI_Bcast ( & filestatus, 1, MPI_INT, 0, comm_all );

   sprintf ( mode, "rb+" );

   if ( filestatus > 0 )
   {
      PLA_Log_message ( routine_name,
                        "Topology/Blocking/Size Parameters Inconsistant" );
      POOCLA_free ( ooc_object );
      PLA_Stack_pop_current ( routine_name );
      return ( PLA_RETURN_LOG_MESSAGE );
   }

   if ( filestatus < 0 ) sprintf ( mode, "wb+" );

   new_object->template = template;
   new_object->datatype = datatype;
   new_object->filestatus = filestatus;
   new_object->global_length = universal_length;
   new_object->global_width = universal_width;
   new_object->global_align_row = universal_align_row;
   new_object->global_align_col = universal_align_col;
   new_object->universal_length = universal_length;
   new_object->universal_width = universal_width;
   new_object->universal_align_row = universal_align_row;
   new_object->universal_align_col = universal_align_col;
   new_object->objtype = objtype;
   new_object->proj_onto = proj_onto;
   new_object->owner = owner;


   new_object->number_views = pla_calloc ( 1, sizeof ( int ) );
   if ( new_object->number_views == NULL )
      PLA_Log_message ( routine_name,
                        "new_object->number_views allocation not successful" );

   sprintf ( new_object->basefilename, "%s\0", basefilename );

   if ( objtype == PLA_MATRIX )
   {
      PLA_Temp_matrix_local_length ( template, universal_length, universal_align_row,
                                     & ( new_object->file_ldim ) );
      sprintf ( new_object->datafilename, "%s.pe%03d\0", basefilename, my_pe );
      new_object->datafile = fopen ( new_object->datafilename, mode );
   }
   else if ( objtype == PLA_PMVECTOR )
   {
      if ( owner == PLA_ALL_ROWS || owner == PLA_ALL_COLS )
      {
         PLA_Log_message ( routine_name,
                           "Duplicated object not permitted" );
         POOCLA_free ( ooc_object );
         PLA_Stack_pop_current ( routine_name );
         return ( PLA_RETURN_LOG_MESSAGE );
      }

      if ( proj_onto == PLA_PROJ_ONTO_COL )
      {
         new_object->global_align_col = 0;
         new_object->universal_align_col = 0;
         PLA_Temp_matrix_local_length ( template, universal_length, universal_align_row,
                                        & ( new_object->file_ldim ) );
         if ( owner == my_col )
         {
            sprintf ( new_object->datafilename, "%s.row%02d\0", basefilename, my_row );
            new_object->datafile = fopen ( new_object->datafilename, mode );
         }
      }
      else if ( proj_onto == PLA_PROJ_ONTO_ROW )
      {
         new_object->global_align_row = 0;
         new_object->universal_align_row = 0;
         new_object->file_ldim = universal_length;
         if ( owner == my_row )
         {
            sprintf ( new_object->datafilename, "%s.col%02d\0", basefilename, my_col );
            new_object->datafile = fopen ( new_object->datafilename, mode );
         }
      }
   }
   else
   {
      PLA_Log_message ( routine_name, "Object type not accepted in Ooc_create" );
      POOCLA_free ( ooc_object );
      return ( PLA_RETURN_LOG_MESSAGE );
   }

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int POOCLA_free   ( POOCLA_Obj *      ooc_object )
{
   int number_views;
   POOCLA_Obj old_object = NULL;
   old_object = * ooc_object;
   if ( old_object )
   {
      poocla_number_views ( old_object, & number_views );
      if ( number_views == 0 )
      {
         if ( old_object->datafile ) fclose ( old_object->datafile );
         if ( old_object->number_views ) pla_free ( old_object->number_views );
      }
      else
         poocla_decrement_number_views ( old_object );

      pla_free ( old_object );
   }
   * ooc_object = NULL;
   return ( PLA_SUCCESS );
}

int POOCLA_view_all ( POOCLA_Obj ooc_object, POOCLA_Obj * new_object )
{
   int number_views;
   char routine_name [ 16 ] = "POOCLA_view_all";
   PLA_Stack_push_current ( routine_name );
   if ( * new_object )
   {
      poocla_number_views ( * new_object, & number_views );
      if ( number_views > 0 )
         poocla_decrement_number_views ( * new_object );
      else
         POOCLA_free ( new_object );
   }

   if ( * new_object == NULL )
      * new_object = ( POOCLA_Obj ) pla_calloc ( 1, sizeof ( POOCLA_OBJECT_STRUCT ) );

   * ( * new_object ) = * ooc_object;

   poocla_increment_number_views ( ooc_object );

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int POOCLA_split_size ( POOCLA_Obj ooc_object, int side, int * size )
{
   extern int POOCLA_NB_ROW_INCORE;
   extern int POOCLA_NB_COL_INCORE;
   int length, width, align_row, align_col,
       nb_row_incore = POOCLA_NB_ROW_INCORE,
       nb_col_incore = POOCLA_NB_COL_INCORE;

   POOCLA_global_length ( ooc_object, & length );
   POOCLA_global_width ( ooc_object, & width );
   POOCLA_global_align_row ( ooc_object, & align_row );
   POOCLA_global_align_col ( ooc_object, & align_col );

   if ( side == PLA_SIDE_TOP )
   {
      * size = nb_row_incore - align_row % nb_row_incore;
      if ( * size > length ) * size = length;
   }
   else if ( side == PLA_SIDE_BOTTOM )
   {
      * size = ( length + align_row ) % nb_row_incore;
      if ( * size == 0 ) * size = nb_row_incore;
      if ( * size > length ) * size = length;
   }
   else if ( side == PLA_SIDE_LEFT )
   {
      * size = nb_col_incore - align_col % nb_col_incore;
      if ( * size > width ) * size = width;
   }
   else if ( side == PLA_SIDE_RIGHT )
   {
      * size = ( width + align_col ) % nb_col_incore;
      if ( * size == 0 ) * size = nb_col_incore;
      if ( * size > width ) * size = width;
   }

   return ( PLA_SUCCESS );
}

int POOCLA_view_shift ( POOCLA_Obj ooc_object,
                         int length_top,
                         int width_left,
                         int width_right,
                         int length_bottom )
{
   int universal_length, universal_width,
       universal_align_row, universal_align_col,
       length, width, align_row, align_col;
   char routine_name [ 32 ] = "POOCLA_view_shift";

   PLA_Stack_push_current ( routine_name );

   POOCLA_global_length ( ooc_object, & length );
   POOCLA_global_width ( ooc_object, & width );
   POOCLA_global_align_row ( ooc_object, & align_row );
   POOCLA_global_align_col ( ooc_object, & align_col );
   POOCLA_universal_length ( ooc_object, & universal_length );
   POOCLA_universal_width ( ooc_object, & universal_width );
   POOCLA_universal_align_row ( ooc_object, & universal_align_row );
   POOCLA_universal_align_col ( ooc_object, & universal_align_col );

   length -= length_top;	length += length_bottom;
   width  -= width_left;	width  += width_right;

   align_row += length_top;
   align_col += width_left;

   if ( align_row < universal_align_row ||
        align_col < universal_align_col ||
        align_row + length > universal_align_row + universal_length ||
        align_col + width  > universal_align_col + universal_width )
   {
      PLA_Log_message ( routine_name, "View Out-Of-Bounds" );
      PLA_Log_abort ( routine_name, "View Out-Of-Bounds", PLA_ERROR_CODE, 0 );
      PLA_Stack_pop_current ( routine_name );
      return ( PLA_RETURN_LOG_MESSAGE );
   }

   ooc_object->global_length = length;
   ooc_object->global_width = width;
   ooc_object->global_align_row = align_row;
   ooc_object->global_align_col = align_col;

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int POOCLA_horz_split_2 ( POOCLA_Obj ooc_object,
                           int size,
                           POOCLA_Obj * top_object,
                           POOCLA_Obj * bottom_object )
{
   int length;
   char routine_name [ 32 ] = "POOCLA_horz_split_2";
   PLA_Stack_push_current ( routine_name );
   POOCLA_global_length ( ooc_object, & length );
   POOCLA_view_all ( ooc_object, top_object );
   POOCLA_view_all ( ooc_object, bottom_object );

   if ( size > 0 )
   {
      POOCLA_view_shift ( * top_object, 0, 0, 0, size - length );
      POOCLA_view_shift ( * bottom_object, size, 0, 0, 0 );
   }
   else if ( size < 0 )
   {
      POOCLA_view_shift ( * top_object, 0, 0, 0, size );
      POOCLA_view_shift ( * bottom_object, length + size, 0, 0, 0 );
   }
   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int POOCLA_vert_split_2 ( POOCLA_Obj ooc_object,
                           int size,
                           POOCLA_Obj * left_object,
                           POOCLA_Obj * right_object )
{
   int width;
   char routine_name [ 32 ] = "POOCLA_vert_split_2";
   PLA_Stack_push_current ( routine_name );
   POOCLA_global_width ( ooc_object, & width );
   POOCLA_view_all ( ooc_object, left_object );
   POOCLA_view_all ( ooc_object, right_object );

   if ( size > 0 )
   {
      POOCLA_view_shift ( * left_object, 0, 0, size - width, 0 );
      POOCLA_view_shift ( * right_object, 0, size, 0, 0 );
   }
   else if ( size < 0 )
   {
      POOCLA_view_shift ( * left_object, 0, 0, size, 0 );
      POOCLA_view_shift ( * right_object, 0, width + size, 0, 0 );
   }
   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int POOCLA_split_4 ( POOCLA_Obj ooc_object,
                      int length, int width,
                      POOCLA_Obj * topleft_object,
                      POOCLA_Obj * topright_object,
                      POOCLA_Obj * bottomleft_object,
                      POOCLA_Obj * bottomright_object )
{
   int global_length, global_width;
   char routine_name [ 32 ] = "POOCLA_split_4";
   PLA_Stack_push_current ( routine_name );

   POOCLA_global_length ( ooc_object, & global_length );
   POOCLA_global_width ( ooc_object, & global_width );
   POOCLA_view_all ( ooc_object, topleft_object );
   POOCLA_view_all ( ooc_object, topright_object );
   POOCLA_view_all ( ooc_object, bottomleft_object );
   POOCLA_view_all ( ooc_object, bottomright_object );

   if ( length > 0 && width > 0 )
   {
      POOCLA_view_shift ( * topleft_object,
                           0, 0, width - global_width, length - global_length );
      POOCLA_view_shift ( * topright_object,
                           0, width, 0, length - global_length );
      POOCLA_view_shift ( * bottomleft_object,
                           length, 0, width - global_width, 0 );
      POOCLA_view_shift ( * bottomright_object,
                           length, width, 0, 0 );
   }
   else if ( length > 0 && width < 0 )
   {
      POOCLA_view_shift ( * topleft_object,
                           0, 0, width, length - global_length );
      POOCLA_view_shift ( * topright_object,
                           0, global_width + width, 0, length - global_length );
      POOCLA_view_shift ( * bottomleft_object,
                           length, 0, width , 0 );
      POOCLA_view_shift ( * bottomright_object,
                           length, global_width + width, 0, 0 );
   }
   else if ( length < 0 && width > 0 )
   {
      POOCLA_view_shift ( * topleft_object,
                           0, 0, width - global_width, length );
      POOCLA_view_shift ( * topright_object,
                           0, width, 0, length );
      POOCLA_view_shift ( * bottomleft_object,
                           global_length + length, 0, width - global_width, 0 );
      POOCLA_view_shift ( * bottomright_object,
                           global_length + length, width, 0, 0 );
   }
   if ( length < 0 && width < 0 )
   {
      POOCLA_view_shift ( * topleft_object,
                           0, 0, width, length );
      POOCLA_view_shift ( * topright_object,
                           0, global_width + width, 0, length );
      POOCLA_view_shift ( * bottomleft_object,
                           global_length + length, 0, width, 0 );
      POOCLA_view_shift ( * bottomright_object,
                           global_length + length, global_width + width, 0, 0 );
   }
   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}


int POOCLA_read   ( POOCLA_Obj        ooc_object,
                     int            incore_objtype,
                     int            incore_proj_onto,
                     int            incore_owner,
                     PLA_Obj *      incore_object )
{
   int ooc_objtype, ooc_owner, filestatus,
       global_length, global_width,
       global_align_row, global_align_col;
   MPI_Datatype datatype;
   PLA_Template template;
   char routine_name [ 16 ] = "POOCLA_read";

   PLA_Stack_push_current ( routine_name );
   POOCLA_set_start_time ();
   POOCLA_objtype ( ooc_object, & ooc_objtype );
   POOCLA_owner ( ooc_object, & ooc_owner );
   POOCLA_filestatus ( ooc_object, & filestatus );
   POOCLA_global_length ( ooc_object, & global_length );
   POOCLA_global_width ( ooc_object, & global_width );
   POOCLA_global_align_row ( ooc_object, & global_align_row );
   POOCLA_global_align_col ( ooc_object, & global_align_col );
   POOCLA_datatype ( ooc_object, & datatype );
   POOCLA_template ( ooc_object, & template );

   if ( incore_objtype == PLA_MATRIX )
   {
      if ( PLA_Matrix_create ( datatype,
                               global_length,
                               global_width,
                               template,
                               global_align_row,
                               global_align_col,
                               incore_object ) != PLA_SUCCESS )
         PLA_Log_abort ( routine_name,
                         "PLA_Matrix_create returned with error", PLA_ERROR_CODE, 0 );
   }
   else if ( incore_objtype == PLA_PMVECTOR )
   {
      if ( incore_owner != ooc_owner )
      {
         PLA_Log_message ( routine_name, "Incore owner not equal Ooc owner" );
         PLA_Stack_pop_current ( routine_name );
         return ( PLA_RETURN_LOG_MESSAGE );
      }
      else
         if ( PLA_Pmvector_create ( datatype,
                                    incore_proj_onto,
                                    incore_owner,
                                    global_length,
                                    global_width,
                                    template,
                                    ( incore_proj_onto == PLA_PROJ_ONTO_COL ?
                                      global_align_row : global_align_col ),
                                    incore_object ) != PLA_SUCCESS )
         PLA_Log_abort ( routine_name,
                         "PLA_Pmvector_create returned with error", PLA_ERROR_CODE, 0 );
   }

   if ( filestatus != 0 )
   {
      PLA_Obj_set_to_zero ( * incore_object );
      PLA_Log_message ( routine_name, "PLA_Obj returned for non-zero filestatus" );
      PLA_Stack_pop_current ( routine_name );
      return ( PLA_RETURN_LOG_MESSAGE );
   }

   if ( ooc_objtype == incore_objtype )
      poocla_read_same_type ( ooc_object, * incore_object );
   else
   {
      if ( incore_objtype == PLA_MATRIX )
         poocla_read_pmvector_to_matrix ( ooc_object, * incore_object );
      else
      {
         PLA_Log_message ( routine_name, "Read from matrix to pmvector not supported" );
         PLA_Obj_free ( incore_object );
         PLA_Stack_pop_current ( routine_name );
         return ( PLA_RETURN_LOG_MESSAGE );
      }
   }

   PLA_Stack_pop_current ( routine_name );
   POOCLA_increment_read_time ();
   return ( PLA_SUCCESS );
}

int poocla_read_same_type ( POOCLA_Obj ooc_object, PLA_Obj incore_object )
{
   int objtype, proj_onto, ldim, file_ldim, typesize,
       file_row_offset, file_col_offset, universal_align_row, universal_align_col,
       global_length, global_width, global_align_row, global_align_col,
       align_row_wrt_ooc, align_col_wrt_ooc, local_length, local_width;
   MPI_Datatype datatype = MPI_DATATYPE_NULL;
   PLA_Template template = NULL;
   FILE * datafile = NULL;
   void * buffer = NULL;
   char routine_name [ 32 ] = "poocla_read_same_type";

   PLA_Stack_push_current ( routine_name );

   POOCLA_objtype ( ooc_object, & objtype );
   POOCLA_proj_onto ( ooc_object, & proj_onto );

   POOCLA_universal_align_row ( ooc_object, & universal_align_row );
   POOCLA_universal_align_col ( ooc_object, & universal_align_col );

   POOCLA_global_length ( ooc_object, & global_length );
   POOCLA_global_width ( ooc_object, & global_width );
   POOCLA_global_align_row ( ooc_object, & global_align_row );
   POOCLA_global_align_col ( ooc_object, & global_align_col );

   POOCLA_datatype ( ooc_object, & datatype );
   POOCLA_template ( ooc_object, & template );
   POOCLA_datafile ( ooc_object, & datafile );
   POOCLA_file_ldim ( ooc_object, & file_ldim );

   MPI_Type_extent ( datatype, ( MPI_Aint * ) & typesize );

   PLA_Obj_local_length ( incore_object, & local_length );
   PLA_Obj_local_width ( incore_object, & local_width );
   PLA_Obj_local_buffer ( incore_object, & buffer );
   PLA_Obj_local_ldim ( incore_object, & ldim );

   align_row_wrt_ooc = global_align_row - universal_align_row;
   align_col_wrt_ooc = global_align_col - universal_align_col;

   if ( objtype == PLA_MATRIX )
   {
      PLA_Temp_matrix_local_length_width ( template,
                                           align_row_wrt_ooc, align_col_wrt_ooc,
                                           universal_align_row, universal_align_col,
                                           & file_row_offset, & file_col_offset );
   }
   else if ( objtype == PLA_PMVECTOR )
   {
      if ( proj_onto == PLA_PROJ_ONTO_COL )
      {
         PLA_Temp_matrix_local_length ( template, align_row_wrt_ooc,
                                        universal_align_row, & file_row_offset );
         file_col_offset = align_col_wrt_ooc;
      }
      else if ( proj_onto == PLA_PROJ_ONTO_ROW )
      {
         file_row_offset = align_row_wrt_ooc;
         PLA_Temp_matrix_local_width ( template, align_col_wrt_ooc,
                                       universal_align_col, & file_col_offset );
      }
   }

   if ( datafile != NULL )
   {
      fseek ( datafile,
              ( file_col_offset * file_ldim + file_row_offset ) * typesize,
              SEEK_SET );
      poocla_read_data ( local_length, local_width, typesize,
                          datafile, file_ldim, buffer, ldim );
      fflush ( datafile );
      fseek ( datafile, 0, SEEK_SET );
   }

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int poocla_read_pmvector_to_matrix ( POOCLA_Obj ooc, PLA_Obj obj )
{
   int proj_onto, obj_dim, nb, size, owner, n_pes;
   PLA_Template template = NULL;
   POOCLA_Obj ooc_slab = NULL, ooc_block = NULL;
   PLA_Obj obj_slab = NULL, obj_block = NULL, buffer = NULL;
   char routine_name [ 32 ] = "poocla_read_pmvector_to_matrix";

   PLA_Stack_push_current ( routine_name );

   POOCLA_proj_onto ( ooc, & proj_onto );
   POOCLA_template ( ooc, & template );
   PLA_Temp_comm_all_size ( template, & n_pes );

   POOCLA_view_all ( ooc, & ooc_block );
   PLA_Obj_view_all ( obj, & obj_block );


   if ( proj_onto == PLA_PROJ_ONTO_COL )
   {
      POOCLA_global_width ( ooc_block, & obj_dim );
      POOCLA_get_nb_col_incore ( & nb );
      nb /= n_pes;
   }
   else if ( proj_onto == PLA_PROJ_ONTO_ROW )
   {
      POOCLA_global_length ( ooc_block, & obj_dim );
      POOCLA_get_nb_row_incore ( & nb );
      nb /= n_pes;
   }

   while ( ( size = min ( nb, obj_dim ) ) > 0 )
   {
      if ( proj_onto == PLA_PROJ_ONTO_COL )
      {
         POOCLA_vert_split_2 ( ooc_block, size, & ooc_slab, & ooc_block );
         PLA_Obj_vert_split_2 ( obj_block, size, & obj_slab, & obj_block );
      }
      else if ( proj_onto == PLA_PROJ_ONTO_ROW )
      {
         POOCLA_horz_split_2 ( ooc_block, size, & ooc_slab, & ooc_block );
         PLA_Obj_horz_split_2 ( obj_block, size, & obj_slab, & obj_block );
      }

      POOCLA_create_PLA_Obj_conf_to ( ooc_slab, PLA_PMVECTOR, & buffer );
      PLA_Obj_set_to_zero ( buffer );
      poocla_read_same_type ( ooc_slab, buffer );
      pla_Copy_pmv2mat ( buffer, obj_slab );

      obj_dim -= size;
   }

   POOCLA_free ( & ooc_slab );
   POOCLA_free ( & ooc_block );

   PLA_Obj_free ( & obj_slab );
   PLA_Obj_free ( & obj_block );
   PLA_Obj_free ( & buffer );

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int POOCLA_write  ( POOCLA_Obj        ooc_object,
                     PLA_Obj        incore_object )
{
   int ooc_objtype, incore_objtype;
   char routine_name [ 16 ] = "POOCLA_write";
   PLA_Stack_push_current ( routine_name );
   POOCLA_set_start_time ();
   POOCLA_objtype ( ooc_object, & ooc_objtype );
   PLA_Obj_objtype ( incore_object, & incore_objtype );

   if ( ooc_objtype == incore_objtype )
      poocla_write_same_type ( ooc_object, incore_object );
   else
   {
      if ( incore_objtype == PLA_MATRIX )
         poocla_write_matrix_to_pmvector ( ooc_object, incore_object );
      else
      {
         PLA_Log_message ( routine_name, "Write to matrix from pmvector no supported" );
         PLA_Stack_pop_current ( routine_name );
         return ( PLA_RETURN_LOG_MESSAGE );
      }
   }

   PLA_Stack_pop_current ( routine_name );
   POOCLA_increment_write_time ();
   return ( PLA_SUCCESS );
}

int poocla_write_same_type ( POOCLA_Obj ooc_object,
                              PLA_Obj incore_object )
{
   int objtype, proj_onto, ldim, file_ldim, typesize,
       file_row_offset, file_col_offset, universal_align_row, universal_align_col,
       global_length, global_width, global_align_row, global_align_col,
       align_row_wrt_ooc, align_col_wrt_ooc, local_length, local_width;
   MPI_Datatype datatype = MPI_DATATYPE_NULL;
   PLA_Template template = NULL;
   FILE * datafile = NULL;
   void * buffer = NULL;
   char routine_name [ 32 ] = "poocla_write_same_type";

   PLA_Stack_push_current ( routine_name );

   POOCLA_objtype ( ooc_object, & objtype );
   POOCLA_proj_onto ( ooc_object, & proj_onto );

   POOCLA_universal_align_row ( ooc_object, & universal_align_row );
   POOCLA_universal_align_col ( ooc_object, & universal_align_col );

   POOCLA_global_length ( ooc_object, & global_length );
   POOCLA_global_width ( ooc_object, & global_width );
   POOCLA_global_align_row ( ooc_object, & global_align_row );
   POOCLA_global_align_col ( ooc_object, & global_align_col );

   POOCLA_datatype ( ooc_object, & datatype );
   POOCLA_template ( ooc_object, & template );
   POOCLA_datafile ( ooc_object, & datafile );
   POOCLA_file_ldim ( ooc_object, & file_ldim );

   MPI_Type_extent ( datatype, ( MPI_Aint * ) & typesize );

   PLA_Obj_local_length ( incore_object, & local_length );
   PLA_Obj_local_width ( incore_object, & local_width );
   PLA_Obj_local_buffer ( incore_object, & buffer );
   PLA_Obj_local_ldim ( incore_object, & ldim );

   align_row_wrt_ooc = global_align_row - universal_align_row;
   align_col_wrt_ooc = global_align_col - universal_align_col;

   if ( objtype == PLA_MATRIX )
   {
      PLA_Temp_matrix_local_length_width ( template,
                                           align_row_wrt_ooc, align_col_wrt_ooc,
                                           universal_align_row, universal_align_col,
                                           & file_row_offset, & file_col_offset );
   }
   else if ( objtype == PLA_PMVECTOR )
   {
      if ( proj_onto == PLA_PROJ_ONTO_COL )
      {
         PLA_Temp_matrix_local_length ( template, align_row_wrt_ooc,
                                        universal_align_row, & file_row_offset );
         file_col_offset = align_col_wrt_ooc;
      }
      else if ( proj_onto == PLA_PROJ_ONTO_ROW )
      {
         file_row_offset = align_row_wrt_ooc;
         PLA_Temp_matrix_local_width ( template, align_col_wrt_ooc,
                                       universal_align_col, & file_col_offset );
      }
   }

   if ( datafile != NULL )
   {
      fseek ( datafile,
              ( file_col_offset * file_ldim + file_row_offset ) * typesize,
              SEEK_SET );
      poocla_write_data ( local_length, local_width, typesize,
                           buffer, ldim, datafile, file_ldim );
      fflush ( datafile );
   }

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

int poocla_write_matrix_to_pmvector ( POOCLA_Obj ooc, PLA_Obj obj )
{
   int proj_onto, obj_dim, nb, size, owner, n_pes;
   PLA_Template template = NULL;
   POOCLA_Obj ooc_slab = NULL, ooc_block = NULL;
   PLA_Obj obj_slab = NULL, obj_block = NULL, buffer = NULL;
   char routine_name [ 64 ] = "poocla_write_matrix_to_pmvector";

   PLA_Stack_push_current ( routine_name );

   POOCLA_proj_onto ( ooc, & proj_onto );
   POOCLA_template ( ooc, & template );
   PLA_Temp_comm_all_size ( template, & n_pes );

   POOCLA_view_all ( ooc, & ooc_block );
   PLA_Obj_view_all ( obj, & obj_block );

   if ( proj_onto == PLA_PROJ_ONTO_COL )
   {
      POOCLA_global_width ( ooc_block, & obj_dim );
      POOCLA_get_nb_col_incore ( & nb );
      nb /= n_pes;
   }
   else
   {
      POOCLA_global_length ( ooc_block, & obj_dim );
      POOCLA_get_nb_row_incore ( & nb );
      nb /= n_pes;
   }

   while ( ( size = min ( nb, obj_dim ) ) > 0 )
   {
      if ( proj_onto == PLA_PROJ_ONTO_COL )
      {
         POOCLA_vert_split_2 ( ooc_block, size, & ooc_slab, & ooc_block );
         PLA_Obj_vert_split_2 ( obj_block, size, & obj_slab, & obj_block );
      }
      else
      {
         POOCLA_horz_split_2 ( ooc_block, size, & ooc_slab, & ooc_block );
         PLA_Obj_horz_split_2 ( obj_block, size, & obj_slab, & obj_block );
      }

      POOCLA_create_PLA_Obj_conf_to ( ooc_slab, PLA_PMVECTOR, & buffer );
      PLA_Obj_set_to_zero ( buffer );

      pla_Copy_mat2pmv ( obj_slab, buffer );
      poocla_write_same_type ( ooc_slab, buffer );

      obj_dim -= size;
   }

   POOCLA_free ( & ooc_slab );
   POOCLA_free ( & ooc_block );

   PLA_Obj_free ( & obj_slab );
   PLA_Obj_free ( & obj_block );
   PLA_Obj_free ( & buffer );

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}


int poocla_read_data ( int length,   /* row dimension */
                        int width,    /* column dimension */
                        int typesize,
                        FILE * source,
                        int ldim_source,
                        void * destin,
                        int ldim_destin )
{
   int i, offset;
   fpos_t position;
   void * buf;

   fgetpos ( source, & position );
   ldim_destin *= typesize;
   offset = ( ldim_source - length ) * typesize;

   for ( i = 0 ; i < width ; i++ )
   {
      buf = ( char * ) destin + i * ldim_destin;
      fread ( buf, typesize, length, source );
      if ( offset ) fseek ( source, offset, SEEK_CUR );
   }

  fsetpos ( source, & position );

  return ( PLA_SUCCESS );
}

int poocla_write_data ( int length,   /* row dimension */
                         int width,    /* column dimension */
                         int typesize,
                         void * source,
                         int ldim_source,
                         FILE * destin,
                         int ldim_destin )
{
   int i, offset;
   fpos_t position;
   void * buf;

   fgetpos ( destin, & position );
   ldim_source *= typesize;
   offset = ( ldim_destin - length ) * typesize;

   for ( i = 0 ; i < width ; i++ )
   {
      buf = ( char * ) source + i * ldim_source;
      fwrite ( buf, typesize, length, destin );
      if ( offset ) fseek ( destin, offset, SEEK_CUR );
   }

  fsetpos ( destin, & position );

  return ( PLA_SUCCESS );
}


int POOCLA_set_nb_row_incore ( int nb_row_incore )
{
   extern int POOCLA_NB_ROW_INCORE;
   POOCLA_NB_ROW_INCORE = nb_row_incore;
   return ( PLA_SUCCESS );
}

int POOCLA_set_nb_col_incore ( int nb_col_incore )
{
   extern int POOCLA_NB_COL_INCORE;
   POOCLA_NB_COL_INCORE = nb_col_incore;
   return ( PLA_SUCCESS );
}

int POOCLA_set_nb_incore ( int nb_row_incore, int nb_col_incore )
{
   extern int POOCLA_NB_ROW_INCORE;
   extern int POOCLA_NB_COL_INCORE;
   POOCLA_NB_ROW_INCORE = nb_row_incore;
   POOCLA_NB_COL_INCORE = nb_col_incore;
   return ( PLA_SUCCESS );
}


int POOCLA_get_nb_row_incore ( int * nb_row_incore )
{
   extern int POOCLA_NB_ROW_INCORE;
   * nb_row_incore = POOCLA_NB_ROW_INCORE;
   return ( PLA_SUCCESS );
}

int POOCLA_get_nb_col_incore ( int * nb_col_incore )
{
   extern int POOCLA_NB_COL_INCORE;
   * nb_col_incore = POOCLA_NB_COL_INCORE;
   return ( PLA_SUCCESS );
}

int POOCLA_get_nb_incore ( int * nb_row_incore, int * nb_col_incore )
{
   extern int POOCLA_NB_ROW_INCORE;
   extern int POOCLA_NB_COL_INCORE;
   * nb_row_incore = POOCLA_NB_ROW_INCORE;
   * nb_col_incore = POOCLA_NB_COL_INCORE;
   return ( PLA_SUCCESS );
}


int POOCLA_template            ( POOCLA_Obj ooc_object, PLA_Template * template )
{
   * template = ooc_object->template;
   return ( PLA_SUCCESS );
}

int POOCLA_datatype            ( POOCLA_Obj ooc_object, MPI_Datatype * datatype )
{
   * datatype = ooc_object->datatype;
   return ( PLA_SUCCESS );
}

int POOCLA_filestatus          ( POOCLA_Obj ooc_object, int * filestatus )
{
   * filestatus = ooc_object->filestatus;
   return ( PLA_SUCCESS );
}

int POOCLA_global_length    ( POOCLA_Obj ooc_object, int * global_length )
{
   * global_length = ooc_object->global_length;
   return ( PLA_SUCCESS );
}

int POOCLA_global_width     ( POOCLA_Obj ooc_object, int * global_width )
{
   * global_width = ooc_object->global_width;
   return ( PLA_SUCCESS );
}

int POOCLA_global_align_row ( POOCLA_Obj ooc_object, int * global_align_row )
{
   * global_align_row = ooc_object->global_align_row;
   return ( PLA_SUCCESS );
}

int POOCLA_global_align_col ( POOCLA_Obj ooc_object, int * global_align_col )
{
   * global_align_col = ooc_object->global_align_col;
   return ( PLA_SUCCESS );
}

int POOCLA_objtype     ( POOCLA_Obj ooc_object, int * objtype )
{
   * objtype = ooc_object->objtype;
   return ( PLA_SUCCESS );
}

int POOCLA_proj_onto   ( POOCLA_Obj ooc_object, int * proj_onto )
{
   * proj_onto = ooc_object->proj_onto;
   return ( PLA_SUCCESS );
}

int POOCLA_owner       ( POOCLA_Obj ooc_object, int * owner )
{
   * owner = ooc_object->owner;
   return ( PLA_SUCCESS );
}

int POOCLA_universal_length    ( POOCLA_Obj ooc_object, int * universal_length )
{
   * universal_length = ooc_object->universal_length;
   return ( PLA_SUCCESS );
}

int POOCLA_universal_width     ( POOCLA_Obj ooc_object, int * universal_width )
{
   * universal_width = ooc_object->universal_width;
   return ( PLA_SUCCESS );
}

int POOCLA_universal_align_row ( POOCLA_Obj ooc_object, int * universal_align_row )
{
   * universal_align_row = ooc_object->universal_align_row;
   return ( PLA_SUCCESS );
}

int POOCLA_universal_align_col ( POOCLA_Obj ooc_object, int * universal_align_col )
{
   * universal_align_col = ooc_object->universal_align_col;
   return ( PLA_SUCCESS );
}

int POOCLA_basefilename        ( POOCLA_Obj ooc_object, char ** basefilename )
{
   * basefilename = ooc_object->basefilename;
   return ( PLA_SUCCESS );
}

int POOCLA_datafilename        ( POOCLA_Obj ooc_object, char ** datafilename )
{
   * datafilename = ooc_object->datafilename;
   return ( PLA_SUCCESS );
}

int POOCLA_datafile            ( POOCLA_Obj ooc_object, FILE ** datafile )
{
   * datafile = ooc_object->datafile;
   return ( PLA_SUCCESS );
}

int POOCLA_file_ldim ( POOCLA_Obj ooc_object, int * file_ldim )
{
   * file_ldim = ooc_object->file_ldim;
   return ( PLA_SUCCESS );
}

int poocla_number_views ( POOCLA_Obj ooc_object, int * number_views )
{
   * number_views = ooc_object->number_views[0];
   return ( PLA_SUCCESS );
}

int poocla_increment_number_views ( POOCLA_Obj ooc_object )
{
   ooc_object->number_views[0]++;
   return ( PLA_SUCCESS );
}

int poocla_decrement_number_views ( POOCLA_Obj ooc_object )
{
   ooc_object->number_views[0]--;
   return ( PLA_SUCCESS );
}


int POOCLA_activate_instrumentation ()
{
   extern int POOCLA_INSTRUMENT_IO;
   POOCLA_INSTRUMENT_IO = 1;
   return ( PLA_SUCCESS );
}

int POOCLA_deactivate_instrumentation ()
{
   extern int POOCLA_INSTRUMENT_IO;
   POOCLA_INSTRUMENT_IO = 0;
   return ( PLA_SUCCESS );
}

int POOCLA_set_start_time ()
{
   extern int POOCLA_INSTRUMENT_IO;
   extern double POOCLA_START_TIME;
   if ( POOCLA_INSTRUMENT_IO ) POOCLA_START_TIME = MPI_Wtime ();
   return ( PLA_SUCCESS );
}

int POOCLA_increment_read_time ()
{
   extern int POOCLA_INSTRUMENT_IO;
   extern double POOCLA_START_TIME;
   extern double POOCLA_READ_TIME;
   if ( POOCLA_INSTRUMENT_IO ) POOCLA_READ_TIME += MPI_Wtime () - POOCLA_START_TIME;
   return ( PLA_SUCCESS );
}

int POOCLA_increment_write_time ()
{
   extern int POOCLA_INSTRUMENT_IO;
   extern double POOCLA_START_TIME;
   extern double POOCLA_WRITE_TIME;
   if ( POOCLA_INSTRUMENT_IO ) POOCLA_WRITE_TIME += MPI_Wtime () - POOCLA_START_TIME;
   return ( PLA_SUCCESS );
}

int POOCLA_reset_read_time ()
{
   extern double POOCLA_READ_TIME;
   POOCLA_READ_TIME = 0.0e+0;
   return ( PLA_SUCCESS );
}

int POOCLA_reset_write_time ()
{
   extern double POOCLA_WRITE_TIME;
   POOCLA_WRITE_TIME = 0.0e+0;
   return ( PLA_SUCCESS );
}

int POOCLA_read_time ( double * read_time )
{
   extern double POOCLA_READ_TIME;
   * read_time = POOCLA_READ_TIME;
   return ( PLA_SUCCESS );
}

int POOCLA_write_time ( double * write_time )
{
   extern double POOCLA_WRITE_TIME;
   * write_time = POOCLA_WRITE_TIME;
   return ( PLA_SUCCESS );
}

int POOCLA_validate_filestatus ( POOCLA_Obj ooc )
{
   ooc->filestatus = 0;
   return ( PLA_SUCCESS );
}

int POOCLA_create_PLA_Obj_conf_to ( POOCLA_Obj ooc, int objtype, PLA_Obj * obj )
{
   int proj_onto, owner,
       global_length, global_width, global_align_row, global_align_col;
   MPI_Datatype datatype = MPI_DATATYPE_NULL;
   PLA_Template template = NULL;
   char routine_name [ 32 ] = "POOCLA_create_PLA_Obj_conf_to";

   PLA_Stack_push_current ( routine_name );

   POOCLA_proj_onto ( ooc, & proj_onto );
   POOCLA_owner ( ooc, & owner );
   POOCLA_global_length ( ooc, & global_length );
   POOCLA_global_width ( ooc, & global_width );
   POOCLA_global_align_row ( ooc, & global_align_row );
   POOCLA_global_align_col ( ooc, & global_align_col );
   POOCLA_datatype ( ooc, & datatype );
   POOCLA_template ( ooc, & template );

   if ( objtype == PLA_MATRIX )
   {
      if ( PLA_Matrix_create ( datatype,
                               global_length,
                               global_width,
                               template,
                               global_align_row,
                               global_align_col,
                               obj ) != PLA_SUCCESS )
      {
         PLA_Obj_free ( obj );
         PLA_Log_message ( routine_name, "PLA_Matrix_create returned with error" );
         PLA_Stack_pop_current ( routine_name );
         return ( PLA_RETURN_LOG_MESSAGE );
      }
   }
   else if ( objtype == PLA_PMVECTOR )
   {
      if ( PLA_Pmvector_create ( datatype,
                                 proj_onto,
                                 owner, 
                                 global_length,
                                 global_width,
                                 template,
                                 ( proj_onto == PLA_PROJ_ONTO_COL ?
                                   global_align_row : global_align_col ),
                                 obj ) != PLA_SUCCESS )
      {
         PLA_Obj_free ( obj );
         PLA_Log_message ( routine_name, "PLA_Pmvector_create returned with error" );
         PLA_Stack_pop_current ( routine_name );
         return ( PLA_RETURN_LOG_MESSAGE );
      }
   }
   else
   {
      PLA_Obj_free ( obj );
      PLA_Log_message ( routine_name, "Object type not valid" );
      PLA_Stack_pop_current ( routine_name );
      return ( PLA_RETURN_LOG_MESSAGE );
   }

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}


int POOCLA_view ( POOCLA_Obj ooc, int length, int width,
                   int align_row, int align_col, POOCLA_Obj * new_ooc )
{
   int global_length, global_width;
   char routine_name [ 16 ] = "POOCLA_view";

   PLA_Stack_push_current ( routine_name );
   POOCLA_global_length ( ooc, & global_length );
   POOCLA_global_width ( ooc, & global_width );

   POOCLA_view_all ( ooc, new_ooc );
   POOCLA_view_shift ( * new_ooc, align_row,
                                   align_col,
                                   align_col + width - global_width,
                                   align_row + length - global_length );

   PLA_Stack_pop_current ( routine_name );
   return ( PLA_SUCCESS );
}

