#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_mean(PLA_Obj X, PLA_Obj lambda_mean) {
/*
    Compute the mean eigenvalue of symmetric matrix X.
    method: lambda_mean = trace(X) / global_length(X) 
*/
    int      length;
    double   scale_value;
    PLA_Obj  trace=NULL, alpha=NULL;
    MPI_Datatype datatype;

    /* Extract info from matrix X */
    PLA_Obj_global_length (X, &length);
    PLA_Obj_datatype      (X, &datatype);

    /* Create temporary objects */
    PLA_Mscalar_create_conf_to(lambda_mean, PLA_ALL_ROWS, PLA_ALL_COLS, &trace);
    PLA_Mscalar_create_conf_to(lambda_mean, PLA_ALL_ROWS, PLA_ALL_COLS, &alpha);

    /* Compute trace(X) */
    PRISM_trace(X, trace);

    /* Compute lambda_mean = trace(X) / length(X) */
    scale_value = 1. / length;
    PLA_Obj_set_to_zero(lambda_mean);
    PLA_Obj_set(alpha, datatype, &scale_value);
    PLA_Local_axpy(alpha, trace, lambda_mean);

    PLA_Obj_free(&alpha);
    PLA_Obj_free(&trace);

#ifdef DB_GLOBAL_
    if (me==ROOT) {
       printf(">> PRISM_mean\n");
       printf("lambda_mean="); prism_pla_prt_m(lambda_mean);
    }
#endif

}

