/* various print functions for PLAPACK objects */

#include <stdio.h>
#include "mpi.h"
#include "PLA.h"
#include "PRISM.h"
 
void prism_pla_ginfo(PLA_Obj A)
{
  PLA_Template  template;
  int m, n, im, in, lld,l_stride;

  PLA_Matrix_global_all( A, &m, &n, &im, &in, &template);
  PLA_Obj_local_ldim( A, &lld);
  l_stride=A->local_stride;
  printf("of size=(%d,%d) lds=(%d,%d) at template (%d,%d)\n",
	 m,n,lld,l_stride,im,in);

}
 
void prism_pla_linfo(PLA_Obj A)
{
  printf("of local size=(%d,%d)\n",
	 A->local_length, A->local_width);

}
 
 
void prism_pla_prt_m(PLA_Obj A)
{
  void *A_buf=NULL;
  MPI_Datatype datatype;
  int lda, local_l, local_w, width,i,j,k,l;
  
  PLA_Matrix_local_all(A, &datatype, &local_l, &local_w, &A_buf, &lda);
  
  for (k=0; k<local_w; k+=6) {
    width = (local_w-k>6)? 6: local_w-k;
    for (i=0; i<local_l; i++) {
      for (j=0;j<width;j++) {
	l=(k+j)*lda;
	printf("%8.4e ",*((double *)A_buf+(i+l)));
      }
      printf("\n");
    }
    printf("\n");
  }
  printf("----------\n");
}


void prism_pla_prt_Qresid(PLA_Obj A, double *resid)
{
  void *A_buf;
  double l_resid, abs_a;
  MPI_Datatype datatype;
  int lda, local_l, local_w, i,k,l;
  
  PLA_Matrix_local_all(A, &datatype, &local_l, &local_w, &A_buf, 
		      &lda);
  
  l_resid=0.0;
  for (k=0; k<local_w; k++){
    l=k*lda;
    for (i=0; i<local_l; i++) {
      abs_a = ABS(*((double *)A_buf+(i+l))); 
      if (l_resid< abs_a ) { 
	l_resid=abs_a;
      }
    }
  }
  *resid=l_resid;
}
 
 
void prism_loc_prt_m(double *A, int local_l, int local_w, int lda)
{
  int k, width, j, i,l;  

  for (k=0; k<local_w; k+=6) {
    width = (local_w-k>6)? 6: local_w-k;
    for (i=0; i<local_l; i++) {
      printf("%d: ",me);
      for (j=0;j<width;j++) {
	l=(k+j)*lda;
	printf("%8.4e ",A[i+l]);
      }
      printf("\n");
    }
    printf("\n");
  }
  printf("%d: ----------\n",me);
}



