#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_print_result(PLA_Obj A, PLA_Obj Z, PLA_Obj D)
{
/* 
   Print results of PRISM_divide - TEST PHASE.
*/
   int length;
   double local_resid, global_resid, max_a;
   MPI_Comm comm;
   PLA_Obj W=NULL, X=NULL;

   /* Extract base communicator */
   PLA_Base_comm(&comm);

   /* Create temporary objects */
   PLA_Matrix_create_conf_to( A, &W );
   PLA_Matrix_create_conf_to( A, &X );

   /* Print results if size < 10 */
   PLA_Obj_global_length(A, &length);
   if (length < 10) {
      if (me==ROOT) printf("Matrix A:\n"); PRISM_print(A);
      if (me==ROOT) printf("Matrix Z:\n"); PRISM_print(Z);
      if (me==ROOT) printf("Vector D:\n"); PRISM_print(D);
   }

   /* Check that Z'Z = I */
   PRISM_set_to_ident( X );
   PLA_Gemm(PLA_NO_TRANS, PLA_TRANS, PRISM_mone, Z, Z, PRISM_one, X);
   prism_pla_prt_Qresid( X, &local_resid);
   /* printf("me=%d  |Z'Z-I|:: Local = %8.4e ;;;", me,local_resid); */
   MPI_Reduce(&local_resid, &global_resid, 1, MPI_DOUBLE, MPI_MAX, 0, comm);
   MPI_Bcast(&global_resid, 1, MPI_DOUBLE, 0, comm);
   if (me==ROOT) printf("|Z'Z-I|:: Max = %8.4e \n", global_resid);

   /* Check SZ = ZW, W a diagonal matrix containing D */
   prism_pla_prt_Qresid( A, &max_a);
   PLA_Obj_set_to_zero(X);
   PRISM_set_diag( W, D );
   PLA_Gemm(PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, A, Z, PRISM_zero, X);
   PLA_Gemm(PLA_NO_TRANS, PLA_NO_TRANS, PRISM_mone, Z, W, PRISM_one, X);
   prism_pla_prt_Qresid( X, &local_resid);
   /* printf("me=%d  |SZ-ZD|:: Local = %8.4e ;;;", me,local_resid); */
   MPI_Reduce(&local_resid, &global_resid, 1, MPI_DOUBLE, MPI_MAX, 0, comm);
   MPI_Bcast(&global_resid, 1, MPI_DOUBLE, 0, comm);
   if (me==ROOT) printf("|SZ-ZD|:: Max = %8.4e \n", global_resid/max_a);

   /* free objects */
   PLA_Obj_free(&W);
   PLA_Obj_free(&X);
}
